from codeable_models import CClass, add_links, CBundle
from metamodels.guidance_metamodel import includes, practice, decision, \
    add_decision_option_link, \
    do_nothing_design_solution, decide_for_all_instances_of, can_use, \
    consider_if_not_decided_yet, very_negative, \
    positive, force, negative, very_positive, is_a, can_be_combined_with, \
    neutral, \
    decide_for_some_instances_of, can_include, pattern, has_variant, uses, \
    can_be_realized_by, enables, requires, \
    alternative_to, extension, rules_out, \
    positive_and_negative

from add_models.ml_domain_model import ml_model, model_and_application, \
    ml_application


def add_force_relations(force_relations_definition):
    for practice in force_relations_definition.keys():
        for force in force_relations_definition[practice]:
            stereotype = (force_relations_definition[practice])[force]
            add_links({practice: force}, role_name="forces", stereotype_instances=stereotype)


# domain classes

# misc. practices
# s23 - initial context: monitoring
iterative_model_building = CClass(practice, "Iterative model building")
trigger_model_retraining = CClass(practice, "Trigger model retraining")
model_performance_analysis = CClass(practice, "Model performance analysis")

cross_team_collaboration = CClass(practice, "Cross-team collaboration")
production_operations = CClass(practice, "Production operations")
production_maintenance = CClass(practice, "Production maintenance")
production_support = CClass(practice, "Production support")

# s24 - initial context: monitoring
humans_in_the_loop = CClass(practice, "Humans-in-the-loop")
use_monitoring_schedule = CClass(practice, "Use monitoring schedule")
set_trigger_conditions_for_human_loop_activation = CClass(practice,
                                                          "Set trigger conditions for humans-in-the-loop activation")

# forces
division_of_labour = CClass(force, "Division of labour")
distributed_development = CClass(force, "Distributed development")
independent_development = CClass(force, "Independent development")
iterative_development = CClass(force, "Iterative development")
production_ready_development = CClass(force, "Production-ready development")
development_agility = CClass(force, "Development agility")

automated_data_collection = CClass(force, "Automated data collection")
process_and_work_automation = CClass(force, "Process and work automation")

maintainability = CClass(force, "Maintainability")
modifiability = CClass(force, "Modifiability")
robustness = CClass(force, "Robustness")
encapsulation = CClass(force, "Encapsulation")
separation_of_concerns = CClass(force, "Separation of concerns")
model_portability = CClass(force, "Model portability")
interactive_prototyping = CClass(force, "Interactive prototyping")
unified_data_access = CClass(force, "Unified data access")
computational_load = CClass(force, "Computational load")
development_velocity = CClass(force, "Development velocity")
tool_development_effort = CClass(force, "Tool development effort")
deployment_velocity = CClass(force, "Deployment velocity")
modularity = CClass(force, "Modularity")
reusability = CClass(force, "Reusability")
explainability = CClass(force, "Explainability")
testability = CClass(force, "Testability")
auditability = CClass(force, "Auditability")
continuous_improvement = CClass(force, "Continuous improvement")
flexibility = CClass(force, "Flexibility")
handling_massive_amounts_of_data = CClass(force, "Handling massive amounts of data")

dependable_releases = CClass(force, "Dependable releases")
reproducibility = CClass(force, "Reproducibility")
experimental_operational_symmetry = CClass(force, "Experimental operational symmetry")

infrastructure_costs = CClass(force, "Infrastructure costs")

scalability = CClass(force, "Scalability")
independence_of_provisioning = CClass(force, "Independence of provisioning")
auto_scaling = CClass(force, "Auto-scaling")
near_real_time_support = CClass(force, "Near real-time support")
efficiency = CClass(force, "Efficiency")
data_throughput = CClass(force, "Data throughput")
independent_processing = CClass(force, "Independent processing")
reliability = CClass(force, "Reliability")
low_latency = CClass(force, "Low latency")
loose_coupling = CClass(force, "Loose coupling")
chatty_data_access_api = CClass(force, "Chatty data access API")
data_access_flexibility = CClass(force, "Data access flexibility")
data_consistency = CClass(force, "Data consistency")
avoid_cold_start_problem_for_training_data = CClass(force, "Avoid cold start problem for training data")
handling_large_data_files = CClass(force, "Handling large data files")

speed_performance = CClass(force, "Speed performance")
model_accuracy = CClass(force, "Model accuracy")
continuous_model_improvement = CClass(force, "Continuous model improvement")
observability = CClass(force, "Observability")
debugging_support = CClass(force, "Debugging support")
identify_potential_bias_or_overfit = CClass(force, "Identify potential bias or overfit")

offline_support = CClass(force, "Offline support")
online_support = CClass(force, "Online support")

interchangeability_of_ml_algorithms = CClass(force,
                                                "Interchangeability of machine learning algorithms")
feature_reuse = CClass(force, "Feature reuse")
feature_discovery = CClass(force, "Feature discovery")
avoid_training_serving_skew = CClass(force, "Avoid training/serving skew")

safe_model_transition = CClass(force, "Safe transition between production models")
rolling_upgrades = CClass(force, "Rolling upgrades")

react_to_unforeseen_changes_in_the_data = CClass(force, "React to unforeseen changes in the data")
react_to_user_events_and_interactions = CClass(force, "React to user event and interactions")
memory_requirements = CClass(force, "Memory requirements")
variety_of_ml_algorithm_choices = CClass(force, "Variety of machine learning algorithm choices")

# s20
time_to_market_reduction = CClass(force, "Reduced time to market")
improve_user_experience = CClass(force, "Better user experience")
improve_customer_satisfaction = CClass(force, "Better customer satisfaction")
improve_prediction_quality = CClass(force, "Higher quality of predictions")

recognise_model_degradation = CClass(force, "Ability to recognise model degradation")
address_model_degradation = CClass(force, "Ability to address model degradation")

recognise_data_drift = CClass(force, "Ability to identify whether data has changed over time")

improve_model_management = CClass(force, "Improve model management")
change_tracking = CClass(force, "Change tracking")
shadow_mode_testing = CClass(force, "Shadow mode testing")

simplified_deployment = CClass(force, "Simplified deployment")
simplified_model_updates = CClass(force, "Simplified model updates")
improved_client_heterogeneity = CClass(force, "Improved client heterogeneity")
improved_support_for_numerous_clients = CClass(force, "Improved support for numerous clients")
facilitates_testing_in_production = CClass(force, "Facilitates testing in production")

# s21
align_models_with_business_needs = CClass(force, "Align models with business needs")
align_models_with_regulatory_requirements = CClass(force, "Align models with regulatory requirements")
model_consistency = CClass(force, "Model consistency")
unlock_new_sources_of_revenue = CClass(force, "Unlock new sources of revenue")
save_time = CClass(force, "Save time")
reduce_cost = CClass(force, "Reduce cost")
management_agility = CClass(force, "Management agility")
rapid_experimentation = CClass(force, "Rapid experimentation")

# s22
reduce_technical_debt = CClass(force, "Reduce technical debt")

# s23 - initial context: monitoring
model_successful_operation = CClass(force, "Successful model operation")
model_stability = CClass(force, "Model stability")
model_performance = CClass(force, "Model performance")
valuable_models = CClass(force, "Valuable models")
incident_detection = CClass(force, "Incident detection")
incident_resolution = CClass(force, "Incident resolution")
fresh_models = CClass(force, "Fresh models")
risk_mitigation = CClass(force, "Risk mitigation")

# s26 - initial context: development environment
demonstrability = CClass(force, "Demonstrability")
ease_of_hosting = CClass(force, "Ease of hosting")
use_ide_features = CClass(force, "Use IDE features")
code_quality = CClass(force, "Code quality")

# s33 - initial context: integration testing
data_usability = CClass(force, "Data usability")

# s35 - initial context: fata parallelization
error_reduction = CClass(force, "Error reduction")

# decisions, options, and contexts

#
# ** Decision: Integrating ML model **
#
model_as_a_service = CClass(practice, "Model as a service")
model_in_monolith = CClass(practice, "Model in monolith")
model_as_module_in_modular_monolith = CClass(practice, "Model as module in modular monolith")
model_in_batch_prediction_system = CClass(practice, "Model in batch prediction system")
event_driven_model_as_a_service = CClass(practice, "Event-driven model as a service")
serverless_model_as_a_service = CClass(practice, "Serverless model as a service")
no_integration_of_model = CClass(do_nothing_design_solution, "No integration of the model")

model_as_module_in_modular_monolith_is_a_model_in_monolith = \
    model_as_module_in_modular_monolith.add_links(model_in_monolith, role_name="to", stereotype_instances=is_a)[0]

event_driven_model_as_a_service_is_a_model_as_a_service = \
    event_driven_model_as_a_service.add_links(model_as_a_service, role_name="to", stereotype_instances=is_a)[0]

serverless_model_as_a_service_is_a_model_as_a_service = \
    serverless_model_as_a_service.add_links(model_as_a_service, role_name="to", stereotype_instances=is_a)[0]

integrating_ml_model_decision = CClass(decision,
                                       "How to integrate a machine learning model in an application?")
add_decision_option_link(integrating_ml_model_decision, model_as_a_service,
                         "create one service per model")
add_decision_option_link(integrating_ml_model_decision, model_in_monolith,
                         "embed the model in the code of the application")
add_decision_option_link(integrating_ml_model_decision, model_as_module_in_modular_monolith,
                         "create a module for the model in the application")
add_decision_option_link(integrating_ml_model_decision, model_in_batch_prediction_system,
                         "call an external batch prediction system")
add_decision_option_link(integrating_ml_model_decision, event_driven_model_as_a_service,
                         "event-driven pub/sub or queueing interface for model")
add_decision_option_link(integrating_ml_model_decision, serverless_model_as_a_service,
                         "serverless/function-as-a-service computing model for offering the model")
add_decision_option_link(integrating_ml_model_decision, no_integration_of_model,
                         "model is not integrated, maybe results are used, e.g. as a forecast")

add_force_relations({model_as_a_service: {division_of_labour: very_positive,
                                          distributed_development: very_positive,
                                          independent_development: very_positive,
                                          scalability: very_positive,
                                          offline_support: negative,
                                          online_support: positive,
                                          model_portability: very_positive,
                                          low_latency: positive,
                                          data_throughput: positive,
                                          encapsulation: very_positive,
                                          separation_of_concerns: very_positive,
                                          modularity: very_positive,
                                          reusability: very_positive,
                                          loose_coupling: very_positive,
                                          production_ready_development: very_positive,
                                          simplified_deployment: positive,
                                          simplified_model_updates: positive,
                                          improved_client_heterogeneity: positive,
                                          improved_support_for_numerous_clients: positive,
                                          facilitates_testing_in_production: positive,
                                          model_consistency: positive
                                          },
                     model_in_monolith: {division_of_labour: negative,
                                         distributed_development: negative,
                                         independent_development: negative,
                                         scalability: negative,
                                         offline_support: very_positive,
                                         online_support: neutral,
                                         model_portability: very_negative,
                                         low_latency: neutral,
                                         data_throughput: negative,
                                         encapsulation: very_negative,
                                         separation_of_concerns: very_negative,
                                         modularity: very_negative,
                                         reusability: negative,
                                         loose_coupling: very_negative,
                                         production_ready_development: neutral,
                                         simplified_deployment: neutral,
                                         simplified_model_updates: neutral,
                                         improved_client_heterogeneity: neutral,
                                         improved_support_for_numerous_clients: neutral,
                                         facilitates_testing_in_production: neutral,
                                         model_consistency: neutral,
                                         },
                     model_as_module_in_modular_monolith: {division_of_labour: neutral,
                                                           distributed_development: neutral,
                                                           independent_development: neutral,
                                                           scalability: neutral,
                                                           offline_support: very_positive,
                                                           online_support: neutral,
                                                           model_portability: neutral,
                                                           low_latency: neutral,
                                                           data_throughput: negative,
                                                           encapsulation: positive,
                                                           separation_of_concerns: positive,
                                                           modularity: positive,
                                                           reusability: positive,
                                                           loose_coupling: positive,
                                                           production_ready_development: positive,
                                                           simplified_deployment: neutral,
                                                           simplified_model_updates: neutral,
                                                           improved_client_heterogeneity: neutral,
                                                           improved_support_for_numerous_clients: neutral,
                                                           facilitates_testing_in_production: neutral
                                                           },
                     model_in_batch_prediction_system: {division_of_labour: neutral,
                                                        distributed_development: neutral,
                                                        independent_development: neutral,
                                                        scalability: negative,
                                                        offline_support: neutral,
                                                        online_support: negative,
                                                        model_portability: neutral,
                                                        low_latency: negative,
                                                        data_throughput: negative,
                                                        encapsulation: neutral,
                                                        separation_of_concerns: positive,
                                                        modularity: neutral,
                                                        reusability: neutral,
                                                        loose_coupling: neutral,
                                                        production_ready_development: neutral,
                                                        simplified_deployment: neutral,
                                                        simplified_model_updates: neutral,
                                                        improved_client_heterogeneity: neutral,
                                                        improved_support_for_numerous_clients: neutral,
                                                        facilitates_testing_in_production: neutral
                                                        },
                     event_driven_model_as_a_service: {division_of_labour: very_positive,
                                                       distributed_development: very_positive,
                                                       independent_development: very_positive,
                                                       scalability: very_positive,
                                                       offline_support: negative,
                                                       online_support: positive,
                                                       model_portability: very_positive,
                                                       low_latency: positive,
                                                       data_throughput: positive,
                                                       encapsulation: very_positive,
                                                       separation_of_concerns: very_positive,
                                                       modularity: very_positive,
                                                       reusability: very_positive,
                                                       loose_coupling: very_positive,
                                                       production_ready_development: very_positive,
                                                       simplified_deployment: neutral,
                                                       simplified_model_updates: neutral,
                                                       improved_client_heterogeneity: neutral,
                                                       improved_support_for_numerous_clients: neutral,
                                                       facilitates_testing_in_production: neutral
                                                       },
                     serverless_model_as_a_service: {division_of_labour: very_positive,
                                                     distributed_development: very_positive,
                                                     independent_development: very_positive,
                                                     scalability: very_positive,
                                                     offline_support: negative,
                                                     online_support: positive,
                                                     model_portability: very_positive,
                                                     low_latency: positive,
                                                     data_throughput: positive,
                                                     encapsulation: very_positive,
                                                     separation_of_concerns: very_positive,
                                                     modularity: very_positive,
                                                     reusability: very_positive,
                                                     loose_coupling: very_positive,
                                                     production_ready_development: very_positive,
                                                     simplified_deployment: neutral,
                                                     simplified_model_updates: neutral,
                                                     improved_client_heterogeneity: neutral,
                                                     improved_support_for_numerous_clients: neutral,
                                                     facilitates_testing_in_production: neutral
                                                     },
                     no_integration_of_model: {division_of_labour: negative,
                                               distributed_development: negative,
                                               independent_development: negative,
                                               scalability: very_negative,
                                               offline_support: positive,
                                               online_support: very_negative,
                                               model_portability: very_negative,
                                               low_latency: neutral,
                                               data_throughput: neutral,
                                               encapsulation: neutral,
                                               separation_of_concerns: negative,
                                               modularity: negative,
                                               reusability: very_negative,
                                               loose_coupling: negative,
                                               production_ready_development: very_negative,
                                               simplified_deployment: neutral,
                                               simplified_model_updates: neutral,
                                               improved_client_heterogeneity: neutral,
                                               improved_support_for_numerous_clients: neutral,
                                               facilitates_testing_in_production: neutral
                                               },
                     })

# context
add_links({integrating_ml_model_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_all_instances_of)

#
# ** Decision: store data  **
#
shared_data_store = CClass(pattern, "Shared data store")
data_warehouse = CClass(practice, "Data warehouse")
data_lake = CClass(practice, "Data lake")
shared_database = CClass(practice, "Shared database")
in_memory_data_store = CClass(practice, "In-memory data store")
real_time_database = CClass(practice, "Real-time database")

data_service = CClass(pattern, "Data service")
database_per_service = CClass(pattern, "Database per service")

event_store = CClass(practice, "Event store")
event_sourcing = CClass(practice, "Event sourcing")

no_data_store = CClass(do_nothing_design_solution, "No data store")

data_warehouse_is_a_shared_data_store = \
    data_warehouse.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]
data_lake_is_a_shared_data_store = \
    data_lake.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]
shared_database_is_a_shared_data_store = \
    shared_database.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]
real_time_database_is_a_shared_data_store = \
    real_time_database.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]
event_store_is_a_shared_data_store = \
    event_store.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]

event_store_can_offer_event_sourcing = \
    event_store.add_links(event_sourcing, role_name="to", stereotype_instances=can_be_combined_with,
                          label="can offer event sourcing capabilities")[0]
data_service_can_use_db_per_service = \
    data_service.add_links(database_per_service, role_name="to", stereotype_instances=can_use,
                           label="can have per service database")[0]

database_per_service_has_variant_in_memory_data_store = \
    database_per_service.add_links(in_memory_data_store, role_name="to", stereotype_instances=has_variant,
                                   label="used for low-latency online results")[0]
shared_database_has_variant_in_memory_data_store = \
    shared_database.add_links(in_memory_data_store, role_name="to", stereotype_instances=has_variant,
                              label="used for low-latency online results")[0]

store_data_decision = CClass(decision, "How to store data used in machine learning applications?")
add_decision_option_link(store_data_decision, no_data_store,
                         "get data without storing it, e.g. manual ingested raw data or only event streaming")
add_decision_option_link(store_data_decision, shared_data_store,
                         "store data in shared data store")
add_decision_option_link(store_data_decision, data_service,
                         "store data in services")
add_decision_option_link(store_data_decision, event_store,
                         "store data as events being exchanged in the systems")

add_force_relations({no_data_store: {chatty_data_access_api: positive,
                                     speed_performance: positive,
                                     scalability: positive,
                                     data_access_flexibility: negative,
                                     data_consistency: very_negative,
                                     maintainability: very_negative,
                                     modifiability: very_negative,
                                     reliability: neutral,
                                     loose_coupling: neutral,
                                     avoid_cold_start_problem_for_training_data: negative,
                                     computational_load: positive,
                                     infrastructure_costs: very_positive,
                                     development_velocity: very_negative,
                                     },
                     shared_data_store: {chatty_data_access_api: positive,
                                         speed_performance: positive,
                                         scalability: negative,
                                         data_access_flexibility: positive,
                                         data_consistency: very_positive,
                                         maintainability: neutral,
                                         modifiability: neutral,
                                         reliability: neutral,
                                         loose_coupling: very_negative,
                                         avoid_cold_start_problem_for_training_data: very_positive,
                                         computational_load: positive,
                                         infrastructure_costs: positive,
                                         development_velocity: neutral,
                                         },
                     data_service: {chatty_data_access_api: negative,
                                    speed_performance: negative,
                                    scalability: negative,
                                    data_access_flexibility: negative,
                                    data_consistency: negative,
                                    maintainability: positive,
                                    modifiability: neutral,
                                    reliability: neutral,
                                    loose_coupling: positive,
                                    avoid_cold_start_problem_for_training_data: positive,
                                    computational_load: neutral,
                                    infrastructure_costs: neutral,
                                    development_velocity: positive,
                                    },
                     event_store: {chatty_data_access_api: positive,
                                   speed_performance: positive,
                                   scalability: positive,
                                   data_access_flexibility: positive,
                                   data_consistency: positive,
                                   maintainability: positive,
                                   modifiability: positive,
                                   reliability: positive,
                                   loose_coupling: very_positive,
                                   avoid_cold_start_problem_for_training_data: positive,
                                   computational_load: negative,
                                   infrastructure_costs: negative,
                                   development_velocity: very_positive,
                                   },
                     })

# context
add_links({store_data_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

#
# ** Decision: How to design the interface for data access?  **
#
local_data_access_api = CClass(practice, "Local data access API")
service_based_data_access_api = CClass(practice, "Service-based data access API")
pub_sub_data_access_api = CClass(practice, "Pub/sub data access API")
message_event_queue_based_data_access_api = CClass(practice, "Message/event queueing-based data access API")
direct_data_provider_access = CClass(do_nothing_design_solution, "Data provider is accessed directly")

service_based_data_access_api_contains_data_services = \
    service_based_data_access_api.add_links(data_service, role_name="to", stereotype_instances=includes,
                                            label="data services and/or wrapper services for other kinds of data stores")[
        0]

data_access_interface_decision = CClass(decision, "How to design the interface for data access? ")

add_decision_option_link(data_access_interface_decision, direct_data_provider_access)
add_decision_option_link(data_access_interface_decision, local_data_access_api,
                         "data provider encapsulated by local data access API abstraction")
add_decision_option_link(data_access_interface_decision, service_based_data_access_api,
                         "data provider encapsulated by data access services")
add_decision_option_link(data_access_interface_decision, pub_sub_data_access_api,
                         "subscription for data access; publisher notifies upon data availability")
add_decision_option_link(data_access_interface_decision, message_event_queue_based_data_access_api,
                         "new data queued in messages/events; message/event raised upon data availability")

add_force_relations({direct_data_provider_access: {unified_data_access: negative,
                                                   near_real_time_support: very_negative,
                                                   reliability: neutral,
                                                   low_latency: very_negative,
                                                   data_throughput: very_negative,
                                                   observability: negative
                                                   },
                     local_data_access_api: {unified_data_access: positive,
                                             near_real_time_support: negative,
                                             reliability: neutral,
                                             low_latency: very_negative,
                                             data_throughput: very_negative,
                                             observability: positive,
                                             },
                     service_based_data_access_api: {unified_data_access: positive,
                                                     near_real_time_support: neutral,
                                                     reliability: neutral,
                                                     low_latency: positive,
                                                     data_throughput: positive,
                                                     observability: positive
                                                     },
                     pub_sub_data_access_api: {unified_data_access: positive,
                                               near_real_time_support: very_positive,
                                               reliability: positive,
                                               low_latency: very_positive,
                                               data_throughput: very_positive,
                                               observability: positive
                                               },
                     message_event_queue_based_data_access_api: {unified_data_access: positive,
                                                                 near_real_time_support: positive,
                                                                 reliability: positive,
                                                                 low_latency: very_positive,
                                                                 data_throughput: very_positive,
                                                                 observability: positive
                                                                 },
                     })

add_links({data_access_interface_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({store_data_decision: data_access_interface_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="offer an abstraction layer for data access")

#
# ** Decision:  persist and provide access to features  **
#
feature_store = CClass(practice, "Feature store")
batch_feature_api = CClass(practice, "Batch feature API")
real_time_feature_api = CClass(practice, "Real-time feature API")
store_features_without_specific_support = \
    CClass(do_nothing_design_solution, "Store features in data stores without specific support")

feature_store_can_be_realized_by_event_sourcing = \
    feature_store.add_links(event_sourcing, role_name="to", stereotype_instances=can_be_realized_by)[0]

feature_store_can_include_batch_feature_api = \
    feature_store.add_links(batch_feature_api, role_name="to", stereotype_instances=can_include)[0]
feature_store_can_include_real_time_feature_api = \
    feature_store.add_links(real_time_feature_api, role_name="to", stereotype_instances=can_include)[0]

feature_persistence_and_access_decision = CClass(decision, "How to persist and provide access to features?")
add_decision_option_link(feature_persistence_and_access_decision, store_features_without_specific_support)
add_decision_option_link(feature_persistence_and_access_decision, feature_store)

add_force_relations({store_features_without_specific_support: {feature_reuse: negative,
                                                               feature_discovery: negative,
                                                               maintainability: negative,
                                                               avoid_training_serving_skew: negative,
                                                               online_support: neutral,
                                                               offline_support: neutral,
                                                               data_throughput: neutral,
                                                               low_latency: neutral,
                                                               near_real_time_support: neutral
                                                               },
                     feature_store: {feature_reuse: positive,
                                     feature_discovery: positive,
                                     maintainability: positive,
                                     avoid_training_serving_skew: positive,
                                     online_support: positive,
                                     offline_support: positive,
                                     data_throughput: positive,
                                     low_latency: positive,
                                     near_real_time_support: positive
                                     },
                     })

add_links({feature_persistence_and_access_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({feature_store: store_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to persist the features")

add_links({store_features_without_specific_support: store_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to persist the features")

add_links({batch_feature_api: data_access_interface_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to design data access api")
add_links({real_time_feature_api: data_access_interface_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to design data access api")

#
# ** Decision: Model Serving Decision (i.e. providing ML predictions or scores) **
#

# please note: can be a distributed service (i.e., ml_service) or DDD service in monolith.
service_based_model_serving = CClass(practice, "Service based model serving")
microservice_based_model_serving = CClass(practice, "Microservice based model serving")
local_service_api_based_model_serving = CClass(practice, "Local service/API based model serving")
data_batch_based_model_serving = CClass(practice, "Data batch based model serving")
precompute_serving = CClass(practice, "Precompute serving")
predictions_data_store = CClass(practice, "Predictions data store")
hybrid_online_offline_model_serving = CClass(practice, "Hybrid online/offline model serving")

microservice_based_model_serving_uses_model_as_a_service = \
    microservice_based_model_serving.add_links(model_as_a_service, role_name="to", stereotype_instances=uses,
                                               label="distributed service or microservice")[0]

local_service_api_based_model_serving_uses_model_as_a_service = \
    local_service_api_based_model_serving.add_links(model_as_module_in_modular_monolith, role_name="to",
                                                    stereotype_instances=uses, label="service as local module")[0]

microservice_based_model_serving_is_a_service_based_model_serving = \
    microservice_based_model_serving.add_links(service_based_model_serving, role_name="to",
                                               stereotype_instances=is_a)[0]
local_service_api_based_model_serving_is_a_service_based_model_serving = \
    local_service_api_based_model_serving.add_links(service_based_model_serving, role_name="to",
                                                    stereotype_instances=is_a)[0]

service_based_model_serving_can_use_predictions_data_store = \
    service_based_model_serving.add_links(predictions_data_store, role_name="to",
                                          stereotype_instances=can_use,
                                          label="persist predictions")[0]
data_batch_based_model_serving_can_use_predictions_data_store = \
    data_batch_based_model_serving.add_links(predictions_data_store, role_name="to",
                                             stereotype_instances=can_use,
                                             label="persist predictions")[0]

hybrid_online_offline_model_serving_can_use_data_batch_based_model_serving = \
    hybrid_online_offline_model_serving.add_links(data_batch_based_model_serving, role_name="to",
                                                  stereotype_instances=can_use,
                                                  label="precalculate parts of the results or " +
                                                        "use offline processing as a fallback")[0]
hybrid_online_offline_model_serving_can_use_service_based_model_serving = \
    hybrid_online_offline_model_serving.add_links(service_based_model_serving, role_name="to",
                                                  stereotype_instances=can_use,
                                                  label="use services to provide online computations")[0]

model_serving_decision = CClass(decision, "How to serve model predictions or scores?")
add_decision_option_link(model_serving_decision, microservice_based_model_serving,
                         "deliver predictions or scores in real time by distributed service")
add_decision_option_link(model_serving_decision, local_service_api_based_model_serving,
                         "deliver predictions or scores in real time by local service")
add_decision_option_link(model_serving_decision, data_batch_based_model_serving,
                         "process data in batches on schedule and write predictions or scores to data storage")
add_decision_option_link(model_serving_decision, hybrid_online_offline_model_serving,
                         "combine online computations by services with offline computations in batches")
add_decision_option_link(model_serving_decision, precompute_serving,
                         "precompute models e.g. as forecasts and deliver as reports")

add_force_relations({microservice_based_model_serving: {near_real_time_support: very_positive,
                                                        efficiency: positive,
                                                        maintainability: very_positive,
                                                        scalability: very_positive,
                                                        online_support: very_positive,
                                                        offline_support: negative,
                                                        loose_coupling: very_positive,
                                                        robustness: neutral,
                                                        modularity: very_positive,
                                                        react_to_user_events_and_interactions: positive,
                                                        variety_of_ml_algorithm_choices: negative,
                                                        data_throughput: negative
                                                        },
                     local_service_api_based_model_serving: {near_real_time_support: very_positive,
                                                             efficiency: neutral,
                                                             maintainability: positive,
                                                             scalability: neutral,
                                                             online_support: very_positive,
                                                             offline_support: positive,
                                                             loose_coupling: positive,
                                                             robustness: neutral,
                                                             modularity: positive,
                                                             react_to_user_events_and_interactions: positive,
                                                             variety_of_ml_algorithm_choices: negative,
                                                             data_throughput: negative
                                                             },
                     data_batch_based_model_serving: {near_real_time_support: very_negative,
                                                      efficiency: very_negative,
                                                      maintainability: negative,
                                                      scalability: very_negative,
                                                      online_support: negative,
                                                      offline_support: positive,
                                                      loose_coupling: neutral,
                                                      robustness: neutral,
                                                      modularity: neutral,
                                                      react_to_user_events_and_interactions: very_negative,
                                                      variety_of_ml_algorithm_choices: positive,
                                                      data_throughput: very_positive
                                                      },
                     precompute_serving: {near_real_time_support: negative,
                                          efficiency: negative,
                                          maintainability: negative,
                                          scalability: very_negative,
                                          online_support: negative,
                                          offline_support: positive,
                                          loose_coupling: negative,
                                          robustness: neutral,
                                          modularity: very_negative,
                                          react_to_user_events_and_interactions: very_negative,
                                          variety_of_ml_algorithm_choices: positive,
                                          data_throughput: positive
                                          },
                     hybrid_online_offline_model_serving: {near_real_time_support: positive,
                                                           efficiency: positive,
                                                           maintainability: negative,
                                                           scalability: positive,
                                                           online_support: very_positive,
                                                           offline_support: very_positive,
                                                           loose_coupling: positive,
                                                           robustness: very_positive,
                                                           modularity: positive,
                                                           react_to_user_events_and_interactions: positive,
                                                           variety_of_ml_algorithm_choices: very_negative,
                                                           data_throughput: neutral
                                                           }
                     })

add_links({model_serving_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({integrating_ml_model_decision: model_serving_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

add_links({predictions_data_store: [store_data_decision]},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to persist predictions")

add_links({model_serving_decision: feature_persistence_and_access_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to access features for model serving")

#
# ** Decision: How to inform about prediction results availability **
#
immediate_result_notification = CClass(force, "Immediate result notification")
non_blocking_clients = CClass(force, "Non-blocking clients")
clients_must_follow_asynchronous_model = CClass(force, "Clients must follow asynchronous programming model")

push_notification = CClass(practice, "Push notification")
polling = CClass(practice, "Polling")
block_on_result = CClass(practice, "Block on result")

result_availability_information_decision = CClass(decision,
                                                  "How to inform clients about prediction result availability?")
add_decision_option_link(result_availability_information_decision, push_notification,
                         "notify clients about results")
add_decision_option_link(result_availability_information_decision, polling,
                         "clients poll for the results")
add_decision_option_link(result_availability_information_decision, block_on_result,
                         "clients blocks until the results are available")

add_force_relations({push_notification: {immediate_result_notification: positive,
                                         non_blocking_clients: positive,
                                         clients_must_follow_asynchronous_model: negative
                                         },
                     polling: {immediate_result_notification: negative,
                               non_blocking_clients: positive,
                               clients_must_follow_asynchronous_model: positive
                               },
                     block_on_result: {immediate_result_notification: very_negative,
                                       non_blocking_clients: very_negative,
                                       clients_must_follow_asynchronous_model: positive
                                       },

                     })

add_links({result_availability_information_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({model_serving_decision: result_availability_information_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: data ingestion  **
#
streamed_data_ingestion = CClass(practice, "Streamed data ingestion")
data_ingestion_by_request = CClass(practice, "Data ingestion by request")

data_ingestor = CClass(practice, "Data ingestor")
ground_truth_collector = CClass(practice, "Ground-truth collector")
data_ingestion_queue = CClass(practice, "Data ingestion queue")
data_ingestion_in_batches = CClass(practice, "Data ingestion in batches")
manual_data_ingestion = CClass(practice, "Manual data ingestion")
automatic_data_labeller = CClass(practice, "Automatic data labeller")
manual_data_labelling = CClass(practice, "Manual data labelling")
raw_data_store = CClass(practice, "Raw data store")

streamed_data_ingestion_can_be_realized_by_data_ingestor = \
    streamed_data_ingestion.add_links(data_ingestor, role_name="to", stereotype_instances=can_be_realized_by,
                                      label="data ingestion service or component")[0]
data_ingestion_by_request_can_be_realized_by_data_ingestor = \
    data_ingestion_by_request.add_links(data_ingestor, role_name="to", stereotype_instances=can_be_realized_by,
                                        label="data ingestion service or component")[0]
data_ingestion_in_batches_can_be_realized_by_data_ingestor = \
    data_ingestion_in_batches.add_links(data_ingestor, role_name="to", stereotype_instances=can_be_realized_by,
                                        label="data ingestion service or component; requires fully automated " +
                                              "batch ingestion")[0]

ground_truth_collector_is_an_data_ingestor = \
    ground_truth_collector.add_links(data_ingestor, role_name="to", stereotype_instances=is_a,
                                     label="service to collect data on predictions in production, " +
                                           "e.g. from users when predictions are wrong")[0]
data_ingestion_queue_is_an_data_ingestor = \
    data_ingestion_queue.add_links(data_ingestor, role_name="to", stereotype_instances=is_a,
                                   label="queue up the ingested data, e.g for online incremental learning")[0]

data_ingestor_can_use_data_labeller = \
    data_ingestor.add_links(automatic_data_labeller, role_name="to", stereotype_instances=can_use,
                            label="use service to perform automatic data labelling")[0]
data_ingestor_can_use_manual_data_labelling = \
    data_ingestor.add_links(manual_data_labelling, role_name="to", stereotype_instances=can_use,
                            label="data is manually labelled by humans")[0]
data_labeller_alternative_to_manual_data_labelling = \
    automatic_data_labeller.add_links(manual_data_labelling, role_name="to", stereotype_instances=alternative_to)[0]

streamed_data_ingestion_can_use_raw_data_store = \
    streamed_data_ingestion.add_links(raw_data_store, role_name="to", stereotype_instances=can_use)[0]
data_ingestion_by_request_can_use_raw_data_store = \
    data_ingestion_by_request.add_links(raw_data_store, role_name="to", stereotype_instances=can_use)[0]
data_ingestion_in_batches_can_use_raw_data_store = \
    data_ingestion_in_batches.add_links(raw_data_store, role_name="to", stereotype_instances=can_use)[0]
manual_data_ingestion_can_use_raw_data_store = \
    manual_data_ingestion.add_links(raw_data_store, role_name="to", stereotype_instances=can_use)[0]

data_ingestion_decision = CClass(decision, "How to ingest data into ML projects or applications?")
add_decision_option_link(data_ingestion_decision, streamed_data_ingestion,
                         "receive a continuous stream of data")
add_decision_option_link(data_ingestion_decision, data_ingestion_by_request,
                         "actively get raw data from data sources")
add_decision_option_link(data_ingestion_decision, data_ingestion_in_batches,
                         "get data in one or more batches, possibly only in part automated")
add_decision_option_link(data_ingestion_decision, manual_data_ingestion,
                         "provide data for ingestion manually")

add_links({data_ingestion_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({raw_data_store: [store_data_decision]},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to store raw ingested data")

add_force_relations({streamed_data_ingestion: {online_support: positive,
                                               offline_support: negative,
                                               data_throughput: positive,
                                               low_latency: very_positive,
                                               near_real_time_support: very_positive,
                                               tool_development_effort: negative,
                                               },
                     data_ingestion_by_request: {online_support: neutral,
                                                 offline_support: positive,
                                                 data_throughput: neutral,
                                                 low_latency: neutral,
                                                 near_real_time_support: positive,
                                                 tool_development_effort: negative,
                                                 },
                     data_ingestion_in_batches: {online_support: negative,
                                                 offline_support: neutral,
                                                 data_throughput: very_positive,
                                                 low_latency: negative,
                                                 near_real_time_support: very_negative,
                                                 tool_development_effort: negative,
                                                 },
                     manual_data_ingestion: {online_support: very_negative,
                                             offline_support: negative,
                                             data_throughput: very_negative,
                                             low_latency: very_negative,
                                             near_real_time_support: very_negative,
                                             tool_development_effort: positive,
                                             }
                     })

add_links({model_serving_decision: data_ingestion_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="ingest data required during model serving")

#
# ** Decision: How to automate integration and delivery ***
#
ci_cd_pipeline = CClass(practice, "CI/CD pipeline")
build_and_deployment_scripts = CClass(practice, "Build and deployment scripts")
ml_metadata_store = CClass(practice, "Machine learning metadata store")
model_registry = CClass(practice, "Model registry")
ml_orchestrator = CClass(practice, "Machine learning orchestrator")
no_integration_or_delivery_automation = CClass(do_nothing_design_solution, "No integration or delivery automation")

ci_cd_pipeline_can_use_ml_metadata_store = \
    ci_cd_pipeline.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                             label="store all pipeline parameters and versions as metadata")[0]
ci_cd_pipeline_can_use_model_registry = \
    ci_cd_pipeline.add_links(model_registry, role_name="to", stereotype_instances=can_use,
                             label="access models and model candidates")[0]
ci_cd_pipeline_can_be_combined_with_ml_orchestrator = \
    ci_cd_pipeline.add_links(ml_orchestrator, role_name="to", stereotype_instances=can_be_combined_with,
                             label="orchestration of pipelines and experiments")[0]

automated_integration_and_delivery_decision = CClass(decision, "How to automate integration and " +
                                                     "delivery in a machine learning context?")
add_decision_option_link(automated_integration_and_delivery_decision, no_integration_or_delivery_automation)
add_decision_option_link(automated_integration_and_delivery_decision, build_and_deployment_scripts,
                         "manually perform automated deployments")
add_decision_option_link(automated_integration_and_delivery_decision, ci_cd_pipeline,
                         "continuously automate the deployment e.g. upon commits")
add_decision_option_link(automated_integration_and_delivery_decision, ml_orchestrator,
                         "orchestrate multiple pipelines and/or components")

add_links({automated_integration_and_delivery_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({no_integration_or_delivery_automation: {process_and_work_automation: very_negative,
                                                             iterative_development: negative,
                                                             development_velocity: negative,
                                                             deployment_velocity: very_negative,
                                                             dependable_releases: very_negative,
                                                             reliability: very_negative,
                                                             independent_development: negative,
                                                             modifiability: very_negative,
                                                             reproducibility: very_negative,
                                                             testability: negative,
                                                             auditability: very_negative,
                                                             continuous_improvement: very_negative
                                                             },
                     build_and_deployment_scripts: {process_and_work_automation: positive,
                                                    iterative_development: positive,
                                                    development_velocity: positive,
                                                    deployment_velocity: very_positive,
                                                    dependable_releases: negative,
                                                    reliability: negative,
                                                    reproducibility: positive,
                                                    independent_development: neutral,
                                                    modifiability: negative,
                                                    testability: neutral,
                                                    auditability: negative,
                                                    continuous_improvement: negative
                                                    },
                     ci_cd_pipeline: {process_and_work_automation: very_positive,
                                      iterative_development: very_positive,
                                      development_velocity: very_positive,
                                      deployment_velocity: very_positive,
                                      dependable_releases: very_positive,
                                      reliability: positive,
                                      independent_development: positive,
                                      modifiability: very_positive,
                                      reproducibility: very_positive,
                                      testability: very_positive,
                                      auditability: very_positive,
                                      continuous_improvement: very_positive
                                      },
                     ml_orchestrator: {process_and_work_automation: positive,
                                          recognise_data_drift: positive
                                          }
                     })

add_links({ml_metadata_store: store_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

add_links({model_registry: store_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: Which tasks can be performed by a delivery pipeline or component? ***
#
packaging = CClass(practice, "Packaging")
testing = CClass(practice, "Testing")
building = CClass(practice, "Building")
deployment = CClass(practice, "Deployment")
containerization = CClass(practice, "Containerization")
canary_testing = CClass(practice, "Canary testing")
a_b_testing = CClass(practice, "A/B testing")
hypothesis_testing = CClass(practice, "Hypothesis testing")

version_control_repository = CClass(practice, "Version control repository")

delivery_tasks_decision = CClass(decision, "Which tasks can be performed by a delivery pipeline or component?")

add_decision_option_link(delivery_tasks_decision, packaging)
add_decision_option_link(delivery_tasks_decision, testing)
add_decision_option_link(delivery_tasks_decision, building)
add_decision_option_link(delivery_tasks_decision, deployment)
add_decision_option_link(delivery_tasks_decision, containerization)

add_links({delivery_tasks_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({ci_cd_pipeline: delivery_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the delivery pipeline")

add_links({build_and_deployment_scripts: delivery_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the deployment scripts")

testing.add_links(canary_testing, role_name="to", stereotype_instances=can_include,
                             label="testing technique")[0]

testing.add_links(a_b_testing, role_name="to", stereotype_instances=can_include,
                             label="testing technique")[0]

testing.add_links(hypothesis_testing, role_name="to", stereotype_instances=can_include,
                             label="testing technique")[0]

# extending the integration decision
model_as_a_service_can_use_containerization = \
    model_as_a_service.add_links(containerization, role_name="to", stereotype_instances=can_use)[0]

#
# ** Decision: model versions in production ***
#
n_versions_in_production = CClass(pattern, "N versions in production")
rollback_to_previous_model_version = CClass(practice, "Rollback to previous model version")
single_model_in_production = CClass(do_nothing_design_solution, "Single model in production")

deploying_model_versions_in_production_decision = CClass(decision, "Which model versions should be deployed " +
                                                         "and how?")

add_decision_option_link(deploying_model_versions_in_production_decision, single_model_in_production)
add_decision_option_link(deploying_model_versions_in_production_decision, n_versions_in_production,
                         "deploy multiple model versions at the same time")
add_decision_option_link(deploying_model_versions_in_production_decision, rollback_to_previous_model_version,
                         "redeploy previous model version and then switch to previous model")

rollback_to_previous_model_version_can_use_n_versions_in_production = \
    rollback_to_previous_model_version.add_links(n_versions_in_production, role_name="to", stereotype_instances=can_use,
                                                 label="keep previous version in production to " +
                                                       "enable immediate rollback")[0]

n_versions_in_production_can_use_metadata_store = \
    n_versions_in_production.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                       label="manage version metadata")[0]
rollback_to_previous_model_version_can_use_metadata_store = \
    rollback_to_previous_model_version.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                                 label="manage version metadata")[0]
single_model_in_production_can_use_metadata_store = \
    single_model_in_production.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                         label="manage version metadata")[0]

add_links({deploying_model_versions_in_production_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({single_model_in_production: {safe_model_transition: very_negative,
                                                  rolling_upgrades: very_negative
                                                  },
                     n_versions_in_production: {safe_model_transition: very_positive,
                                                rolling_upgrades: very_positive,
                                                shadow_mode_testing: very_positive
                                                },
                     rollback_to_previous_model_version: {safe_model_transition: positive,
                                                          rolling_upgrades: positive},
                     })

n_versions_in_production_enables_canary_testing = n_versions_in_production.add_links(canary_testing, role_name="to", stereotype_instances=enables,
                    label="canary testing")[0]
n_versions_in_production_enables_a_b_testing = n_versions_in_production.add_links(a_b_testing, role_name="to", stereotype_instances=enables,
                    label="A/B testing")[0]


add_links({automated_integration_and_delivery_decision: deploying_model_versions_in_production_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: how to version data ***
#
data_version_repository = CClass(pattern, "Data version repository")
data_versioning_in_code_repository = CClass(pattern, "Data versioning in code repository")
no_data_versioning = CClass(do_nothing_design_solution, "No data versioning")

deploying_how_to_version_data_decision = CClass(decision, "How to version data?")

data_versioning_in_code_repository_uses_version_control_repository = \
    data_versioning_in_code_repository.add_links(version_control_repository, role_name="to",
                                                 stereotype_instances=uses)[0]

add_decision_option_link(deploying_how_to_version_data_decision, no_data_versioning)
add_decision_option_link(deploying_how_to_version_data_decision, data_version_repository,
                         "version repository with specific technology-support for data versioning")
add_decision_option_link(deploying_how_to_version_data_decision, data_versioning_in_code_repository)

add_links({deploying_how_to_version_data_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({no_data_versioning: {handling_large_data_files: negative,
                                          reproducibility: very_negative
                                          },
                     data_versioning_in_code_repository: {handling_large_data_files: very_negative,
                                                          reproducibility: very_negative,
                                                          development_velocity: very_negative,
                                                          flexibility: very_negative
                                                          },
                     data_version_repository: {handling_large_data_files: very_positive,
                                               reproducibility: very_positive,
                                               development_velocity: very_positive
                                               }
                     })

add_links({automated_integration_and_delivery_decision: deploying_how_to_version_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: how to trigger the machine learning pipeline or orchestrator ***
#
on_demand_trigger = CClass(practice, "On-demand trigger")
on_commit_trigger = CClass(practice, "On commit trigger")
on_schedule_trigger = CClass(practice, "On schedule trigger")
on_availability_of_new_training_data_trigger = CClass(practice, "On availability of new training data trigger")
on_model_performance_degradation_trigger = CClass(practice, "On model performance degradation trigger")
on_changes_in_the_data_distribution_trigger = CClass(practice, "On changes in the data distribution trigger")

trigger_pipeline_or_orchestrator_decision = CClass(decision, "How to trigger a machine learning " +
                                                   "pipeline or orchestrator?")

add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_demand_trigger,
                         "ad-hoc manual execution of the pipeline, or triggered by other pipeline or service")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_commit_trigger,
                         "when a commit to a source or data version control repository happens")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_schedule_trigger,
                         "frequent pipeline runs e.g. on a daily, weekly, or monthly basis")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_availability_of_new_training_data_trigger,
                         "when new data is collected and made available in data stores")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_model_performance_degradation_trigger,
                         "when a noticeable performance degradation is observed")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_changes_in_the_data_distribution_trigger,
                         "when a significant changes in the data distributions of features is observed")

add_links({trigger_pipeline_or_orchestrator_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({automated_integration_and_delivery_decision: trigger_pipeline_or_orchestrator_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** decision: how to provision environments (test, production) ***
#
automated_provisioning = CClass(practice, "Automated provisioning")
test_environment_provisioning = CClass(practice, "Test environment provisioning")
pre_production_environment_provisioning = CClass(practice, "Pre-production environment provisioning")
production_environment_provisioning = CClass(practice, "Production environment provisioning")
no_automated_provisioning = CClass(do_nothing_design_solution, "No automated provisioning")

ci_cd_pipeline_enables_automated_provisioning = \
    ci_cd_pipeline.add_links(automated_provisioning, role_name="to", stereotype_instances=enables)[0]

automated_provisioning_can_include_test_environment_provisioning = \
    automated_provisioning.add_links(test_environment_provisioning, role_name="to", stereotype_instances=can_include)[0]
automated_provisioning_can_include_pre_production_environment_provisioning = \
    automated_provisioning.add_links(pre_production_environment_provisioning, role_name="to",
                                     stereotype_instances=can_include)[0]
automated_provisioning_can_include_production_environment_provisioning = \
    automated_provisioning.add_links(production_environment_provisioning, role_name="to",
                                     stereotype_instances=can_include)[0]

provisioning_decision = CClass(decision, "How to automatically provision environments?")

add_decision_option_link(provisioning_decision, no_automated_provisioning)
add_decision_option_link(provisioning_decision, automated_provisioning,
                         "automatically provision environments")

add_links({provisioning_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({no_automated_provisioning: {process_and_work_automation: negative,
                                                 independence_of_provisioning: very_negative,
                                                 auto_scaling: very_negative,
                                                 scalability: negative
                                                 },
                     automated_provisioning: {process_and_work_automation: very_positive,
                                              independence_of_provisioning: very_positive,
                                              auto_scaling: very_positive,
                                              scalability: very_positive},
                     })

#
# ** decision: How to automatically extract, validate, and prepare the data used for model building ***
#
data_pipeline = CClass(practice, "Data pipeline")
etl_pipeline = CClass(practice, "ETL pipeline")
data_processing_component = CClass(practice, "Data processing component")
no_data_processing_automation = CClass(do_nothing_design_solution, "No data processing automation")

build_and_deployment_scripts_can_be_combined_with_data_pipeline = \
    build_and_deployment_scripts.add_links(data_pipeline, role_name="to", stereotype_instances=can_be_combined_with)[0]

data_pipeline_can_be_combined_with_ml_orchestrator = \
    data_pipeline.add_links(ml_orchestrator, role_name="to",
                            stereotype_instances=can_be_combined_with,
                            label="orchestration of data pipelines")[0]

data_processing_component_can_be_combined_with_ml_orchestrator = \
    data_processing_component.add_links(ml_orchestrator, role_name="to",
                                        stereotype_instances=can_be_combined_with,
                                        label="orchestration of data processing")[0]

data_processing_automation_decision = CClass(decision,
                                             "How to automatically process the data used for model building?")

add_decision_option_link(data_processing_automation_decision, no_data_processing_automation)
add_decision_option_link(data_processing_automation_decision, data_pipeline,
                         "process data in dedicated, flexible pipeline")
add_decision_option_link(data_processing_automation_decision, etl_pipeline,
                         "process data through extract, transform, and load steps")
add_decision_option_link(data_processing_automation_decision, data_processing_component,
                         "process data in dedicated component")

add_links({data_processing_automation_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({data_processing_automation_decision: feature_persistence_and_access_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to store resulting features")

add_force_relations({no_data_processing_automation: {automated_data_collection: very_negative,
                                                     process_and_work_automation: negative,
                                                     iterative_development: negative,
                                                     reproducibility: very_negative,
                                                     tool_development_effort: positive,
                                                     },
                     data_pipeline: {automated_data_collection: very_positive,
                                     process_and_work_automation: very_positive,
                                     iterative_development: positive,
                                     reproducibility: positive,
                                     tool_development_effort: negative
                                     },
                     etl_pipeline: {automated_data_collection: positive,
                                    process_and_work_automation: positive,
                                    iterative_development: neutral,
                                    reproducibility: positive,
                                    tool_development_effort: negative
                                    },
                     data_processing_component: {automated_data_collection: positive,
                                                 process_and_work_automation: positive,
                                                 iterative_development: negative,
                                                 reproducibility: neutral,
                                                 tool_development_effort: negative,
                                                 },
                     })

add_links({data_processing_automation_decision: data_ingestion_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="ingest data for data processing and later model building")

#
# ** Decision: Should data be processed in batches or in realtime? ***
#
batch_based_data_processing = CClass(practice, "Batch-based data processing")
real_time_data_processing = CClass(practice, "Real-time, stream-based data processing")

batch_or_realtime_data_processing_decision = CClass(decision, "Should data be processed in batches or in realtime?")
add_decision_option_link(batch_or_realtime_data_processing_decision, batch_based_data_processing)
add_decision_option_link(batch_or_realtime_data_processing_decision, real_time_data_processing)

add_links({batch_or_realtime_data_processing_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({batch_based_data_processing: {reliability: very_positive,
                                                   data_throughput: very_positive,
                                                   low_latency: negative,
                                                   near_real_time_support: very_negative,
                                                   online_support: negative,
                                                   offline_support: positive,
                                                   },
                     real_time_data_processing: {reliability: positive,
                                                 data_throughput: positive,
                                                 low_latency: very_positive,
                                                 near_real_time_support: very_positive,
                                                 online_support: very_positive,
                                                 offline_support: neutral,
                                                 }})

add_links({data_pipeline: batch_or_realtime_data_processing_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)
add_links({data_processing_component: batch_or_realtime_data_processing_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)
add_links({etl_pipeline: batch_or_realtime_data_processing_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: Which data processing tasks can be performed by data processing pipeline or component? ***
#
data_extraction = CClass(practice, "Data extraction")
data_preparation = CClass(practice, "Data preparation")
data_validation = CClass(practice, "Data validation")
data_transformation = CClass(practice, "Data transformation")
data_selection = CClass(practice, "Data selection")
feature_engineering = CClass(practice, "Feature engineering")
data_processing_hyperparameter_tuning = CClass(practice, "Data processing hyperparameter tuning")

data_processing_tasks_decision = CClass(decision, "Which data processing tasks can be performed by a " +
                                        "data processing pipeline or component?")

add_decision_option_link(data_processing_tasks_decision, data_extraction)
add_decision_option_link(data_processing_tasks_decision, data_transformation)
add_decision_option_link(data_processing_tasks_decision, data_preparation)
add_decision_option_link(data_processing_tasks_decision, data_validation)
add_decision_option_link(data_processing_tasks_decision, data_selection)
add_decision_option_link(data_processing_tasks_decision, feature_engineering)
add_decision_option_link(data_processing_tasks_decision, data_processing_hyperparameter_tuning)

add_links({data_processing_tasks_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({data_pipeline: data_processing_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the data pipeline")

add_links({data_processing_component: data_processing_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the data processing component")

#
# ** Decision: data processing parallelization
#
data_pipeline_per_data_set = CClass(practice, "Data pipeline per data set")
data_partitioning_in_pipeline = CClass(practice, "Data partitioning in data pipeline")
hybrid_data_processing_parallelization = CClass(practice, "Hybrid data processing parallelization")
no_data_processing_parallelization = CClass(do_nothing_design_solution, "No parallel data processing")

data_processing_parallelization_decision = CClass(decision, "How to parallelize data processing?")

add_decision_option_link(data_processing_parallelization_decision, no_data_processing_parallelization)
add_decision_option_link(data_processing_parallelization_decision, data_pipeline_per_data_set,
                         "dedicated pipeline for each dataset")
add_decision_option_link(data_processing_parallelization_decision, data_partitioning_in_pipeline,
                         "within a pipeline, the data is partitioned to benefit from parallel processing")
add_decision_option_link(data_processing_parallelization_decision, hybrid_data_processing_parallelization,
                         "combining one or more of the other parallelization options")

hybrid_data_processing_parallelization_can_use_data_pipeline_per_data_set = \
    hybrid_data_processing_parallelization.add_links(data_pipeline_per_data_set, role_name="to",
                                                     stereotype_instances=can_use)[0]

hybrid_data_processing_parallelization_can_use_data_partitioning_in_pipeline = \
    hybrid_data_processing_parallelization.add_links(data_partitioning_in_pipeline, role_name="to",
                                                     stereotype_instances=can_use)[0]

add_force_relations({no_data_processing_parallelization: {data_throughput: very_negative,
                                                          speed_performance: very_negative,
                                                          low_latency: negative,
                                                          independent_processing: very_negative,
                                                          },
                     data_pipeline_per_data_set: {data_throughput: positive,
                                                  speed_performance: positive,
                                                  low_latency: positive,
                                                  independent_processing: very_positive},
                     data_partitioning_in_pipeline: {data_throughput: positive,
                                                     speed_performance: positive,
                                                     low_latency: positive,
                                                     independent_processing: neutral},
                     hybrid_data_processing_parallelization: {data_throughput: very_positive,
                                                              speed_performance: very_positive,
                                                              low_latency: very_positive,
                                                              independent_processing: very_positive},
                     })

add_links({data_processing_parallelization_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({data_processing_automation_decision: data_processing_parallelization_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: how to automate model building ***
#
model_building_pipeline = CClass(practice, "Model building pipeline")
model_builder_component = CClass(practice, "Model builder component")
model_building_in_development_tool = CClass(practice, "Model building in development tool")

model_building_pipeline_can_be_combined_with_data_pipeline = \
    model_building_pipeline.add_links(data_pipeline, role_name="to", stereotype_instances=can_be_combined_with)[0]

model_building_pipeline_can_use_ml_metadata_store = \
    model_building_pipeline.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                      label="store all model building, model versioning, " +
                                            "and auditability metrics metadata")[0]

model_builder_component_can_use_ml_metadata_store = \
    model_builder_component.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                      label="store all model building, model versioning, " +
                                            "and validation metrics metadata")[0]

model_building_pipeline_can_use_model_registry = \
    model_building_pipeline.add_links(model_registry, role_name="to", stereotype_instances=can_use,
                                      label="store all models and model candidates")[0]

ci_cd_pipeline_can_include_model_building_pipeline = \
    ci_cd_pipeline.add_links(model_building_pipeline, role_name="to", stereotype_instances=can_include)[0]

ci_cd_pipeline_can_include_data_pipeline = \
    ci_cd_pipeline.add_links(data_pipeline, role_name="to", stereotype_instances=can_include)[0]

build_and_deployment_scripts_can_be_combined_with_model_building_pipeline = \
    build_and_deployment_scripts.add_links(model_building_pipeline, role_name="to",
                                           stereotype_instances=can_be_combined_with)[0]

model_building_pipeline_can_be_combined_with_ml_orchestrator = \
    model_building_pipeline.add_links(ml_orchestrator, role_name="to",
                                      stereotype_instances=can_be_combined_with,
                                      label="orchestration of pipelines and experiments")[0]

model_builder_component_can_be_combined_with_ml_orchestrator = \
    model_builder_component.add_links(ml_orchestrator, role_name="to",
                                      stereotype_instances=can_be_combined_with,
                                      label="orchestration of model building and experiments")[0]

model_building_decision = CClass(decision, "How to perform model building in an ML project?")
add_decision_option_link(model_building_decision, model_building_in_development_tool,
                         "model building performed in development sandbox")
add_decision_option_link(model_building_decision, model_building_pipeline,
                         "model training, evaluation, and validation in automated pipeline")
add_decision_option_link(model_building_decision, model_builder_component,
                         "model training, evaluation, and validation in dedicated component")

add_links({model_building_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({model_building_in_development_tool: {interchangeability_of_ml_algorithms: negative,
                                                           interactive_prototyping: positive,
                                                           low_latency: very_negative,
                                                           loose_coupling: negative,
                                                           scalability: very_negative,
                                                           iterative_development: positive,
                                                           production_ready_development: negative,
                                                           experimental_operational_symmetry: negative,
                                                           reproducibility: negative,
                                                           development_agility: negative,
                                                           flexibility: negative,
                                                           tool_development_effort: positive},
                     model_building_pipeline: {interchangeability_of_ml_algorithms: positive,
                                               interactive_prototyping: negative,
                                               low_latency: positive,
                                               loose_coupling: very_positive,
                                               scalability: positive,
                                               iterative_development: neutral,
                                               production_ready_development: positive,
                                               experimental_operational_symmetry: positive,
                                               reproducibility: positive,
                                               development_agility: positive,
                                               flexibility: positive,
                                               tool_development_effort: negative},
                     model_builder_component: {interchangeability_of_ml_algorithms: positive,
                                               interactive_prototyping: negative,
                                               low_latency: positive,
                                               loose_coupling: neutral,
                                               scalability: neutral,
                                               iterative_development: neutral,
                                               production_ready_development: neutral,
                                               experimental_operational_symmetry: positive,
                                               reproducibility: neutral,
                                               development_agility: neutral,
                                               flexibility: negative,
                                               tool_development_effort: negative},
                     })

add_links({model_building_decision: data_processing_automation_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to automate data processing tasks used by or part of model building")

add_links({model_building_decision: feature_persistence_and_access_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to access features for model training")

add_links({model_building_decision: trigger_pipeline_or_orchestrator_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# ** Decision: When and how to train the model?
#
batch_based_learning = CClass(practice, "Batch-based learning")
incremental_learning = CClass(practice, "Incremental learning")
hybrid_learning = CClass(practice, "Hybrid batch-based and incremental learning")

# s35
ensemble_learning = CClass(practice, "Ensemble learning")

hybrid_learning_can_use_batch_based_learning = \
    hybrid_learning.add_links(batch_based_learning, role_name="to", stereotype_instances=can_use,
                              label="precompute some factors")[0]
hybrid_learning_can_use_incremental_learning = \
    hybrid_learning.add_links(incremental_learning, role_name="to", stereotype_instances=can_use,
                              label="compute some factors incrementally")[0]

when_and_how_to_train_the_model_decision = CClass(decision, "When and how to train the model?")

add_decision_option_link(when_and_how_to_train_the_model_decision, batch_based_learning,
                         "train the model in batches, possibly offline")
add_decision_option_link(when_and_how_to_train_the_model_decision, incremental_learning,
                         "train the model incrementally, possibly online")
add_decision_option_link(when_and_how_to_train_the_model_decision, hybrid_learning,
                         "train some factors of the model in batches, possibly offline, and " +
                         "incrementally train other factors, possibly online")
add_decision_option_link(when_and_how_to_train_the_model_decision, ensemble_learning,
                         "combine multiple base models to produce a single optimal model")

add_links({when_and_how_to_train_the_model_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_force_relations({batch_based_learning: {react_to_unforeseen_changes_in_the_data: very_negative,
                                            memory_requirements: very_negative,
                                            handling_massive_amounts_of_data: very_positive,
                                            online_support: negative,
                                            offline_support: positive,
                                            speed_performance: very_negative,
                                            near_real_time_support: negative,
                                            scalability: positive,
                                            variety_of_ml_algorithm_choices: very_positive
                                            },
                     incremental_learning: {react_to_unforeseen_changes_in_the_data: very_positive,
                                            memory_requirements: very_positive,
                                            handling_massive_amounts_of_data: very_negative,
                                            online_support: positive,
                                            offline_support: negative,
                                            speed_performance: positive,
                                            near_real_time_support: positive,
                                            scalability: negative,
                                            variety_of_ml_algorithm_choices: negative, },
                     hybrid_learning: {react_to_unforeseen_changes_in_the_data: positive,
                                       memory_requirements: positive,
                                       handling_massive_amounts_of_data: positive,
                                       online_support: positive,
                                       offline_support: positive,
                                       speed_performance: neutral,
                                       near_real_time_support: neutral,
                                       scalability: positive,
                                       variety_of_ml_algorithm_choices: very_negative
                                       },
                     })

#
# ** Decision: Model building tasks
#
model_training = CClass(practice, "Model training")
data_splitting = CClass(practice, "Data splitting")
data_checkpoints = CClass(practice, "Data checkpoints")
model_validation = CClass(practice, "Model validation")
model_selection = CClass(practice, "Model selection")
model_packaging = CClass(practice, "Model packaging")
development_tool_facade = CClass(practice, "Development tool facade")
development_tool_export = CClass(practice, "Development tool export")
train_multiple_model_versions = CClass(practice,
                                       "Train multiple model versions with different parameters and/or algorithms")
model_hyperparameter_tuning = CClass(practice, "Model hyperparameter tuning")

development_tool_facade_can_be_combined_with_model_building_in_development_tool = \
    development_tool_facade.add_links(model_building_in_development_tool, role_name="to",
                                       stereotype_instances=enables)[0]

development_tool_export_can_be_combined_with_model_building_in_development_tool = \
    development_tool_export.add_links(model_building_in_development_tool, role_name="to",
                                       stereotype_instances=enables)[0]

add_force_relations({model_validation: {
    auditability: very_positive,
    maintainability: very_positive,
    explainability: very_positive,
    align_models_with_business_needs: very_positive,
    align_models_with_regulatory_requirements: very_positive,
    demonstrability: very_positive
}
})

model_building_tasks_decision = CClass(decision, "Which tasks can be performed by a " +
                                       "model building pipeline or component?")

add_decision_option_link(model_building_tasks_decision, model_training)
add_decision_option_link(model_building_tasks_decision, data_splitting)
add_decision_option_link(model_building_tasks_decision, data_checkpoints)
add_decision_option_link(model_building_tasks_decision, model_validation)
add_decision_option_link(model_building_tasks_decision, model_selection)
add_decision_option_link(model_building_tasks_decision, train_multiple_model_versions)
add_decision_option_link(model_building_tasks_decision, model_packaging)
add_decision_option_link(model_building_tasks_decision, model_hyperparameter_tuning)
add_decision_option_link(model_building_tasks_decision, development_tool_facade,
                         "use API to integrate development tool")
add_decision_option_link(model_building_tasks_decision, development_tool_export,
                         "use exported file to work in development tool")

add_links({model_building_tasks_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({model_building_pipeline: model_building_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the model building pipeline")

add_links({model_builder_component: model_building_tasks_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="tasks to be performed by the model builder component")

add_links({model_training: when_and_how_to_train_the_model_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

model_selection_uses_train_multiple_model_versions = \
    model_selection.add_links(train_multiple_model_versions, role_name="to",
                              stereotype_instances=uses)[0]

#
# ** Decision: model building parallelization
#
model_building_pipeline_per_model = CClass(practice, "Model building pipeline per model")
training_data_partitioning = CClass(practice, "Training data partitioning")
model_partitioning = CClass(practice, "Model partitioning")
hybrid_model_building_parallelization = CClass(practice, "Hybrid model building parallelization")
no_model_building_parallelization = CClass(do_nothing_design_solution, "No parallel model building")

model_building_parallelization_decision = CClass(decision, "How to parallelize model building?")

add_decision_option_link(model_building_parallelization_decision, no_model_building_parallelization)
add_decision_option_link(model_building_parallelization_decision, model_building_pipeline_per_model,
                         "all models run concurrently")
add_decision_option_link(model_building_parallelization_decision, training_data_partitioning,
                         "the data is partitioned and each partition has a replica of the model")
add_decision_option_link(model_building_parallelization_decision, model_partitioning,
                         "the model is partitioned and each partition updates a portion of parameters")
add_decision_option_link(model_building_parallelization_decision, hybrid_model_building_parallelization,
                         "combining one or more of the other parallelization options")

add_links({model_building_parallelization_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

# s35
training_data_partitioning_can_be_combined_with_model_hyperparameter_tuning = \
    training_data_partitioning.add_links(model_hyperparameter_tuning, role_name="to",
                                         stereotype_instances=can_be_combined_with)[0]
training_data_partitioning_can_be_combined_with_data_processing_hyperparameter_tuning = \
    training_data_partitioning.add_links(data_processing_hyperparameter_tuning, role_name="to",
                                         stereotype_instances=can_be_combined_with)[0]

hybrid_model_building_parallelization_can_use_model_building_pipeline_per_model = \
    hybrid_model_building_parallelization.add_links(model_building_pipeline_per_model, role_name="to",
                                                    stereotype_instances=can_use)[0]

hybrid_model_building_parallelization_can_use_training_data_partitioning = \
    hybrid_model_building_parallelization.add_links(training_data_partitioning, role_name="to",
                                                    stereotype_instances=can_use)[0]

hybrid_model_building_parallelization_can_use_training_model_partitioning = \
    hybrid_model_building_parallelization.add_links(model_partitioning, role_name="to",
                                                    stereotype_instances=can_use)[0]

add_force_relations({no_model_building_parallelization: {speed_performance: very_negative,
                                                         low_latency: negative,
                                                         },
                     model_building_pipeline_per_model: {speed_performance: positive,
                                                         low_latency: positive,
                                                         },
                     training_data_partitioning: {speed_performance: positive,
                                                  low_latency: positive,
                                                  handling_massive_amounts_of_data: positive,
                                                  near_real_time_support: positive
                                                  },
                     model_partitioning: {speed_performance: positive,
                                          low_latency: positive,
                                          },
                     hybrid_model_building_parallelization: {speed_performance: very_positive,
                                                             low_latency: very_positive,
                                                             },
                     })

add_links({model_building_decision: model_building_parallelization_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# Decision: monitoring and tracking in production
#
model_performance_monitoring = CClass(practice, "Model performance monitoring")
speed_performance_monitoring = CClass(practice, "Speed performance monitoring")
monitoring_or_logging_data_and_predictions = CClass(practice, "Monitoring or logging data and predictions")
monitoring_domain_specific_metrics_per_model = CClass(practice, "Monitoring domain-specific metrics per model")
model_accuracy_monitoring = CClass(practice, "Model accuracy monitoring")
no_in_production_monitoring = CClass(do_nothing_design_solution, "No in-production monitoring")
model_in_production_tracking = CClass(practice, "Model in-production tracking")
production_data_store = CClass(practice, "Production data store")

model_performance_monitoring_can_use_speed_performance_monitoring = \
    model_performance_monitoring.add_links(speed_performance_monitoring, role_name="to",
                                           stereotype_instances=can_use)[0]

model_performance_monitoring_can_use_model_accuracy_monitoring = \
    model_performance_monitoring.add_links(model_accuracy_monitoring, role_name="to",
                                           stereotype_instances=can_use)[0]

model_performance_monitoring_can_use_domain_specific_monitoring = \
    model_performance_monitoring.add_links(monitoring_domain_specific_metrics_per_model, role_name="to",
                                           stereotype_instances=can_use)[0]

model_performance_monitoring_requires_ml_metadata_store = \
    model_performance_monitoring.add_links(ml_metadata_store, role_name="to", stereotype_instances=requires,
                                           label="access model and prediction metadata")[0]

on_model_performance_degradation_trigger_requires_model_performance_monitoring = \
    on_model_performance_degradation_trigger.add_links(model_performance_monitoring, role_name="to",
                                                       stereotype_instances=requires)[0]

model_performance_monitoring_can_use_production_data_store = \
    model_performance_monitoring.add_links(production_data_store, role_name="to",
                                           stereotype_instances=can_use, label="store monitored production data")[0]

model_in_production_tracking_can_use_production_data_store = \
    model_in_production_tracking.add_links(production_data_store, role_name="to",
                                           stereotype_instances=can_use, label="store tracked production data")[0]

model_performance_monitoring_can_be_combined_with_model_in_production_tracking = \
    model_performance_monitoring.add_links(model_in_production_tracking, role_name="to",
                                           stereotype_instances=can_be_combined_with)[0]

monitoring_and_tracking_in_production_decision = CClass(decision, "How to monitor and track the model in production?")

add_decision_option_link(monitoring_and_tracking_in_production_decision, no_in_production_monitoring)
add_decision_option_link(monitoring_and_tracking_in_production_decision, model_performance_monitoring,
                         "observe specific model performance parameters")
add_decision_option_link(monitoring_and_tracking_in_production_decision, monitoring_or_logging_data_and_predictions,
                         "log and monitor system data and predictions")
add_decision_option_link(monitoring_and_tracking_in_production_decision, model_in_production_tracking,
                         "track model usage, inputs, outputs, interpretability, and fairness")

add_force_relations({no_in_production_monitoring: {speed_performance: negative,
                                                   model_accuracy: negative,
                                                   continuous_model_improvement: very_negative,
                                                   observability: very_negative,
                                                   avoid_training_serving_skew: negative,
                                                   identify_potential_bias_or_overfit: very_negative,
                                                   robustness: very_negative,
                                                   debugging_support: very_negative,
                                                   model_successful_operation: negative,
                                                   model_stability: negative,
                                                   model_performance: negative,
                                                   align_models_with_business_needs: negative,
                                                   align_models_with_regulatory_requirements: negative,
                                                   maintainability: negative,
                                                   reduce_cost: negative,
                                                   reliability: negative,
                                                   valuable_models: negative,
                                                   incident_detection: negative,
                                                   incident_resolution: negative,
                                                   fresh_models: negative,
                                                   risk_mitigation: negative,
                                                   recognise_data_drift: negative,
                                                   recognise_model_degradation: negative,
                                                   address_model_degradation: negative
                                                   },
                     model_performance_monitoring: {speed_performance: positive,
                                                    model_accuracy: positive,
                                                    continuous_model_improvement: very_positive,
                                                    observability: very_positive,
                                                    avoid_training_serving_skew: neutral,
                                                    identify_potential_bias_or_overfit: positive,
                                                    robustness: neutral,
                                                    debugging_support: neutral,
                                                    model_successful_operation: positive,
                                                    model_stability: positive,
                                                    model_performance: positive,
                                                    align_models_with_business_needs: positive,
                                                    align_models_with_regulatory_requirements: positive,
                                                    maintainability: positive,
                                                    reduce_cost: positive,
                                                    reliability: positive,
                                                    valuable_models: positive,
                                                    incident_detection: positive,
                                                    incident_resolution: positive,
                                                    fresh_models: positive,
                                                    risk_mitigation: positive,
                                                    recognise_data_drift: positive,
                                                    recognise_model_degradation: positive,
                                                    address_model_degradation: positive,
                                                    },
                     monitoring_or_logging_data_and_predictions: {speed_performance: neutral,
                                                                  model_accuracy: neutral,
                                                                  continuous_model_improvement: positive,
                                                                  observability: positive,
                                                                  avoid_training_serving_skew: neutral,
                                                                  identify_potential_bias_or_overfit: neutral,
                                                                  robustness: very_positive,
                                                                  debugging_support: very_positive,
                                                                  model_successful_operation: positive,
                                                                  model_stability: positive,
                                                                  model_performance: positive,
                                                                  align_models_with_business_needs: positive,
                                                                  align_models_with_regulatory_requirements: positive,
                                                                  maintainability: positive,
                                                                  reduce_cost: positive,
                                                                  reliability: positive,
                                                                  valuable_models: positive,
                                                                  incident_detection: positive,
                                                                  incident_resolution: positive,
                                                                  fresh_models: positive,
                                                                  risk_mitigation: positive
                                                                  },
                     model_in_production_tracking: {speed_performance: neutral,
                                                    model_accuracy: positive,
                                                    continuous_model_improvement: very_positive,
                                                    observability: very_positive,
                                                    avoid_training_serving_skew: very_positive,
                                                    identify_potential_bias_or_overfit: very_positive,
                                                    robustness: positive,
                                                    debugging_support: positive,
                                                    model_successful_operation: positive,
                                                    model_stability: positive,
                                                    model_performance: positive,
                                                    align_models_with_business_needs: positive,
                                                    align_models_with_regulatory_requirements: positive,
                                                    maintainability: positive,
                                                    reduce_cost: positive,
                                                    reliability: positive,
                                                    valuable_models: positive,
                                                    incident_detection: positive,
                                                    incident_resolution: positive,
                                                    fresh_models: positive,
                                                    risk_mitigation: positive
                                                    }
                     })

add_links({monitoring_and_tracking_in_production_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

#
# using AutoML or no decisions
#
auto_ml = CClass(practice, "AutoML")
no_auto_ml = CClass(practice, "No AutoML")

auto_ml_can_use_model_training = \
    auto_ml.add_links(model_training, role_name="to", stereotype_instances=can_use,
                      label="automatically train models")[0]
auto_ml_can_use_model_selection = \
    auto_ml.add_links(model_selection, role_name="to", stereotype_instances=can_use,
                      label="automatically select models")[0]
auto_ml_can_use_model_hyperparameter_tuning = \
    auto_ml.add_links(model_hyperparameter_tuning, role_name="to", stereotype_instances=can_use,
                      label="automatically tune hyperparameters of selected models")[0]
auto_ml_requires_model_validation = \
    auto_ml.add_links(model_validation, role_name="to", stereotype_instances=requires,
                      label="use model validation to score models")[0]
auto_ml_can_use_data_processing_hyperparameter_tuning = \
    auto_ml.add_links(data_processing_hyperparameter_tuning, role_name="to", stereotype_instances=can_use,
                      label="automatically tune hyperparameters of data processing")[0]
auto_ml_can_use_data_preparation = \
    auto_ml.add_links(data_preparation, role_name="to", stereotype_instances=can_use,
                      label="automatically preprocess data")[0]
auto_ml_can_use_feature_engineering = \
    auto_ml.add_links(feature_engineering, role_name="to", stereotype_instances=can_use,
                      label="auto-generate new features and select meaningful ones")[0]

auto_ml_decision = CClass(decision, "Should AutoML be applied and, if so, where?")
add_decision_option_link(auto_ml_decision, no_auto_ml)
add_decision_option_link(auto_ml_decision, auto_ml)

add_force_relations({auto_ml: {reproducibility: positive,
                               explainability: positive,
                               development_velocity: positive,
                               development_agility: positive,
                               # positive when auto_ml is applied, but unclear extra effort
                               # for production-ready development and building the auto-ml solution
                               tool_development_effort: neutral,
                               process_and_work_automation: positive,
                               production_ready_development: negative,
                               },
                     no_auto_ml: {reproducibility: neutral,
                                  explainability: neutral,
                                  development_velocity: neutral,
                                  development_agility: neutral,
                                  tool_development_effort: neutral,
                                  process_and_work_automation: neutral,
                                  production_ready_development: positive},
                     })

add_links({model_building_tasks_decision: auto_ml_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether AutoML should be applied")

add_links({data_processing_tasks_decision: auto_ml_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether AutoML should be applied")

#
# Decision: monitoring - model adjustment
#

manual_model_adjustment = CClass(practice, "Manual model adjustment")
automated_model_adjustment = CClass(practice, "Automated model adjustment")
switch_to_other_model_version = CClass(practice, "Switch to other model version")
retrain_model = CClass(practice, "Retrain model")
collect_new_data = CClass(practice, "Collect new data")

automated_model_adjustment_can_use_switch_to_other_model_version = \
    automated_model_adjustment.add_links(switch_to_other_model_version, role_name="to",
                                         stereotype_instances=can_use)[0]

automated_model_adjustment_can_use_retrain_model = \
    automated_model_adjustment.add_links(retrain_model, role_name="to", stereotype_instances=can_use)[0]

automated_model_adjustment_can_use_collect_new_data = \
    automated_model_adjustment.add_links(collect_new_data, role_name="to", stereotype_instances=can_use)[0]

switch_to_other_model_version_requires_n_in_production = \
    switch_to_other_model_version.add_links(n_versions_in_production, role_name="to",
                                            stereotype_instances=requires)[0]

retrain_model_can_use_train_multiple_model_versions = \
    retrain_model.add_links(train_multiple_model_versions, role_name="to",
                            stereotype_instances=can_use)[0]

monitoring_model_adjustment_decision = CClass(decision, "How to adjust models based on monitoring results?")
add_decision_option_link(monitoring_model_adjustment_decision, manual_model_adjustment,
                         "humans adjust and redeploy the models")
add_decision_option_link(monitoring_model_adjustment_decision, automated_model_adjustment,
                         "automated measures are triggered to adjust the models")

add_links({monitoring_model_adjustment_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({monitoring_and_tracking_in_production_decision: monitoring_model_adjustment_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

add_links({production_data_store: store_data_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# Decision: how to approach deployment of machine learning models
#
semi_automated_model_deployment = CClass(practice,
                                           "Build, test, and deploy models to be served and other system components in a semi-automated fashion by deploying pre-prepared pipelines")

ci_cd_pipeline_model_deployment = CClass(practice,
                                         "Build, test, and deploy machine learning models " +
                                         "based on CI/CD pipeline automation")

no_automated_deployment = CClass(practice, "No automated deployment")

ci_cd_pipeline_model_deployment_can_include_model_building_pipeline = \
    ci_cd_pipeline_model_deployment.add_links(model_building_pipeline, role_name="to", stereotype_instances=can_include)[
        0]
ci_cd_pipeline_model_deployment_can_include_data_pipelines = \
    ci_cd_pipeline_model_deployment.add_links(data_pipeline, role_name="to", stereotype_instances=can_include)[0]

semi_automated_model_deployment_can_include_model_building_pipeline = \
    semi_automated_model_deployment.add_links(model_building_pipeline, role_name="to", stereotype_instances=can_include)[
        0]
semi_automated_model_deployment_can_include_data_pipelines = \
    semi_automated_model_deployment.add_links(data_pipeline, role_name="to", stereotype_instances=can_include)[0]

deployment_how_to_approach_decision = CClass(decision, "How to approach deployment of machine learning models?")
add_decision_option_link(deployment_how_to_approach_decision, no_automated_deployment)
add_decision_option_link(deployment_how_to_approach_decision, semi_automated_model_deployment)
add_decision_option_link(deployment_how_to_approach_decision, ci_cd_pipeline_model_deployment)
add_links({ci_cd_pipeline_model_deployment: automated_integration_and_delivery_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="consider a holistic integration and delivery approach in a machine learning context")

add_force_relations({no_automated_deployment: {process_and_work_automation: very_negative,
                                               iterative_development: negative,
                                               observability: very_negative,
                                               },
                     semi_automated_model_deployment: {process_and_work_automation: positive,
                                                         iterative_development: positive,
                                                         observability: very_positive,
                                                         },
                     ci_cd_pipeline_model_deployment: {process_and_work_automation: very_positive,
                                                     iterative_development: very_positive,
                                                     observability: very_positive
                                                     },
                     })

add_links({deployment_how_to_approach_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({deployment_how_to_approach_decision: automated_integration_and_delivery_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to perform integration and delivery automation")

#
# Decision: what to to test decision
#
unit_testing_system_components = CClass(practice, "Unit testing system components")
system_integration_testing = CClass(practice, "System integration testing")
unit_testing_feature_engineering_logic = CClass(practice, "Unit testing feature engineering logic")
unit_testing_different_methods_implemented_in_a_model = CClass(practice,
                                                               "Unit testing different methods implemented in a model")
testing_model_training_converges = CClass(practice, "Testing model training converges")
testing_trained_model_quality = CClass(practice, "Testing trained model quality")
unit_testing_pipeline_components = CClass(practice, "Unit testing pipeline components")
integration_testing_pipeline_components = CClass(practice, "Integration testing pipeline components")
prediction_service_api_testing = CClass(practice, "Prediction service API testing")
prediction_service_performance_testing = CClass(practice, "Prediction service performance testing")
no_testing = CClass(practice, "No testing")

what_to_test_decision = CClass(decision, "What to test in a machine learning pipeline?")
add_decision_option_link(what_to_test_decision, no_testing)
add_decision_option_link(what_to_test_decision, unit_testing_system_components)
add_decision_option_link(what_to_test_decision, system_integration_testing)
add_decision_option_link(what_to_test_decision, a_b_testing)
add_decision_option_link(what_to_test_decision, canary_testing)
add_decision_option_link(what_to_test_decision, hypothesis_testing)
add_decision_option_link(what_to_test_decision, unit_testing_feature_engineering_logic)
add_decision_option_link(what_to_test_decision, unit_testing_different_methods_implemented_in_a_model)
add_decision_option_link(what_to_test_decision, testing_model_training_converges)
add_decision_option_link(what_to_test_decision, testing_trained_model_quality)
add_decision_option_link(what_to_test_decision, unit_testing_pipeline_components)
add_decision_option_link(what_to_test_decision, integration_testing_pipeline_components)
add_decision_option_link(what_to_test_decision, prediction_service_api_testing)
add_decision_option_link(what_to_test_decision, prediction_service_performance_testing)

add_force_relations({unit_testing_feature_engineering_logic: {
    code_quality: very_positive,
    debugging_support: very_positive,
    maintainability: very_positive,
    explainability: positive
},
    unit_testing_pipeline_components: {
        code_quality: very_positive,
        debugging_support: very_positive,
        maintainability: very_positive,
        explainability: positive
    },
    unit_testing_system_components: {
        code_quality: very_positive,
        debugging_support: very_positive,
        maintainability: very_positive,
        explainability: positive
    },
    unit_testing_different_methods_implemented_in_a_model: {
        code_quality: very_positive,
        debugging_support: very_positive,
        maintainability: very_positive,
        explainability: positive
    },
    integration_testing_pipeline_components: {
        model_consistency: positive,
        reproducibility: very_positive,
        modifiability: positive,
        data_usability: positive
    },
    system_integration_testing: {
        model_consistency: very_positive,
        reproducibility: very_positive,
        modifiability: very_positive,
        data_usability: very_positive
    }
})

add_links({what_to_test_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({testing: what_to_test_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

#
# Decision: when to use MLOps
# initially based on s20, extended with s21, s22 and s25
#
mlops = CClass(practice, "MLOps")
no_mlops = CClass(practice, "No MLOps")

continuous_training = CClass(practice, "Continuous training")
continuous_training.add_links(model_training, role_name="to", stereotype_instances=can_include)[0]
continuous_training.add_links(retrain_model, role_name="to", stereotype_instances=can_include)[0]
continuous_training.add_links(ci_cd_pipeline, role_name="from", stereotype_instances=can_include)[0]

continuous_validation = CClass(practice, "Continuous validation")
continuous_validation.add_links(model_validation, role_name="to", stereotype_instances=can_include)[0]
continuous_validation.add_links(data_validation, role_name="to", stereotype_instances=can_include)[0]
continuous_validation.add_links(ci_cd_pipeline, role_name="from", stereotype_instances=can_include)[0]

# s21
data_sourcing = CClass(practice, "Data sourcing")
data_sourcing.add_links(data_pipeline, role_name="from", stereotype_instances=can_use)[0]
data_sourcing.add_links(data_processing_component, role_name="from", stereotype_instances=can_use)[0]

data_analysis = CClass(practice, "Data analysis")
data_analysis.add_links(data_sourcing, role_name="to", stereotype_instances=requires)[0]
data_analysis.add_links(data_sourcing, role_name="from", stereotype_instances=enables)[0]

data_labelling = CClass(practice, "Data labeling")
data_labelling.add_links(data_analysis, role_name="to", stereotype_instances=requires)[0]
data_labelling.add_links(data_analysis, role_name="from", stereotype_instances=enables)[0]
data_labelling.add_links(automatic_data_labeller, role_name="to", stereotype_instances=can_use)[0]
data_labelling.add_links(manual_data_labelling, role_name="to", stereotype_instances=can_use)[0]

experiment_tracking = CClass(practice, "Experiment tracking")
model_architecture = CClass(practice, "Model architecture")
model_development = CClass(practice, "Model development")
model_evaluation = CClass(practice, "Model evaluation")

model_versioning = CClass(practice, "Model versioning")
model_versioning.add_links(version_control_repository, role_name="to", stereotype_instances=can_use)[0]
model_versioning.add_links(n_versions_in_production, role_name="to", stereotype_instances=enables)[0]
model_versioning.add_links(rollback_to_previous_model_version, role_name="to", stereotype_instances=enables)[0]
model_versioning.add_links(train_multiple_model_versions, role_name="to", stereotype_instances=enables)[0]
model_versioning.add_links(switch_to_other_model_version, role_name="to", stereotype_instances=enables)[0]

# Can't relate practices to forces?
# model_versioning.add_links(shadow_mode_testing, role_name="to", stereotype_instances=enables)[0]
# model_versioning.add_links(facilitates_testing_in_production, role_name="to", stereotype_instances=enables)[0]
# model_versioning.add_links(canary_testing, role_name="to", stereotype_instances=enables)[0]
# model_versioning.add_links(a_b_testing, role_name="to", stereotype_instances=enables)[0]

hardware_scaling = CClass(practice, "Hardware scaling")
hardware_scaling.add_links(production_environment_provisioning, role_name="to", stereotype_instances=has_variant)[0]

training_at_scale = CClass(practice, "Training at scale")
training_at_scale.add_links(model_training, role_name="to", stereotype_instances=requires)[0]
training_at_scale.add_links(train_multiple_model_versions, role_name="to", stereotype_instances=can_include)[0]
training_at_scale.add_links(retrain_model, role_name="to", stereotype_instances=can_include)[0]
training_at_scale.add_links(continuous_training, role_name="to", stereotype_instances=can_include)[0]

model_benchmarking = CClass(practice, "Model benchmarking")
model_benchmarking.add_links(model_evaluation, role_name="from", stereotype_instances=has_variant)[0]

model_management = CClass(practice, "Model management")

# Can't relate practices to forces?
# model_management.add_links(model_in_production_tracking, role_name="to", stereotype_instances=can_use)[0]
model_management.add_links(model_versioning, role_name="to", stereotype_instances=can_use)[0]

# Can't relate practices to forces?
# model_management.add_links(model_consistency, role_name="to", stereotype_instances=leads_to)[0]
# model_management.add_links(align_models_with_business_needs, role_name="to", stereotype_instances=enables)[0]
# model_management.add_links(align_models_with_regulatory_requirements, role_name="to", stereotype_instances=enables)[0]
# model_management.add_links(reproducibility, role_name="to", stereotype_instances=leads_to)[0]
# model_management.add_links(reusability, role_name="to", stereotype_instances=leads_to)[0]
# model_management.add_links(time_to_market_reduction, role_name="to", stereotype_instances=leads_to)[0]
# model_management.add_links(infrastructure_costs, role_name="to", stereotype_instances=influences)[0]

automated_model_development = CClass(practice, "Automated model development")
automated_model_development.add_links(model_development, role_name="to", stereotype_instances=requires)[0]
# Can't relate practices to forces?
# automated_model_development.add_links(time_to_market_reduction, role_name="to", stereotype_instances=leads_to)[0]
# automated_model_development.add_links(infrastructure_costs, role_name="to", stereotype_instances=influences)[0]

configuration_versioning = CClass(practice, "Configuration versioning")

automated_regression_tests = CClass(practice, "Automated regression tests")
automated_regression_tests.add_links(testing, role_name="from", stereotype_instances=extension)[0]

static_code_analysis = CClass(practice, "Static code analysis")
training_in_production = CClass(practice, "Training in production")

# s22
experimentation = CClass(practice, "Automated regression tests")
prediction_service = CClass(practice, "Prediction service")

skew_testing = CClass(practice, "Skew tests")
skew_testing.add_links(testing, role_name="from", stereotype_instances=extension)[0]

stress_testing = CClass(practice, "Stress testing")
stress_testing.add_links(testing, role_name="from", stereotype_instances=extension)[0]

system_monitoring = CClass(practice, "System monitoring")

application_pipeline = CClass(practice, "Application pipeline")

# s20
mlops_enables_deployment = \
    mlops.add_links(deployment, role_name="to", stereotype_instances=enables,
                    label="deployment of models")[0]

# s25
model_service_profiling = CClass(practice, "Model service profiling")

mlops_can_use_continuous_training = \
    mlops.add_links(continuous_training, role_name="to", stereotype_instances=can_use,
                    label="automatically and continually train models")[0]

mlops_can_use_continuous_validation = \
    mlops.add_links(continuous_validation, role_name="to", stereotype_instances=can_use,
                    label="automatically and continually validate models")[0]

mlops_can_use_model_registry = \
    mlops.add_links(model_registry, role_name="to", stereotype_instances=can_use,
                    label="keep track of models and their metadata in a registry")[0]

mlops_can_use_model_validation = \
    mlops.add_links(model_validation, role_name="to", stereotype_instances=can_use,
                    label="automatically validate models")[0]

mlops_can_use_model_training = \
    mlops.add_links(model_training, role_name="to", stereotype_instances=can_use,
                    label="automatically train models")[0]

mlops_can_use_ci_cd_pipeline = \
    mlops.add_links(ci_cd_pipeline, role_name="to", stereotype_instances=can_use,
                    label="use a CI/CD pipeline")[0]

# s21
mlops_can_use_data_sourcing = \
    mlops.add_links(data_sourcing, role_name="to", stereotype_instances=can_use,
                    label="source data")[0]

mlops_can_use_data_analysis = \
    mlops.add_links(data_analysis, role_name="to", stereotype_instances=can_use,
                    label="analyse data")[0]

mlops_can_use_data_labelling = \
    mlops.add_links(data_labelling, role_name="to", stereotype_instances=can_use,
                    label="label data")[0]

mlops_enables_experiment_tracking = \
    mlops.add_links(experiment_tracking, role_name="to", stereotype_instances=enables,
                    label="track experiments")[0]

mlops_can_use_model_architecture = \
    mlops.add_links(model_architecture, role_name="to", stereotype_instances=can_use,
                    label="architect models")[0]

mlops_can_use_model_development = \
    mlops.add_links(model_development, role_name="to", stereotype_instances=can_use,
                    label="develop models")[0]

mlops_can_use_model_evaluation = \
    mlops.add_links(model_evaluation, role_name="to", stereotype_instances=can_use,
                    label="evaluate models")[0]

mlops_can_use_model_versioning = \
    mlops.add_links(model_versioning, role_name="to", stereotype_instances=can_use,
                    label="version models")[0]

mlops_can_use_hardware_scaling = \
    mlops.add_links(hardware_scaling, role_name="to", stereotype_instances=can_use,
                    label="automatically scale hardware")[0]

mlops_enables_training_at_scale = \
    mlops.add_links(training_at_scale, role_name="to", stereotype_instances=enables,
                    label="train at scale")[0]

mlops_can_use_model_management = \
    mlops.add_links(model_management, role_name="to", stereotype_instances=can_use,
                    label="model management")[0]

mlops_can_use_automated_model_development = \
    mlops.add_links(automated_model_development, role_name="to", stereotype_instances=can_use,
                    label="automated model development")[0]

mlops_can_use_configuration_versioning = \
    mlops.add_links(configuration_versioning, role_name="to", stereotype_instances=can_use,
                    label="configuration versioning")[0]

mlops_can_use_automated_regression_tests = \
    mlops.add_links(automated_regression_tests, role_name="to", stereotype_instances=can_use,
                    label="automated regression tests")[0]

mlops_can_use_static_code_analysis = \
    mlops.add_links(static_code_analysis, role_name="to", stereotype_instances=can_use,
                    label="static code analysis")[0]

mlops_enables_training_in_production = \
    mlops.add_links(training_in_production, role_name="to", stereotype_instances=enables,
                    label="training in production")[0]

mlops_can_use_on_demand_trigger = \
    mlops.add_links(on_demand_trigger, role_name="to", stereotype_instances=can_use,
                    label="on-demand trigger")[0]
mlops_can_use_on_schedule_trigger = \
    mlops.add_links(on_schedule_trigger, role_name="to", stereotype_instances=can_use,
                    label="on schedule trigger")[0]

mlops_can_use_on_availability_of_new_training_data_trigger = \
    mlops.add_links(on_availability_of_new_training_data_trigger, role_name="to", stereotype_instances=can_use,
                    label="on availability of new training data trigger")[0]

mlops_can_use_on_model_performance_degradation_trigger = \
    mlops.add_links(on_model_performance_degradation_trigger, role_name="to", stereotype_instances=can_use,
                    label="on model performance degradation trigger")[0]

mlops_can_use_on_changes_in_the_data_distribution_trigger = \
    mlops.add_links(on_changes_in_the_data_distribution_trigger, role_name="to", stereotype_instances=can_use,
                    label="on changes in the data distribution trigger")[0]

# s22
mlops_enables_experimentation = \
    mlops.add_links(experimentation, role_name="to", stereotype_instances=enables,
                    label="experimentation")[0]

mlops_enables_prediction_service = \
    mlops.add_links(prediction_service, role_name="to", stereotype_instances=enables,
                    label="prediction service")[0]

mlops_can_use_skew_testing = \
    mlops.add_links(skew_testing, role_name="to", stereotype_instances=can_use,
                    label="skew testing")[0]

mlops_can_use_system_monitoring = \
    mlops.add_links(system_monitoring, role_name="to", stereotype_instances=can_use,
                    label="system_monitoring")[0]

mlops_can_use_application_pipeline = \
    mlops.add_links(application_pipeline, role_name="to", stereotype_instances=can_use,
                    label="application_pipeline")[0]

mlops_can_use_stress_testing = \
    mlops.add_links(stress_testing, role_name="to", stereotype_instances=can_use,
                    label="stress testing")[0]

mlops_can_use_canary_testing = \
    mlops.add_links(canary_testing, role_name="to", stereotype_instances=can_use,
                    label="canary testing")[0]

mlops_can_use_data_versioning_in_code_repository = \
    mlops.add_links(data_versioning_in_code_repository, role_name="to", stereotype_instances=can_use,
                    label="data versioning in code repository")[0]

mlops_can_use_data_version_repository = \
    mlops.add_links(data_version_repository, role_name="to", stereotype_instances=can_use,
                    label="data version repository")[0]

mlops_enables_model_benchmarking = \
    mlops.add_links(model_benchmarking, role_name="to", stereotype_instances=enables,
                    label="model benchmarking")[0]

mlops_can_use_a_b_testing = \
    mlops.add_links(a_b_testing, role_name="to", stereotype_instances=can_use,
                    label="A/B testing")[0]

mlops_can_use_prediction_service_api_testing = \
    mlops.add_links(prediction_service_api_testing, role_name="to", stereotype_instances=can_use,
                    label="prediction service api testing")[0]

mlops_can_use_prediction_service_performance_testing = \
    mlops.add_links(prediction_service_performance_testing, role_name="to", stereotype_instances=can_use,
                    label="prediction service performance testing")[0]

# s25
mlops_can_use_packaging = \
    mlops.add_links(packaging, role_name="to", stereotype_instances=can_use,
                    label="packaging")[0]

mlops_can_use_model_service_profiling = \
    mlops.add_links(model_service_profiling, role_name="to", stereotype_instances=can_use,
                    label="model_service_profiling")[0]

mlops_can_use_model_in_batch_prediction_system = \
    mlops.add_links(model_in_batch_prediction_system, role_name="to", stereotype_instances=can_use,
                    label="model in batch prediction system")[0]

mlops_can_use_ci_cd_pipeline_model_deployment = \
    mlops.add_links(ci_cd_pipeline_model_deployment, role_name="to", stereotype_instances=can_use,
                    label="deployment of ML and DS pipelines")[0]

mlops_can_use_data_ingestor = \
    mlops.add_links(data_ingestor, role_name="to", stereotype_instances=can_use,
                    label="data ingestor")[0]

mlops_decision = CClass(decision, "Should MLOps be applied and, if so, when?")
add_decision_option_link(mlops_decision, no_mlops)
add_decision_option_link(mlops_decision, mlops)

# existing practices without decisions
mlops_can_use_retrain_model = \
    mlops.add_links(retrain_model, role_name="to", stereotype_instances=can_use,
                    label="retrain model")[0]

# decisions for existing practices
# model_registry, shared_data_store
add_links({store_data_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# model_validation, data_splitting, model_hyperparameter_tuning
add_links({model_building_tasks_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# model_training
add_links({when_and_how_to_train_the_model_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# ci_cd_pipeline, ml_metadata_store, ml_orchestrator
add_links({automated_integration_and_delivery_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# deployment, version_control_repository, a_b_testing, canary_testing
add_links({delivery_tasks_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# model_performance_monitoring
add_links({monitoring_and_tracking_in_production_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# data_validation, data_preparation, data_transformation, feature_engineering
# data_processing_hyperparameter_tuning
add_links({data_processing_tasks_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# semi_automated_model_deployment
add_links({deployment_how_to_approach_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# unit_testing_system_components, system_integration_testing
add_links({what_to_test_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# automated_provisioning
add_links({provisioning_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# data_processing_component
add_links({data_processing_automation_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# data_pipeline_per_data_set, model_building_pipeline_per_model
add_links({data_processing_parallelization_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# automated_model_adjustment
add_links({model_building_parallelization_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# on_demand_trigger, on_commit_trigger, on_schedule_trigger
# on_availability_of_new_training_data_trigger
# on_model_performance_degradation_trigger
# on_changes_in_the_data_distribution_trigger
add_links({trigger_pipeline_or_orchestrator_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# feature_store
add_links({feature_persistence_and_access_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

# rollback_to_previous_model_version
add_links({deploying_model_versions_in_production_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_force_relations({mlops: {division_of_labour: positive,
                             process_and_work_automation: positive,
                             time_to_market_reduction: positive,
                             improve_user_experience: positive,
                             improve_customer_satisfaction: positive,
                             improve_prediction_quality: positive,
                             recognise_model_degradation: positive,
                             address_model_degradation: positive,
                             reproducibility: positive,
                             recognise_data_drift: positive,
                             dependable_releases: positive,
                             improve_model_management: positive,
                             change_tracking: positive,
                             shadow_mode_testing: positive,
                             # s21
                             simplified_deployment: positive,
                             align_models_with_business_needs: positive,
                             align_models_with_regulatory_requirements: positive,
                             explainability: positive,
                             reusability: positive,
                             infrastructure_costs: positive,
                             development_agility: positive,
                             model_consistency: positive,
                             unlock_new_sources_of_revenue: positive,
                             save_time: positive,
                             reduce_cost: positive,
                             management_agility: positive,
                             continuous_improvement: positive,
                             experimental_operational_symmetry: positive,
                             modularity: positive,
                             rapid_experimentation: positive,
                             # s22
                             reduce_technical_debt: positive,
                             iterative_development: positive,
                             # s25
                             development_velocity: positive,
                             deployment_velocity: positive,
                             reusability: positive,
                             auditability: positive,
                             explainability: positive,
                             },
                     no_mlops: {division_of_labour: neutral,
                                process_and_work_automation: neutral,
                                time_to_market_reduction: neutral,
                                improve_user_experience: neutral,
                                improve_customer_satisfaction: neutral,
                                improve_prediction_quality: neutral,
                                recognise_model_degradation: neutral,
                                address_model_degradation: neutral,
                                reproducibility: neutral,
                                recognise_data_drift: neutral,
                                dependable_releases: neutral,
                                improve_model_management: neutral,
                                change_tracking: neutral,
                                shadow_mode_testing: neutral,
                                # s21
                                simplified_deployment: neutral,
                                align_models_with_business_needs: neutral,
                                align_models_with_regulatory_requirements: neutral,
                                explainability: neutral,
                                reusability: neutral,
                                infrastructure_costs: neutral,
                                development_agility: neutral,
                                model_consistency: neutral,
                                unlock_new_sources_of_revenue: neutral,
                                save_time: neutral,
                                reduce_cost: neutral,
                                management_agility: neutral,
                                continuous_improvement: neutral,
                                experimental_operational_symmetry: neutral,
                                modularity: neutral,
                                rapid_experimentation: neutral,
                                # s22
                                reduce_technical_debt: neutral,
                                iterative_development: neutral,
                                # s25
                                development_velocity: neutral,
                                deployment_velocity: neutral,
                                reusability: neutral,
                                auditability: neutral,
                                explainability: neutral
                                },
                     })

add_links({data_ingestion_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_links({deploying_how_to_version_data_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_links({batch_or_realtime_data_processing_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_links({model_building_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_links({model_performance_monitoring: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

add_links({monitoring_model_adjustment_decision: mlops_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="whether to apply MLOps")

#
# Decision: what development environment to use
# initially based on s26, extended with s27, and s28
#
# s26
development_environment_decision = CClass(practice, "What development environment should be used?")

use_computational_notebooks = CClass(practice, "Use computational notebooks")
use_ide = CClass(practice, "Use an IDE")

add_decision_option_link(development_environment_decision, use_computational_notebooks, "use computational notebooks")
add_decision_option_link(development_environment_decision, use_ide, "use an IDE")

use_computational_notebooks_rules_out_version_control_repository = \
    use_computational_notebooks.add_links(version_control_repository, role_name="to", stereotype_instances=rules_out,
                                          label="version control repository")[0]
use_computational_notebooks_rules_out_static_code_analysis = \
    use_computational_notebooks.add_links(static_code_analysis, role_name="to", stereotype_instances=rules_out,
                                          label="static code analysis")[0]

use_ide_can_use_version_control_repository = \
    use_ide.add_links(version_control_repository, role_name="to", stereotype_instances=can_use,
                      label="version control repository")[0]

add_force_relations({use_computational_notebooks:
    {  # s26
        reusability: positive_and_negative,
        flexibility: very_positive,
        rapid_experimentation: positive,
        interactive_prototyping: very_positive,
        demonstrability: very_positive,
        ease_of_hosting: positive,
        maintainability: very_negative,
        development_agility: positive_and_negative,
        change_tracking: very_negative,
        use_ide_features: very_negative,
        development_velocity: positive_and_negative,
        reduce_technical_debt: very_negative,
        code_quality: very_negative,
        reproducibility: positive_and_negative,
        # s27
        efficiency: positive,
        # s28
        testability: negative,
        # s29
        production_ready_development: negative,
        time_to_market_reduction: negative,
        debugging_support: negative,
        explainability: positive,
        tool_development_effort: positive,
        simplified_deployment: negative,
        model_performance: negative,
    },
    use_ide: {
        # s26
        development_agility: very_positive,
        reproducibility: very_positive,
        interactive_prototyping: positive,
        development_velocity: very_positive,
        use_ide_features: very_positive,
        code_quality: very_positive,
        change_tracking: very_positive,
        reduce_technical_debt: positive,
        maintainability: positive,
        # s27
        # IDEs not discussed in this source
        # s28

        # IDEs not discussed in this source
        # s29
    },
})

# s34
add_force_relations({stress_testing: {align_models_with_business_needs: positive,
                                      speed_performance: very_positive,
                                      model_accuracy: neutral,
                                      maintainability: neutral,
                                      reduce_cost: positive
                                      },
                     prediction_service_performance_testing: {align_models_with_business_needs: positive,
                                                              speed_performance: very_positive,
                                                              model_accuracy: neutral,
                                                              maintainability: neutral,
                                                              reduce_cost: positive
                                                              },
                     prediction_service_api_testing: {align_models_with_business_needs: positive,
                                                      speed_performance: positive,
                                                      model_accuracy: positive,
                                                      maintainability: positive,
                                                      reduce_cost: positive
                                                      }
                     })

# s35
add_force_relations({ensemble_learning: {error_reduction: positive
                                         }
                     })

training_data_partitioning_can_be_combined_with_ensemble_learning = \
    training_data_partitioning.add_links(ensemble_learning, role_name="to",
                                         stereotype_instances=can_be_combined_with)[0]

model_training_can_use_ensemble_learning = \
    model_training.add_links(ensemble_learning, role_name="to",
                             stereotype_instances=can_use)[0]

# decision views
#

# Please note: we do not show the contexts of the decisions, as in this ADD model, they are almost always
# a model and application in combination

forces_class_objects = [f.class_object for f in force.all_classes]
model_and_app_class_objects = [ml_model.class_object, ml_application.class_object]

integrating_ml_model_decision_view = CBundle("integrating_ml_model_decision", elements=[
    o.class_object for o in [integrating_ml_model_decision, model_as_a_service, containerization,
                             model_in_monolith, model_as_module_in_modular_monolith,
                             model_in_batch_prediction_system, event_driven_model_as_a_service,
                             serverless_model_as_a_service,
                             model_serving_decision, no_integration_of_model, mlops_decision]])

model_serving_decision_view = CBundle("model_serving_decision", elements=[
    o.class_object for o in [model_serving_decision, microservice_based_model_serving, model_as_a_service,
                             service_based_model_serving, local_service_api_based_model_serving,
                             model_as_module_in_modular_monolith, data_batch_based_model_serving,
                             predictions_data_store, feature_persistence_and_access_decision,
                             store_data_decision, hybrid_online_offline_model_serving, precompute_serving,
                             result_availability_information_decision, data_ingestion_decision]])

model_serving_result_availability_decision_view = CBundle("model_serving_result_availability_decision", elements=[
    o.class_object for o in [result_availability_information_decision, block_on_result, push_notification,
                             polling]])

store_data_decision_view = CBundle("store_data_decision", elements=[
    o.class_object for o in [store_data_decision, no_data_store, event_store, event_sourcing, real_time_database,
                             data_lake, shared_database, shared_data_store, data_warehouse, data_service,
                             data_access_interface_decision, database_per_service, in_memory_data_store,
                             mlops_decision]])

feature_persistence_and_access_decision_view = CBundle("feature_persistence_and_access_decision", elements=[
    o.class_object for o in [feature_persistence_and_access_decision, store_features_without_specific_support,
                             feature_store, event_sourcing, batch_feature_api, real_time_feature_api,
                             data_access_interface_decision, store_data_decision, mlops_decision]])

data_ingestion_decision_view = CBundle("data_ingestion_decision", elements=[
    o.class_object for o in [data_ingestion_decision, data_ingestion_in_batches, data_ingestion_by_request,
                             streamed_data_ingestion, manual_data_ingestion, raw_data_store,
                             data_ingestor, ground_truth_collector, data_ingestion_queue,
                             manual_data_labelling, automatic_data_labeller,
                             store_data_decision, mlops_decision]])

automated_integration_and_delivery_in_ml_context_decision_view = \
    CBundle("automated_integration_and_delivery_in_ml_context_decision", elements=[
        o.class_object for o in [automated_integration_and_delivery_decision,
                                 ci_cd_pipeline, ml_orchestrator, no_integration_or_delivery_automation,
                                 build_and_deployment_scripts, trigger_pipeline_or_orchestrator_decision,
                                 deploying_how_to_version_data_decision, automated_provisioning,
                                 ml_metadata_store, model_registry, delivery_tasks_decision,
                                 deploying_model_versions_in_production_decision, store_data_decision,
                                 mlops_decision]])

pipeline_combinations_view = CBundle("pipeline_combinations", elements=[
    o.class_object for o in [ci_cd_pipeline, data_pipeline, build_and_deployment_scripts,
                             model_building_pipeline, ml_orchestrator]])

delivery_tasks_decision_view = CBundle("delivery_tasks_decision", elements=[
    o.class_object for o in [packaging, testing, building,
                             deployment, containerization, canary_testing,
                             a_b_testing, hypothesis_testing, delivery_tasks_decision,
                             what_to_test_decision, mlops_decision]])

deployment_how_to_approach_decision_view = \
    CBundle("deployment_how_to_approach_decision", elements=[
        o.class_object for o in [deployment_how_to_approach_decision, no_automated_deployment,
                                 semi_automated_model_deployment, ci_cd_pipeline_model_deployment,
                                 model_building_pipeline, data_pipeline,
                                 automated_integration_and_delivery_decision, mlops_decision]])

deploying_how_to_version_data_decision_view = CBundle("deploying_how_to_version_data_decision", elements=[
    o.class_object for o in [deploying_how_to_version_data_decision, no_data_versioning,
                             data_version_repository, data_versioning_in_code_repository,
                             version_control_repository, mlops_decision]])

deploying_model_versions_in_production_decision_view = CBundle("deploying_model_versions_in_production_decision",
                                                               elements=[
                                                                   o.class_object for o in [
                                                                       deploying_model_versions_in_production_decision,
                                                                       single_model_in_production,
                                                                       n_versions_in_production, ml_metadata_store,
                                                                       rollback_to_previous_model_version,
                                                                       mlops_decision,
                                                                       canary_testing,
                                                                       a_b_testing]])

trigger_pipeline_or_orchestrator_decision_view = \
    CBundle("trigger_pipeline_or_orchestrator_decision", elements=[
        o.class_object for o in [trigger_pipeline_or_orchestrator_decision, on_demand_trigger,
                                 on_commit_trigger, on_schedule_trigger, on_availability_of_new_training_data_trigger,
                                 on_model_performance_degradation_trigger, model_performance_monitoring,
                                 on_changes_in_the_data_distribution_trigger, mlops_decision]])

provisioning_decision_view = \
    CBundle("provisioning_decision", elements=[
        o.class_object for o in [provisioning_decision, automated_provisioning,
                                 test_environment_provisioning, pre_production_environment_provisioning,
                                 production_environment_provisioning, no_automated_provisioning, ci_cd_pipeline,
                                 mlops_decision]])

data_processing_automation_decision_view = \
    CBundle("data_processing_automation_decision", elements=[
        o.class_object for o in [data_processing_automation_decision, no_data_processing_automation,
                                 data_processing_component, batch_or_realtime_data_processing_decision,
                                 data_pipeline, etl_pipeline,
                                 data_processing_tasks_decision, ml_orchestrator,
                                 feature_persistence_and_access_decision, data_processing_parallelization_decision,
                                 data_ingestion_decision, mlops_decision]])

batch_or_realtime_data_processing_decision_view = \
    CBundle("batch_or_realtime_data_processing_decision", elements=[
        o.class_object for o in [batch_or_realtime_data_processing_decision, real_time_data_processing,
                                 batch_based_data_processing, mlops_decision]])

data_processing_tasks_decision_view = \
    CBundle("data_processing_tasks_decision", elements=[
        o.class_object for o in [data_processing_tasks_decision, data_extraction, data_transformation,
                                 data_preparation, data_validation, data_selection, feature_engineering,
                                 data_processing_hyperparameter_tuning, auto_ml_decision, mlops_decision]])

data_processing_parallelization_decision_view = \
    CBundle("data_processing_parallelization_decision", elements=[
        o.class_object for o in [data_processing_parallelization_decision, no_data_processing_parallelization,
                                 data_pipeline_per_data_set, hybrid_data_processing_parallelization,
                                 data_partitioning_in_pipeline, mlops_decision]])

data_access_interface_decision_view = \
    CBundle("data_access_interface_decision", elements=[
        o.class_object for o in [data_access_interface_decision, direct_data_provider_access, local_data_access_api,
                                 service_based_data_access_api, pub_sub_data_access_api,
                                 message_event_queue_based_data_access_api, data_service]])

model_building_decision_view = \
    CBundle("model_building_decision", elements=[
        o.class_object for o in [model_building_decision, model_building_tasks_decision,
                                 model_building_in_development_tool,
                                 model_building_pipeline, ml_orchestrator,
                                 model_registry, ml_metadata_store, store_data_decision, model_builder_component,
                                 model_building_parallelization_decision, feature_persistence_and_access_decision,
                                 trigger_pipeline_or_orchestrator_decision, data_processing_automation_decision,
                                 mlops_decision]])

model_building_tasks_decision_view = \
    CBundle("model_building_tasks_decision", elements=[
        o.class_object for o in [model_training, data_splitting,
                                 data_checkpoints, model_validation, model_selection,
                                 train_multiple_model_versions, model_building_tasks_decision, model_packaging,
                                 development_tool_facade, model_hyperparameter_tuning, development_tool_export,
                                 model_building_in_development_tool, when_and_how_to_train_the_model_decision,
                                 auto_ml_decision]])

when_and_how_to_train_the_model_decision_view = \
    CBundle("when_and_how_to_train_the_model_decision", elements=[
        o.class_object for o in [batch_based_learning, incremental_learning, hybrid_learning,
                                 when_and_how_to_train_the_model_decision]])

model_building_parallelization_decision_view = \
    CBundle("model_building_parallelization_decision", elements=[
        o.class_object for o in [model_building_parallelization_decision, no_model_building_parallelization,
                                 model_building_pipeline_per_model, hybrid_model_building_parallelization,
                                 training_data_partitioning, model_partitioning, mlops_decision]])

monitoring_in_production_decision_view = \
    CBundle("monitoring_in_production_decision", elements=[
        o.class_object for o in [monitoring_and_tracking_in_production_decision, no_in_production_monitoring,
                                 model_performance_monitoring, speed_performance_monitoring, model_accuracy_monitoring,
                                 monitoring_domain_specific_metrics_per_model,
                                 ml_metadata_store, production_data_store,
                                 model_in_production_tracking, monitoring_or_logging_data_and_predictions,
                                 monitoring_model_adjustment_decision, store_data_decision,
                                 mlops_decision]])

monitoring_model_adjustment_decision_view = \
    CBundle("monitoring_model_adjustment_decision", elements=[
        o.class_object for o in [monitoring_model_adjustment_decision, manual_model_adjustment,
                                 automated_model_adjustment, switch_to_other_model_version,
                                 n_versions_in_production, retrain_model, collect_new_data,
                                 train_multiple_model_versions, mlops_decision]])

what_to_test_decision_view = \
    CBundle("what_to_test_decision", elements=[
        o.class_object for o in [what_to_test_decision, no_testing, unit_testing_system_components,
                                 system_integration_testing, a_b_testing, hypothesis_testing,
                                 canary_testing,
                                 unit_testing_feature_engineering_logic,
                                 unit_testing_different_methods_implemented_in_a_model,
                                 testing_model_training_converges,
                                 testing_trained_model_quality, unit_testing_pipeline_components,
                                 integration_testing_pipeline_components, prediction_service_api_testing,
                                 prediction_service_performance_testing, mlops_decision]])

auto_ml_decision_view = \
    CBundle("auto_ml_decision", elements=[
        o.class_object for o in
        [auto_ml_decision, auto_ml, no_auto_ml, data_preparation, feature_engineering, model_training, model_selection,
         model_validation, data_processing_hyperparameter_tuning, model_hyperparameter_tuning]])

# originally based on S20
mlops_decision_view = \
    CBundle("mlops_decision", elements=[
        o.class_object for o in
        [mlops_decision,
         mlops,
         no_mlops,
         semi_automated_model_deployment,
         model_training,
         deployment,
         retrain_model,
         ci_cd_pipeline,
         data_validation,
         model_validation,
         continuous_training,
         continuous_validation,
         division_of_labour,
         process_and_work_automation,
         time_to_market_reduction,
         improve_user_experience,
         improve_customer_satisfaction,
         improve_prediction_quality,
         recognise_model_degradation,
         address_model_degradation,
         reproducibility,
         recognise_data_drift,
         dependable_releases,
         improve_model_management,
         change_tracking,
         shadow_mode_testing,
         model_registry,
         data_sourcing,
         data_analysis,
         data_labelling,
         experiment_tracking,
         model_architecture,
         model_development,
         model_evaluation,
         model_versioning,
         hardware_scaling,
         training_at_scale,
         model_benchmarking,
         model_management,
         automated_model_development,
         configuration_versioning,
         automated_regression_tests,
         static_code_analysis,
         training_in_production,
         experimentation,
         prediction_service,
         skew_testing,
         stress_testing,
         system_monitoring,
         application_pipeline,
         packaging,
         model_service_profiling,
         model_in_batch_prediction_system,
         ci_cd_pipeline_model_deployment,
         data_ingestor
         ]])

# originally based on S26
development_environment_decision_view = \
    CBundle("development_environment_decision", elements=[
        o.class_object for o in
        [experimentation,
         use_computational_notebooks,
         use_ide,
         version_control_repository,
         development_environment_decision
         ]])

everything_view = \
    CBundle("everything_view",
            elements=[o.class_object for o in [integrating_ml_model_decision,
                                               containerization,
                                               model_in_monolith,
                                               model_as_module_in_modular_monolith,
                                               model_in_batch_prediction_system,
                                               event_driven_model_as_a_service,
                                               serverless_model_as_a_service,
                                               no_integration_of_model,
                                               mlops_decision,
                                               model_serving_decision,
                                               microservice_based_model_serving,
                                               model_as_a_service,
                                               service_based_model_serving,
                                               local_service_api_based_model_serving,
                                               data_batch_based_model_serving,
                                               predictions_data_store,
                                               hybrid_online_offline_model_serving,
                                               precompute_serving,
                                               result_availability_information_decision,
                                               block_on_result,
                                               push_notification,
                                               polling,
                                               no_data_store,
                                               event_store,
                                               real_time_database,
                                               data_lake,
                                               shared_database,
                                               shared_data_store,
                                               data_warehouse,
                                               database_per_service,
                                               in_memory_data_store,
                                               store_features_without_specific_support,
                                               feature_store,
                                               event_sourcing,
                                               batch_feature_api,
                                               real_time_feature_api,
                                               data_ingestion_in_batches,
                                               data_ingestion_by_request,
                                               streamed_data_ingestion,
                                               manual_data_ingestion,
                                               raw_data_store,
                                               ground_truth_collector,
                                               data_ingestion_queue,
                                               manual_data_labelling,
                                               automatic_data_labeller,
                                               no_integration_or_delivery_automation,
                                               build_and_deployment_scripts,
                                               testing,
                                               building,
                                               delivery_tasks_decision,
                                               deployment_how_to_approach_decision,
                                               no_automated_deployment,
                                               automated_integration_and_delivery_decision,
                                               deploying_how_to_version_data_decision,
                                               no_data_versioning,
                                               data_version_repository,
                                               data_versioning_in_code_repository,
                                               deploying_model_versions_in_production_decision,
                                               single_model_in_production,
                                               rollback_to_previous_model_version,
                                               on_demand_trigger,
                                               on_commit_trigger,
                                               on_schedule_trigger,
                                               on_availability_of_new_training_data_trigger,
                                               on_model_performance_degradation_trigger,
                                               on_changes_in_the_data_distribution_trigger,
                                               provisioning_decision,
                                               automated_provisioning,
                                               test_environment_provisioning,
                                               pre_production_environment_provisioning,
                                               production_environment_provisioning,
                                               no_automated_provisioning,
                                               no_data_processing_automation,
                                               data_processing_component,
                                               data_pipeline,
                                               etl_pipeline,
                                               data_ingestion_decision,
                                               batch_or_realtime_data_processing_decision,
                                               real_time_data_processing,
                                               batch_based_data_processing,
                                               data_processing_tasks_decision,
                                               data_extraction,
                                               data_transformation,
                                               data_selection,
                                               data_processing_parallelization_decision,
                                               no_data_processing_parallelization,
                                               data_pipeline_per_data_set,
                                               hybrid_data_processing_parallelization,
                                               data_partitioning_in_pipeline,
                                               data_access_interface_decision,
                                               direct_data_provider_access,
                                               local_data_access_api,
                                               service_based_data_access_api,
                                               pub_sub_data_access_api,
                                               message_event_queue_based_data_access_api,
                                               data_service,
                                               model_building_decision,
                                               model_building_pipeline,
                                               ml_orchestrator,
                                               model_builder_component,
                                               feature_persistence_and_access_decision,
                                               trigger_pipeline_or_orchestrator_decision,
                                               data_processing_automation_decision,
                                               data_splitting,
                                               data_checkpoints,
                                               model_building_tasks_decision,
                                               model_packaging,
                                               development_tool_facade,
                                               development_tool_export,
                                               model_building_in_development_tool,
                                               batch_based_learning,
                                               incremental_learning,
                                               hybrid_learning,
                                               when_and_how_to_train_the_model_decision,
                                               model_building_parallelization_decision,
                                               no_model_building_parallelization,
                                               model_building_pipeline_per_model,
                                               hybrid_model_building_parallelization,
                                               training_data_partitioning,
                                               model_partitioning,
                                               monitoring_and_tracking_in_production_decision,
                                               no_in_production_monitoring,
                                               model_performance_monitoring,
                                               speed_performance_monitoring,
                                               model_accuracy_monitoring,
                                               monitoring_domain_specific_metrics_per_model,
                                               ml_metadata_store,
                                               production_data_store,
                                               model_in_production_tracking,
                                               monitoring_or_logging_data_and_predictions,
                                               store_data_decision,
                                               monitoring_model_adjustment_decision,
                                               manual_model_adjustment,
                                               automated_model_adjustment,
                                               switch_to_other_model_version,
                                               n_versions_in_production,
                                               collect_new_data,
                                               train_multiple_model_versions,
                                               what_to_test_decision,
                                               no_testing,
                                               unit_testing_system_components,
                                               system_integration_testing,
                                               a_b_testing,
                                               hypothesis_testing,
                                               canary_testing,
                                               unit_testing_feature_engineering_logic,
                                               unit_testing_different_methods_implemented_in_a_model,
                                               testing_model_training_converges,
                                               testing_trained_model_quality,
                                               unit_testing_pipeline_components,
                                               integration_testing_pipeline_components,
                                               prediction_service_api_testing,
                                               prediction_service_performance_testing,
                                               auto_ml_decision,
                                               auto_ml,
                                               no_auto_ml,
                                               data_preparation,
                                               feature_engineering,
                                               model_selection,
                                               data_processing_hyperparameter_tuning,
                                               model_hyperparameter_tuning,
                                               mlops,
                                               no_mlops,
                                               semi_automated_model_deployment,
                                               model_training,
                                               deployment,
                                               retrain_model,
                                               ci_cd_pipeline,
                                               data_validation,
                                               model_validation,
                                               continuous_training,
                                               continuous_validation,
                                               division_of_labour,
                                               process_and_work_automation,
                                               time_to_market_reduction,
                                               improve_user_experience,
                                               improve_customer_satisfaction,
                                               improve_prediction_quality,
                                               recognise_model_degradation,
                                               address_model_degradation,
                                               reproducibility,
                                               recognise_data_drift,
                                               dependable_releases,
                                               improve_model_management,
                                               change_tracking,
                                               shadow_mode_testing,
                                               model_registry,
                                               data_sourcing,
                                               data_analysis,
                                               data_labelling,
                                               experiment_tracking,
                                               model_architecture,
                                               model_development,
                                               model_evaluation,
                                               model_versioning,
                                               hardware_scaling,
                                               training_at_scale,
                                               model_benchmarking,
                                               model_management,
                                               automated_model_development,
                                               configuration_versioning,
                                               automated_regression_tests,
                                               static_code_analysis,
                                               training_in_production,
                                               prediction_service,
                                               skew_testing,
                                               stress_testing,
                                               system_monitoring,
                                               application_pipeline,
                                               packaging,
                                               model_service_profiling,
                                               ci_cd_pipeline_model_deployment,
                                               data_ingestor,
                                               experimentation,
                                               use_computational_notebooks,
                                               use_ide,
                                               version_control_repository,
                                               development_environment_decision]])

deployment_decisions_and_practices_excluding_mlops_view = \
    CBundle("deployment_decisions_and_practices_excluding_mlops", elements=[
        o.class_object for o in [
            deployment_how_to_approach_decision,
            no_automated_deployment,
            semi_automated_model_deployment,
            ci_cd_pipeline_model_deployment,
            model_building_decision,
            model_building_pipeline,
            model_builder_component,
            automated_integration_and_delivery_decision,
            ci_cd_pipeline,
            no_integration_or_delivery_automation,
            build_and_deployment_scripts,
            automated_provisioning,
            model_registry,
            trigger_pipeline_or_orchestrator_decision,
            on_demand_trigger,
            on_commit_trigger,
            on_schedule_trigger,
            on_availability_of_new_training_data_trigger,
            on_model_performance_degradation_trigger,
            model_performance_monitoring,
            on_changes_in_the_data_distribution_trigger,
            deploying_model_versions_in_production_decision,
            single_model_in_production,
            n_versions_in_production,
            ml_metadata_store,
            rollback_to_previous_model_version,
            no_data_store,
            event_store,
            real_time_database,
            data_lake,
            shared_database,
            shared_data_store,
            data_warehouse,
            database_per_service,
            in_memory_data_store,
            packaging,
            testing,
            building,
            deployment,
            containerization,
            delivery_tasks_decision,
            what_to_test_decision,
            no_testing,
            unit_testing_system_components,
            system_integration_testing,
            a_b_testing,
            hypothesis_testing,
            canary_testing,
            unit_testing_feature_engineering_logic,
            unit_testing_different_methods_implemented_in_a_model,
            testing_model_training_converges,
            testing_trained_model_quality,
            unit_testing_pipeline_components,
            integration_testing_pipeline_components,
            prediction_service_api_testing,
            prediction_service_performance_testing,
            data_ingestion_in_batches,
            data_ingestion_by_request,
            streamed_data_ingestion,
            manual_data_ingestion,
            raw_data_store,
            data_ingestor,
            ground_truth_collector,
            data_ingestion_queue,
            manual_data_labelling,
            automatic_data_labeller,
            deploying_how_to_version_data_decision,
            no_data_versioning,
            data_version_repository,
            data_versioning_in_code_repository,
            version_control_repository,
            data_splitting,
            data_checkpoints,
            train_multiple_model_versions,
            model_building_tasks_decision,
            model_packaging,
            development_tool_facade,
            development_tool_export,
            model_building_in_development_tool,
            batch_based_learning,
            incremental_learning,
            hybrid_learning,
            when_and_how_to_train_the_model_decision,
            auto_ml_decision,
            auto_ml,
            no_auto_ml,
            data_preparation,
            feature_engineering,
            model_training,
            model_selection,
            model_validation,
            data_processing_hyperparameter_tuning,
            model_hyperparameter_tuning,
            direct_data_provider_access,
            local_data_access_api,
            service_based_data_access_api,
            pub_sub_data_access_api,
            message_event_queue_based_data_access_api,
            data_service,
            model_building_parallelization_decision,
            no_model_building_parallelization,
            model_building_pipeline_per_model,
            hybrid_model_building_parallelization,
            training_data_partitioning,
            model_partitioning,
            store_features_without_specific_support,
            feature_store,
            event_sourcing,
            batch_feature_api,
            real_time_feature_api,
            data_access_interface_decision,
            store_data_decision,
            data_processing_automation_decision,
            no_data_processing_automation,
            data_processing_component,
            batch_or_realtime_data_processing_decision,
            data_pipeline,
            etl_pipeline,
            data_processing_tasks_decision,
            ml_orchestrator,
            feature_persistence_and_access_decision,
            data_processing_parallelization_decision,
            data_ingestion_decision,
            mlops_decision,
        ]])

deployment_decisions_and_practices_including_mlops_view = \
    CBundle("deployment_decisions_and_practices_including_mlops", elements=[
        o.class_object for o in [
            deployment_how_to_approach_decision,
            no_automated_deployment,
            semi_automated_model_deployment,
            ci_cd_pipeline_model_deployment,
            model_building_decision,
            model_building_pipeline,
            model_builder_component,
            automated_integration_and_delivery_decision,
            ci_cd_pipeline,
            no_integration_or_delivery_automation,
            build_and_deployment_scripts,
            automated_provisioning,
            model_registry,
            trigger_pipeline_or_orchestrator_decision,
            on_demand_trigger,
            on_commit_trigger,
            on_schedule_trigger,
            on_availability_of_new_training_data_trigger,
            on_model_performance_degradation_trigger,
            model_performance_monitoring,
            on_changes_in_the_data_distribution_trigger,
            deploying_model_versions_in_production_decision,
            single_model_in_production,
            n_versions_in_production,
            ml_metadata_store,
            rollback_to_previous_model_version,
            no_data_store,
            event_store,
            real_time_database,
            data_lake,
            shared_database,
            shared_data_store,
            data_warehouse,
            database_per_service,
            in_memory_data_store,
            packaging,
            testing,
            building,
            deployment,
            containerization,
            delivery_tasks_decision,
            what_to_test_decision,
            no_testing,
            unit_testing_system_components,
            system_integration_testing,
            a_b_testing,
            hypothesis_testing,
            canary_testing,
            unit_testing_feature_engineering_logic,
            unit_testing_different_methods_implemented_in_a_model,
            testing_model_training_converges,
            testing_trained_model_quality,
            unit_testing_pipeline_components,
            integration_testing_pipeline_components,
            prediction_service_api_testing,
            prediction_service_performance_testing,
            data_ingestion_in_batches,
            data_ingestion_by_request,
            streamed_data_ingestion,
            manual_data_ingestion,
            raw_data_store,
            data_ingestor,
            ground_truth_collector,
            data_ingestion_queue,
            manual_data_labelling,
            automatic_data_labeller,
            deploying_how_to_version_data_decision,
            no_data_versioning,
            data_version_repository,
            data_versioning_in_code_repository,
            version_control_repository,
            data_splitting,
            data_checkpoints,
            train_multiple_model_versions,
            model_building_tasks_decision,
            model_packaging,
            development_tool_facade,
            development_tool_export,
            model_building_in_development_tool,
            batch_based_learning,
            incremental_learning,
            hybrid_learning,
            when_and_how_to_train_the_model_decision,
            auto_ml_decision,
            auto_ml,
            no_auto_ml,
            data_preparation,
            feature_engineering,
            model_training,
            model_selection,
            model_validation,
            data_processing_hyperparameter_tuning,
            model_hyperparameter_tuning,
            direct_data_provider_access,
            local_data_access_api,
            service_based_data_access_api,
            pub_sub_data_access_api,
            message_event_queue_based_data_access_api,
            data_service,
            model_building_parallelization_decision,
            no_model_building_parallelization,
            model_building_pipeline_per_model,
            hybrid_model_building_parallelization,
            training_data_partitioning,
            model_partitioning,
            store_features_without_specific_support,
            feature_store,
            event_sourcing,
            batch_feature_api,
            real_time_feature_api,
            data_access_interface_decision,
            store_data_decision,
            data_processing_automation_decision,
            no_data_processing_automation,
            data_processing_component,
            batch_or_realtime_data_processing_decision,
            data_pipeline,
            etl_pipeline,
            data_processing_tasks_decision,
            ml_orchestrator,
            feature_persistence_and_access_decision,
            data_processing_parallelization_decision,
            data_ingestion_decision,
            mlops_decision,
            mlops,
            no_mlops,
            retrain_model,
            data_validation,
            continuous_training,
            continuous_validation,
            division_of_labour,
            process_and_work_automation,
            time_to_market_reduction,
            improve_user_experience,
            improve_customer_satisfaction,
            improve_prediction_quality,
            recognise_model_degradation,
            address_model_degradation,
            reproducibility,
            recognise_data_drift,
            dependable_releases,
            improve_model_management,
            change_tracking,
            shadow_mode_testing,
            data_sourcing,
            data_analysis,
            data_labelling,
            experiment_tracking,
            model_architecture,
            model_development,
            model_evaluation,
            model_versioning,
            hardware_scaling,
            training_at_scale,
            model_benchmarking,
            model_management,
            automated_model_development,
            configuration_versioning,
            automated_regression_tests,
            static_code_analysis,
            training_in_production,
            experimentation,
            prediction_service,
            skew_testing,
            stress_testing,
            system_monitoring,
            application_pipeline,
            model_service_profiling,
            model_in_batch_prediction_system,
        ]])

deployment_decisions_only_view = \
    CBundle("automated_deployment_only_decisions", elements=[
        o.class_object for o in [
            deployment_how_to_approach_decision,
            model_building_decision,
            automated_integration_and_delivery_decision,
            trigger_pipeline_or_orchestrator_decision,
            deploying_model_versions_in_production_decision,
            store_data_decision,
            mlops_decision,
            delivery_tasks_decision,
            what_to_test_decision,
            data_ingestion_decision,
            deploying_how_to_version_data_decision,
            model_building_tasks_decision,
            when_and_how_to_train_the_model_decision,
            auto_ml_decision,
            data_access_interface_decision,
            model_building_parallelization_decision,
            feature_persistence_and_access_decision,
            data_processing_automation_decision,
        ]])

ml_adds_views = [
    integrating_ml_model_decision_view, {},
    model_serving_decision_view, {},
    model_serving_result_availability_decision_view, {},
    store_data_decision_view, {},
    feature_persistence_and_access_decision_view, {},
    data_ingestion_decision_view, {},
    automated_integration_and_delivery_in_ml_context_decision_view, {},
    pipeline_combinations_view, {},
    delivery_tasks_decision_view, {},
    deployment_how_to_approach_decision_view, {},
    deploying_how_to_version_data_decision_view, {},
    deploying_model_versions_in_production_decision_view, {},
    trigger_pipeline_or_orchestrator_decision_view, {},
    provisioning_decision_view, {},
    data_processing_automation_decision_view, {},
    batch_or_realtime_data_processing_decision_view, {},
    data_processing_tasks_decision_view, {},
    data_processing_parallelization_decision_view, {},
    data_access_interface_decision_view, {},
    model_building_decision_view, {},
    model_building_tasks_decision_view, {},
    model_building_parallelization_decision_view, {},
    monitoring_in_production_decision_view, {},
    monitoring_model_adjustment_decision_view, {},
    what_to_test_decision_view, {},
    when_and_how_to_train_the_model_decision_view, {},
    auto_ml_decision_view, {},
    mlops_decision_view, {},
    development_environment_decision_view, {},
    everything_view, {},
    deployment_decisions_only_view, {},
    deployment_decisions_and_practices_including_mlops_view, {},
    deployment_decisions_and_practices_excluding_mlops_view, {},
]
