# source
## ID
26
## url
https://betterprogramming.pub/pros-and-cons-for-jupyter-notebooks-as-your-editor-for-data-science-work-tip-pycharm-is-probably-40e88f7827cb
## archive url
https://tinyurl.com/ml-adds-u26
## title
The Pros and Cons of Using Jupyter Notebooks as Your Editor for Data Science Work TL;DR: PyCharm’s probably better
## source code
no
## example
no
## source type
Practitioner Audience Article
## author type
Practitioner

# coding

> Jupyter notebooks have three particularly strong benefits:
They’re great for showcasing your work. You can see both the code and the results. The notebooks at Kaggle is a particularly great example of this.
It’s easy to use other people’s work as a starting point. You can run cell by cell to better get an understanding of what the code does.
Very easy to host server side, which is useful for security purposes. A lot of data is sensitive and should be protected, and one of the steps toward that is no data is stored on local machines. A server-side Jupyter Notebook setup gives you that for free.

> When prototyping, the cell-based approach of Jupyter notebooks is great. But you quickly end up programming several steps — instead of looking at object-oriented programming.
 
**OPEN CODING TRACE:**

During coding, we shall generalise from e.g. "Jupyter notebooks" to "computational notebooks".
Another abstraction: "PyCharm" etc. shall be generalised as "IDEs".
Another abstraction: "pip", "conda" etc. shall be generalised as "dependency management tools".

The decision is ultimately whether to use computational notebooks or IDEs.

- **benefits/forces of notebooks:**
  - great for showcasing your work
    (-> no existing force)
  - easy to use other people’s work as a starting point
    (-> `reusability`, `flexibility`)  
  - Very easy to host server side
    (-> no existing force)
  - good for prototyping
    (`rapid_experimentation`, `interactive_prototyping`)

- **risks/drawbacks of notebooks:**
  - But you quickly end up programming several steps — instead of looking at object-oriented programming.
    (-> not sure what is meant by this)

- **practices:**
  - use computational notebooks
    (-> no existing practice)
  - good for prototyping
    (-> practice: `experimentation`)
    

**AXIAL CODING TRACE:**

existing forces:
`reusability`
`flexibility`
`rapid_experimentation`
`interactive_prototyping`

new forces:
`demonstrability`
`ease_of_hosting`

existing practice:
`experimentation`

added new decision option (practice):
`use_computational_notebooks`

added new decision:
`development_environment_decision`

added missing practices to source:
``` python
`experimentation`
`use_computational_notebooks`
```

updated force relations for the above forces that were missing for `use_computational_notebooks`:
`reusability`
`flexibility`
`rapid_experimentation`
`interactive_prototyping`


>Downsides of Jupyter notebooks
When we’re writing code in cells instead of functions/classes/objects, you quickly end up with duplicate code that does the same thing, which is very hard to maintain.
Don’t get the support from a powerful IDE.
Consequences of duplicate code:
It’s hard to actually collaborate on code with Jupyter — as we’re copying snippets from each other it’s very easy to get out of sync
Hard to maintain one version of the truth. Which one of these notebooks has the one true solution to the number of xyz?
There’s also a tricky problem related to plotting. How are you sharing plots outside of the data science team? At first, Jupyter Notebook is a great way of sharing plots — just share the notebook! But how do you ensure the data there’s fresh? Easy, just have them run the notebook.
But in large organizations, you might run into a lot of issues as you don’t want too many users having direct access to the underlying data (for GDPR issues or otherwise). In practice, in a workplace, we’ve noticed plots from Jupyter typically get shared by copy/pasting into PowerPoint. It’s highly ineffective to have your data scientists do copy/paste cycles whenever the data changes.

>An example
Let’s look at an example notebook from Kaggle.
Screenshot: Kaggle example

> This was very easy to get started on. I just copied some cells from the introductory examples and then explored on my own. But here we also see the downside — access-credentials management is now duplicated across all of the notebooks. What if they change? Then, every notebook needs to be changed as well.
> The dataset is interesting. But there’s no canonical one — so if you want to reuse it, you’re copying the SQL statement.
> With poor discipline, you can also move into weird versioning issues where you start accumulating multiple Jupyter notebooks that no one remembers.

 
**OPEN CODING TRACE:**

- **risks/drawbacks of notebooks:**
  - very hard to maintain
    (-> force: `maintainability`)
  - no IDE support
    (no existing force)
  - code duplication
    (-> force: `maintainability`)
  - hinders collaboration
    (-> force: `development_agility`)
  - Hard to maintain one version of the truth. Which one of these notebooks has the one true solution to the number of xyz?
    (force -> `change_tracking`)
  - How are you sharing plots outside of the data science team? But how do you ensure the data there’s fresh?
    we’ve noticed plots from Jupyter typically get shared by copy/pasting into PowerPoint.
    It’s highly ineffective to have your data scientists do copy/paste cycles whenever the data changes.
    (-> force: `development_agility`)
  - access-credentials management is now duplicated across all of the notebooks.
    What if they change? Then, every notebook needs to be changed as well.
    (-> force: `maintainability`)
    you can also move into weird versioning issues where you start accumulating multiple Jupyter notebooks that no one remembers.
    (force -> `change_tracking`)
  
**AXIAL CODING TRACE:**

existing forces:
`maintainability`
`development_agility`
`change_tracking`

new forces:
`use_ide_features`

updated force relations for the above forces that were missing for `use_computational_notebooks`:
`maintainability`
`development_agility`
`change_tracking`


>Benefits from an IDE
As alternatives to Jupyter Notebook, you could checkout Spyder and PyCharm.
Spyder has an interesting feature where it’s very good at reloading code while keeping the state of your current session. Thus, Spyder is great if you’re trying to do most of your transformations in pandas or some other tool on your local machine. It also comes with Anaconda, if that’s your Python distribution of choice.
PyCharm is really good at building proper Python code and getting replicable results. There’s a lot of functionality for greater productivity. Both Spyder and PyCharm enable cell-based processing with #%% comments in your code, so we can still prototype code in the same way as in Jupyter Notebook.

>Screenshot: PyCharm

>For one of our previous clients, we wanted to improve code quality but were not allowed to access data on any local machines. So we spent the effort to spin up VMs with PyCharm to get access to data in a secure manner. It paid off quickly — development speed and code quality increased by a factor of a lot. And code made it into production a lot faster as well.

**OPEN CODING TRACE:**

- **benefits/forces of IDEs:**
  - Spyder has an interesting feature where it’s very good at reloading code while keeping the state of your current session.
    (-> force: `development_agility`)
  -  PyCharm is really good at building proper Python code and getting replicable results.
    (-> force: `reproducibility`)
  - we can still prototype code in the same way as in Jupyter Notebook
    (-> force: `interactive_prototyping`)
  - development speed and code quality increased by a factor of a lot.
    (-> force: `development_velocity`, missing force for quality)
  
- **practices:**
  - As alternatives to Jupyter Notebook, you could checkout Spyder and PyCharm
    (-> no existing practice)
  - we can still prototype code in the same way as in Jupyter Notebook
    (-> practice: `experimentation`)


**AXIAL CODING TRACE:**

existing forces:
`development_agility`
`reproducibility`
`interactive_prototyping`
`development_velocity`
`use_ide_features`

new forces:
`code_quality`

existing practice:
`experimentation`

added new decision option (practice):
`use_ide`

added missing practices to source:
`use_ide`

updated force relations for the above forces that were missing for `use_ide`
`development_agility`
`reproducibility`
`interactive_prototyping`
`development_velocity`
`use_ide_features`
`code_quality`


> Getting machine learning into production
Something to think about is where computations are run. For code that’s easy to put into Docker, deploying that to any cloud solution is easy. For notebooks, there are also good options, though you’re more locked into specific solutions.
If you do want to look into Jupyter notebooks, it’s definitely worth looking into Amazon SageMaker and/or Kubeflow. These solutions enable easier deployment to production for code in notebooks.
We’ve had a lot of success with the following approach:
Use PyCharm (which has improved code quality by a far bit)
Every transformation of data needs to exist in exactly one place in our repository. (So any issues with that transformation needs to be fixed in that one place.)
Every transformation needs to be in production (so available as a table/file/output), so other data scientists can reuse it in their models. And if that transformation improves, all subsequent pipelines are automatically improved as well.

**OPEN CODING TRACE:**
Discusses specific technologies - these won't be coded, since we wish to abstract away such details.

The discussion on transformation is irrelevant with regard to development environments.


> Conclusion
Should you remove notebooks completely? Notebooks have a lot of pros. It depends on where you are and what your main concerns are
If all of the machine learning is already on the cloud and only needs some light scripting — notebooks are probably the easiest way there
Be careful with a heavy reliance on notebooks when the data-engineering team is short staffed or when the data scientist team is immature — as this is when you can build up an unmanageable amount of bad habits and technical debt in a short amount of time
If your problems are fairly unique and require a lot of self-developed code, Jupyter notebooks will grow in size and complexity in ways that’ll be hard to maintain
The larger the team, the more concerns we have about collaborative coding and the reuse of results between team members, and we should be moving away from Jupyter
If you have cross-functional teams, with both software engineers and data scientists, getting the most out of version control and object-oriented programming is easier. If you have a cross-functional team, you’ll probably get more benefits by moving away from Jupyter.

**OPEN CODING TRACE:**

- **benefits/forces of notebooks:**
  - If all of the machine learning is already on the cloud and only needs some light scripting — notebooks are probably the easiest way there
    (-> force: `development_velocity`)
  
- **risks/drawbacks of notebooks/IDEs:**
  - you can build up an unmanageable amount of bad habits and technical debt in a short amount of time
    (-> force: `reduce_technical_debt`, `code_quality`)
  - Jupyter notebooks will grow in size and complexity in ways that’ll be hard to maintain
    (-> force: `maintainability`)
  - The larger the team, the more concerns we have about collaborative coding and the reuse of results between team members, and we should be moving away from Jupyter  #
    (-> force: `development_agility`, `reproducibility`)
  - If you have cross-functional teams, with both software engineers and data scientists, getting the most out of version control and object-oriented programming is easier. If you have a cross-functional team, you’ll probably get more benefits by moving away from Jupyter.
    (-> force: `change_tracking` negative for notebooks, positive for IDEs)
    
- **practices:**
  - If you have cross-functional teams, with both software engineers and data scientists, getting the most out of version control and object-oriented programming is easier. If you have a cross-functional team, you’ll probably get more benefits by moving away from Jupyter.
    (-> practice: `version_control_repository` negative for notebooks, positive for IDEs)

**AXIAL CODING TRACE:**

existing forces:
`development_velocity`
`reduce_technical_debt`
`code_quality`
`maintainability`
`development_agility`
`reproducibility`
`change_tracking`

existing practice:
`version_control_repository`

added missing practices to source:
`version_control_repository`

updated force relations for the above forces that were missing for `use_computational_notebooks`:
`development_velocity`
`reduce_technical_debt`
`code_quality`
`maintainability`
`development_agility`
`reproducibility`
`change_tracking`


updated force relations for the above forces that were missing for `use_ide`:
`development_velocity`
`reduce_technical_debt`
`code_quality`
`maintainability`
`development_agility`
`reproducibility`
`change_tracking`

**Concluding Changes:**
**AXIAL CODING TRACE:**

source and references
added `s26`, `s26_codes` and links in `ml_adds_sources_codes.py`:
``` python
s26
s26_codes
add_links({s26: s26_codes}, role_name="contained_code")
```
updated `decision_view`
added decision option links to `development_environment_decision`
added relations between new practices and existing practices

**SELECTIVE CODING TRACE:**
Added to source:
``` python
deployment_how_to_approach_decision,
no_automated_deployment
```