# source

## ID

5

## url

https://streamsql.io/blog/microservices-with-machine-learning-feature-store

## archive url

https://tinyurl.com/ml-adds-u5

## title

Microservices Suck for Machine Learning (and what we did about it)

## source code

yes

## example

yes

## source type

Practitioner Audience Article

## author type

Practitioner

## references

**AXIAL CODING TRACE**: source and references

``` python
s5 = CClass(source, "s5", values={
    "title": "Microservices Suck for Machine Learning (and what we did about it)",
    "url": "https://streamsql.io/blog/microservices-with-machine-learning-feature-store",
    "archive url": "TODO",
    "author type": "Practitioner",
    "type": "Practitioner Audience Article",
    "example": True,
    "source code": True})
``` 

# coding

> Microservices Suck for Machine Learning (and what we did about it)

> I've worked on teams building ML-powered features, everything from personalization to propensity paywalls. Meetings to find and get access to data consumed my time, other days it was consumed building ETLs to get and clean that data. The worst situations were when I had to deal with microservice oriented architectures. I wouldn't advocate that we stop using microservices, but if you want to spin up an ML project in a strict microservice oriented architecture, you're doomed.

**OPEN CODING TRACE:**

- seems to be less about model as a service than data services with db per service
- etl pipeline

**AXIAL CODING TRACE:**

added codes to source:

``` python
data_service, data_service_can_use_db_per_service,
database_per_service, store_data_decision,
etl_pipeline, data_processing_automation_decision,
```

> In this article, I'll describe why microservice oriented architectures suck for machine learning. I'll then lay out the solution that companies like AirBnB and Uber landed on.

> Microservices have become the de-facto architectural choice for tech companies. Microservices allow small teams in large companies to build small, independent components. Teams can solve problems and meet requirements without having to retrofit into a giant monolith. It lets teams move faster. However, when overused, turning a user session token into a user profile can trigger twenty network calls.

**OPEN CODING TRACE:**

- practice: microservices vs monolith
- benefits: independent small teams, speed of development
- drawback/risk: chatty APIs but again this seems to focus mostly on data services (as model as a service is rather
  seldom called in comparison)

**AXIAL CODING TRACE:**

added:

``` python
chatty_data_access_api = CClass(force, "Chatty data access API")

add_force_relations({no_data_store: {chatty_data_access_api: neutral,
                                     },
                     shared_data_store: {chatty_data_access_api: positive,
                                         },
                     data_service: {chatty_data_access_api: negative},
                     event_store: {chatty_data_access_api: positive}
                     })

```

added codes to source:

``` python
integrating_ml_model_decision, model_in_monolith,
division_of_labour, distributed_development, independent_development,
separation_of_concerns, chatty_data_access_api
```

> I saw this video on Twitter recently, and it inspired me to write that post. The video is hilarious and captures this concept perfectly (Relevant text from video below)

> Product Mananger: Why is it so hard to display the birthday date on the profile page?

> Engineer: First we have to call the Bingo service to get the UserID, then we call Papaya to turn it into a user session token. We validate those with LMNOP, then we can pull the user info down from Racoon. But Racoon isn't guaranteed to have [the birthdate], so next we call ... Example: Reddit Recommender System Reddit Example

**OPEN CODING TRACE:**

- ok this sounds like entity service which is often seen as an anti-pattern, is particularly bad for data services

**AXIAL CODING TRACE:**

- no new concept to code

> Microservices are exceptionally painful when building machine learning models on complex data sources like user behavior. In these cases, to make a prediction, you need to pull from a multitude of microservices, that in turn pull from another multitude of services to get all the contextual information they need. For example, if you were to build a personalized reddit feed, you may want to know all communities a user is a part of, the top posts in these communities, any posts they've clicked on and liked, and more details. Adding additional inputs can provide more signal for the model to pick up on. For example, certain users may have drastically different behavior on the weekends than weekdays. Providing the model with the additional input - day of the week - allows it to pick up on those trends.
>

**OPEN CODING TRACE:**

- again this particular microservice design might not be the best one

**AXIAL CODING TRACE:**

- no new concept to code

> Each logical input to a machine learning model is called a feature. The process of hypothesizing, building, and testing inputs to a model is called feature engineering. Typically, feature engineering is one of the most time-consuming tasks for ML teams. Creatively coming up with new features is only part of the process; most of the time is spent finding the data you need, learning its peculiarities and edge cases, and building data pipelines to clean and transform it into a usable form.

> In a Microservice-based architecture, the only way to collect certain data is via an API call. A single model may have a wide variety of features that require API calls. For example, the model may need to query the user service for the communities a user is a part of, then it needs to query the top posts in each of those communities, and then query the likes that each of those garnered. Since systems are almost never built for this sort of usage, it will set in motion a domino effect of downstream network calls. Unlike a web front end, ML models are not flexible around missing features. It will have to wait for all of the requests to complete or risk providing garbage results. ML models themselves are quite computationally heavy and very slow; mix this with microservices, and it becomes impossible to serve real-time recommendations.

**OPEN CODING TRACE:**

- so ML plus entity services for getting data has drawbacks on performance and scalability
- here data for the prediction is meant, not getting training data
- data transformation
- data preparation
- model training


**AXIAL CODING TRACE:**

added speed performance and scalabilty forces to options of store_data_decision

added codes to source:

``` python
speed_performance, scalability,
data_preparation, 
            data_transformation, data_processing_tasks_decision,
            model_training, model_building_tasks_decision
```

> Getting Training Data Sucks

> Machine Learning models work by mimicking what they observe. To do this, a model requires a dataset of observed results and the inputs at the point in time of the observation to train on. Generating a training set in a microservice-based architecture sucks. In our Reddit example, we'd want an set of user upvotes along with the feature set at the moment they performed the upvote. Though it's far from ideal, we can technically get user upvotes and all post data by scraping the user and post microservices. However, point-in-time correctness is what makes this problem impossible in many cases. To train, we would need to know what rank the post was in the subreddit; however, the subreddit microservice is unlikely to support retroactive queries.

**OPEN CODING TRACE:**

- in addition, MS design might make it hard to get the training data, as MS are not designed for arbitrary queries
- mentions microservice-based data access api

**AXIAL CODING TRACE:**

added:

``` python
data_access_flexibility = CClass(force, "Data access flexibility")
```

added data_access_flexibility force to store_data decision options

added codes to source:

``` python
data_access_flexibility, data_access_interface_decision, service_based_data_access_api, 
```

> One option is to avoid the microservices entirely by breaking encapsulation and reading from database dumps.
>
>We could now by-pass the APIs and plug the datasets directly into Apache Spark or some other batch processing system. We can join tables, and work with the data in a far more convenient way. It also makes it possible to generate training datasets as fast as Spark can crunch the data. While using a data lake seems reasonable at first, it leads to the ML services being dependent on the schema of the raw data. The schema will surely change over time. Microservices are retired and replaced. Data inconsistencies, errors, and issues pile up. Eventually, each ML team has to maintain data pipelines that have devolved into archaic messes of spaghetti code to bandaid the situation. The pipelines are fragile to changes and take a vast amount of time and resources.

**OPEN CODING TRACE:**

- alternative: use database dumps to get training data, i.e. manual data ingestion
- use batch-processing system like Apache Spark for this (vs. the manual data ingestion and the data ingestion by
  request / service discussed above)
- use data lake for this -> bad choice: ML services being dependent on the schema of the raw data => data inconsistency,
  errors and issues; modifiability and maintainability issues.
- first mentioning of ML services (model as a service)
- first mentioning of data pipelines
- again data access api, might be local

**AXIAL CODING TRACE:**

added:

``` python
modifiability = CClass(force, "Modifiability")
data_consistency = CClass(force, "Data consistency")
```

added modifiability, data_consistency, maintainability forces to store_data_decision options

added codes to source:

``` python
data_ingestion_in_batches, data_ingestion_decision, data_ingestion_by_request, manual_data_ingestion,
modifiability, data_consistency, maintainability, 
model_as_a_service, data_pipeline, batch_or_realtime_data_processing_decision, batch_based_data_processing, 
data_processing_automation_decision,
            data_lake, data_lake_is_a_shared_data_store, shared_data_store,
            data_ingestion_in_batches_can_use_raw_data_store, data_ingestion_by_request_can_use_raw_data_store,
            manual_data_ingestion_can_use_raw_data_store,
            raw_data_store, local_data_access_api, 
            
```

>
> Data Pipeline Meme

> Another option is to utilize an event steaming platform like Apache Kafka, Apache Pulsar, or Segment to allow ML teams to subscribe to the event streams that they need. Many of the pitfalls for data lakes also apply to event streaming. However, unlike data lake dumps, event streams tend to have higher quality data. Since event streams often power mission-critical services, teams are held to a higher standard regarding data quality and documentation. Conversely, data lakes are exclusive to ML and analytics teams and are not held to a high standard.

**OPEN CODING TRACE:**

- use event streaming platform for ingestion and processing
- also: access data via event-platform

**AXIAL CODING TRACE:**

added codes to source:

``` python
real_time_data_processing, streamed_data_ingestion,  
```

> Event stream processing suffers from the cold start problem. Event streaming platforms are rarely configured to retain events for long, often it's for as little as one week. If you want to generate a new feature you may be stuck with only the last week of data to generate a training set. The cold start problem is further pronounced when creating a new stateful feature. A stateful feature requires you to aggregate events over some window of time. For example, the number of posts a user made in the last week. In these situations, it could take weeks to even start generating the training data set.

**OPEN CODING TRACE:**

- additional issue: cold start (data is only retained for a week or so, so you need to wait until you have enough data)

**AXIAL CODING TRACE:**

added:

``` python
avoid_cold_start_problem_for_training_data = CClass(force, "Avoid cold start problem for training data")
```

changed:

``` python
no_data_store = CClass(do_nothing_design_solution, "No data store")
add_decision_option_link(store_data_decision, no_data_store,
                         "get data without storing it, e.g. manual ingested raw data or only event streaming")
```

added avoid_cold_start_problem_for_training_data as force to store_data_decision

added codes to source:

``` python
avoid_cold_start_problem_for_training_data, no_data_store
```

> Feature engineering is an iterative process. You generate a hypothesis, build an experiment, and run a test. You either merge it into the main model or scrap it. The faster iterations happen, the faster the model gets better. If it takes weeks to perform a single test, ML teams lose their ability to perform their jobs. Teams end up wasting away plumbing data pipes and playing politics to get access to data, rather than building better models.

> Data Pipeline Meme How Teams are Solving the Microservice Issue

> The microservice problem isn't new or unique. Intriguingly, many companies independently landed on the same solution to the problem. AirBnB built Zipline, Uber built Michelangelo, and Lyft built Dryft. These systems are collectively called Feature Stores. What is a Feature Store?

> A feature store provides a standardized way for data scientists to define features. The feature store handles generating training data and providing online features for serving. It abstracts away the data engineering from the ML workflow. Under the hood, it coordinates multiple big data systems to seamlessly process both incoming events and a past events. If you're interested in the specific technologies, here's the layout of our feature store infrastructure.

> In the original "Reddit" microservice architecture, each service owned its own data. The posts microservice is the source of truth for data about posts, the user microservice is the source of truth for data about users, and so on. A feature store attempts to create its own views of this data in its own internal data structures. It does this by processing a stream of domain events into materialized state. Domain events are logical events such as when a user upvotes an article or when a new post is created. A materialized view is the result set from running a query on an event stream. So, if we wanted the model to know the number of posts that a user upvoted, then we could create a materialized view with this logic:


**OPEN CODING TRACE:**

- Feature Store: View on the microservice data with own internal data structures, obtained by processing domain events.
- feature engineering

**AXIAL CODING TRACE:**

- feature store coded below

**SELECTIVE CODING TRACE**:
- added addition testing codes for model training testing:

``` python
          testing, delivery_tasks_decision, testing_model_training_converges, testing_trained_model_quality, what_to_test_decision,
          feature_engineering
```

> SELECT user, COUNT(DISTINCT item) FROM upvote_stream GROUP BY user;

> All the materialized views exist in the same feature store and are preprocessed for ML use. We've merged all the microservice data that we care about into a monolithic data store. This eliminates the problems related to fetching real-time features from microservices. Now features can be fetched in one round trip. The feature store has the added benefit of being fault resistant, since the materialized views are stored in a highly-available and eventually-consistent data store. We own our own business logic in creating features since we are processing the raw events ourselves, so we are loosely coupled to the business logic in each microservice.


**OPEN CODING TRACE:**

- merges all data need for ML tasks into a monolithic data store for ML
- eventually-consistent data store: event store
- benefit: fault resistant (reliable), eventually consistent (leads to loose coupling), decoupled from business logic in
  MS

**AXIAL CODING TRACE:**

added:

``` python
event_store = CClass(practice, "Event store")

add_decision_option_link(store_data_decision, event_store,
                         "store data as events being exchanged in the systems")
                         
event_store_is_a_shared_data_store = \
    event_store.add_links(shared_data_store, role_name="to", stereotype_instances=is_a)[0]
    
```

added reliability and loose_coupling forces to store_data_decision options

added codes to source:

``` python
event_store, event_store_is_a_shared_data_store, reliability, loose_coupling 
```

> Event-Sourcing to generate Training Data
> Our features are kept consistent with incoming events. It's essentially a mirror of the microservices tables, but already preprocessed for use in ML and all in one place. Unlike the microservices, a feature store persists every incoming event indefinitely into a log. If we were to replay every single event in the log to a blank feature store, we would end up with the exact same materialized state. The event log becomes the source of truth for the system. This design pattern is called Event Sourcing.
>
**OPEN CODING TRACE:**

- benefit: data preprocessing possible
- Actually Feature Store uses Event Sourcing

**AXIAL CODING TRACE:**

added:

``` python
event_sourcing = CClass(practice, "Event sourcing")
event_store_can_offer_event_sourcing = \
    event_store.add_links(event_sourcing, role_name="to", stereotype_instances=can_be_combined_with,
                          label="can offer event sourcing capabilities")
```

added codes to source:

``` python
event_sourcing, event_store_can_offer_event_sourcing
```

> Event Sourcing lets us generate a training dataset for our models. To illustrate how this is done, let's take the Reddit example where we want to predict the next post that a user will upvote. The relevant domain events are streamed into the feature store, which then updates the model input features. Observed outcomes should also be streamed to the feature store, in this case every user upvote.

> feature_store.append_observation(userId, postId, now())

> Since the feature store maintains a log of every event and the logic to turn an event stream into state, it can get the state of the features at any point in time. To generate a training set, it loops through the observations and generates the feature set at that point. By combining the two, it ends up with a training dataset.

> def generate_training_set():
for observed in observations:
feature_store.process_events_until(observed.time)
features = feature_store.get_features(observed.userId)
yield (features, observed.postId)

**OPEN CODING TRACE:**
- can be used for training data generation
- feature store

**AXIAL CODING TRACE:**
- training data generation coded above already.

added:

``` python
feature_store_can_be_realized_by_event_sourcing = \
    feature_store.add_links(event_sourcing, role_name="to", stereotype_instances=can_be_realized_by)[0]
```

added codes to source:

``` python
feature_store,
feature_persistence_and_access_decision, store_features_without_specific_support, 
feature_store_can_be_realized_by_event_sourcing,
feature_discovery, feature_reuse
```


> Design Considerations

> A Feature Store allows us to decouple from the microservice architecture, and own our own features. However, building and maintaining a feature store is not free. Teams should consider the following points before deploying feature store infrastructure. System-Wide Event Streaming

**OPEN CODING TRACE:**
- requires Event Streaming system-wide

**AXIAL CODING TRACE:**
- already coded and in the various decisions that are affected.

> A feature store requires that domain events are passed through an event streaming platform like Kafka or Pulsar. This allows the feature store to materialize its state independetly of the microservices. Persisting the event log allows it to materialize the features at any point in time.

> Moving a large microservice-based system to use event streams is a monumental shift. Routines have to be injected to capture important events. This may require updating old, mission-critical microservices with a new dependency and new error conditions to catch. Another option is to use Change-Data-Capture semantics from each database to turn updates to a stream. However, the feature store is then vulnerable to schema changes inside a microservice's database.
>
**OPEN CODING TRACE:**

- how to transition to event streaming

**AXIAL CODING TRACE:**
- as this is not a must for feature stores (see other sources), no need for further coding

> Handling Event Schema Changes

> A feature store is still dependent on the schema of the event streams. If a stream changes its schema, or a microservice misbehaves and uploads garbage data, it can incapacitate the feature store downstream. Event streams schemas should be treated with the same care as a database schema. Migration procedures should be clear and tested. Events should be written with an extensible format like Protobuf or JSON.

**OPEN CODING TRACE:**
- downside: strong dependence on event stream schema

**AXIAL CODING TRACE:**
- more about event streaming as such, not really ML-specific

> Storage and Compute Capacity

> Processing and storing huge amounts of data is not free. In many cases, the feature store will repeat computations performed by individual microservices. A feature store trades infrastructure cost and complexity for developer velocity and ease of use. Building and maintaining a feature store requires both money and specialized engineers.

**OPEN CODING TRACE:**

- downside: higher computation load, costs, complexity for developer velocity and ease of use (in the sense of  Building and maintaining), so maintainability and modifiability

**AXIAL CODING TRACE:**

added:

``` python
computational_load = CClass(force, "Computational load")
infrastructure_costs = CClass(force, "Infrastructure costs")
development_velocity = CClass(force, "Development velocity")
```

added computational_load, infrastructure_costs as force to store_data_decision

added codes to source:

``` python
computational_load, infrastructure_costs, development_velocity
```

> Documenting and Sharing Input Features and Data Sources

> ML features are often applicable in many different use cases. Reddit may have many different ML-powered features that all use a user's activity to make decisions. Discovering and understanding features that others have built can speed up ML development time and provide inspiration in feature engineering. Since feature stores are a relatively new piece of architecture, teams will have to document how features should be published and shared. Check Out the StreamSQL Feature Store

> StreamSQL's Feature Store lets ML teams focus on building models, not data pipelines. ML teams can create a single feature definition in SQL to use in training and serving. Stateful features are automatically backfilled to eliminate waiting time. StreamSQL is foundation to engineer, share, and discover features across your team.


**OPEN CODING TRACE:**
- no new concept


**AXIAL CODING TRACE:**
- nothing coded 

added:

``` python

```

added codes to source:

``` python

```
