# source

## ID

6

## url

https://cloud.google.com/solutions/machine-learning/mlops-continuous-delivery-and-automation-pipelines-in-machine-learning?hl=en

## archive url

https://tinyurl.com/ml-adds-u6

## title

MLOps: Continuous delivery and automation pipelines in machine learning

## source code

no

## example

yes

## source type

Practitioner Audience Article

## author type

Practitioner

## references

**AXIAL CODING TRACE**: source and references

``` python
s6 = CClass(source, "s6", values={
    "title": "MLOps: Continuous delivery and automation pipelines in machine learning",
    "url": "https://cloud.google.com/solutions/machine-learning/mlops-continuous-delivery-and-automation-pipelines-in-machine-learning?hl=en",
    "archive url": "TODO",
    "author type": "Practitioner",
    "type": "Practitioner Audience Article",
    "example": True,
    "source code": False})
``` 

# coding

> MLOps: Continuous delivery and automation pipelines in machine learning

> This document discusses techniques for implementing and automating continuous integration (CI), continuous delivery (CD), and continuous training (CT) for machine learning (ML) systems.

> Data science and ML are becoming core capabilities for solving complex real-world problems, transforming industries, and delivering value in all domains. Currently, the ingredients for applying effective ML are available to you:

> Large datasets Inexpensive on-demand compute resources Specialized accelerators for ML on various cloud platforms Rapid advances in different ML research fields (such as computer vision, natural language understanding, and recommendations AI systems).

> Therefore, many businesses are investing in their data science teams and ML capabilities to develop predictive models that can deliver business value to their users.

> This document is for data scientists and ML engineers who want to apply DevOps principles to ML systems (MLOps). MLOps is an ML engineering culture and practice that aims at unifying ML system development (Dev) and ML system operation (Ops). Practicing MLOps means that you advocate for automation and monitoring at all steps of ML system construction, including integration, testing, releasing, deployment and infrastructure management.

> Data scientists can implement and train an ML model with predictive performance on an offline holdout dataset, given relevant training data for their use case. However, the real challenge isn't building an ML model, the challenge is building an integrated ML system and to continuously operate it in production. With the long history of production ML services at Google, we've learned that there can be many pitfalls in operating ML-based systems in production. Some of these pitfalls are summarized in Machine Learning: The high-interest credit card of technical debt.

> As shown in the following diagram, only a small fraction of a real-world ML system is composed of the ML code. The required surrounding elements are vast and complex.

> ML systems is more than ML code.

> Figure 1. Elements for ML systems. Adapted from Hidden Technical Debt in Machine Learning Systems.

> In this diagram, the rest of the system is composed of configuration, automation, data collection, data verification, testing and debugging, resource management, model analysis, process and metadata management, serving infrastructure, and monitoring.

> To develop and operate complex systems like these, you can apply DevOps principles to ML systems (MLOps). This document covers concepts to consider when setting up an MLOps environment for your data science practices, such as CI, CD, and CT in ML.

> The following topics are discussed:

> DevOps versus MLOps Steps for developing ML models MLOps maturity levels

> DevOps versus MLOps

> DevOps is a popular practice in developing and operating large-scale software systems. This practice provides benefits such as shortening the development cycles, increasing deployment velocity, and dependable releases. To achieve these benefits, you introduce two concepts in the software system development:

> Continuous Integration (CI)
Continuous Delivery (CD)



> An ML system is a software system, so similar practices apply to help guarantee that you can reliably build and operate ML systems at scale.

**OPEN CODING TRACE:**

- ci/cd pipeline
- development velocity
- deployment velocity
- dependable releases
- reliable delivery/operations

**AXIAL CODING TRACE:**

added:

``` python
deployment_velocity = CClass(force, "Deployment velocity")
dependable_releases = CClass(force, "Dependable releases")

```

added development_velocity, deployment_velocity, dependable_releases, reliability forces to automated_integration_and_delivery_in_ml_context_decision

added codes to source:

``` python
ci_cd_pipeline, automated_integration_and_delivery_in_ml_context_decision, development_velocity, deployment_velocity, dependable_releases, reliability
```

**SELECTIVE CODING TRACE**:
- added support for "mlops" in `ml_adds_sources_codes.py` following addition of s20:
``` python
        mlops, mlops_decision, no_mlops
```

> However, ML systems differ from other software systems in the following ways:

> Team skills: In an ML project, the team usually includes data scientists or ML researchers, who focus on exploratory data analysis, model development, and experimentation. These members might not be experienced software engineers who can build production-class services.

> Development: ML is experimental in nature. You should try different features, algorithms, modeling techniques, and parameter configurations to find what works best for the problem as quickly as possible. The challenge is tracking what worked and what didn't, and maintaining reproducibility while maximizing code reusability.

> Testing: Testing an ML system is more involved than testing other software systems. In addition to typical unit and integration tests, you need data validation, trained model quality evaluation, and model validation.

> Deployment: In ML systems, deployment isn't as simple as deploying an offline-trained ML model as a prediction service. ML systems can require you to deploy a multi-step pipeline to automatically retrain and deploy model. This pipeline adds complexity and requires you to automate steps that are manually done before deployment by data scientists to train and validate new models.

> Production: ML models can have reduced performance not only due to suboptimal coding, but also due to constantly evolving data profiles. In other words, models can decay in more ways than conventional software systems, and you need to consider this degradation. Therefore, you need to track summary statistics of your data and monitor the online performance of your model to send notifications or roll back when values deviate from your expectations.


**OPEN CODING TRACE:**

- independent development in different teams and reproducibility as a force
- training multiple model versions
- testing as part of ci/cd; validation during model engineering; validation 
  of data as part of data processing
- retraining and redeploy a model automatically
- speed performance, accuracy
- monitoring

**AXIAL CODING TRACE:**

added:

``` python
reproducibility = CClass(force, "Reproducibility")

```

added reproducibility, independent_development to automated_integration_and_delivery_in_ml_context_decision options

added codes to source:

``` python
reproducibility, independent_development, train_multiple_model_versions,
            testing, delivery_tasks_decision, model_validation,
            model_building_tasks_decision, model_building_pipeline,
            model_building_decision, data_processing_automation_decision,
            retrain_model,
            monitoring_model_adjustment_decision, automated_model_adjustment,
            automated_model_adjustment_can_use_retrain_model,
                        model_performance_monitoring,
            speed_performance_monitoring,
            model_accuracy_monitoring,
            model_performance_monitoring_can_use_model_accuracy_monitoring,
            model_performance_monitoring_can_use_speed_performance_monitoring,
            monitoring_in_production_decision,
            speed_performance, model_accuracy,
```

**SELECTIVE CODING TRACE**:

- After what to test decision modelled below, we added here:

``` python
testing_trained_model_quality = CClass(practice, "Testing trained model quality")
add_decision_option_link(what_to_test_decision, testing_trained_model_quality)
```

- and added codes to source:

``` python
unit_testing_system_components, system_integration_testing, testing_trained_model_quality

```

> ML and other software systems are similar in continuous integration of source control, unit testing, integration testing, and continuous delivery of the software module or the package. However, in ML, there are a few notable differences:

>    CI is no longer only about testing and validating code and components, but also testing and validating data, data schemas, and models.
    CD is no longer about a single software package or a service, but a system (an ML training pipeline) that should automatically deploy another service (model prediction service).
    CT is a new property, unique to ML systems, that's concerned with automatically retraining and serving the models.

**OPEN CODING TRACE:**
- version control in CI/CD
- again testing
- packaging
- model eneingeering pipeline (again)
- retrain (again)
- data validation

**AXIAL CODING TRACE:**

added:

``` python
data_validation = CClass(practice, "Data validation")

```

added codes to source:

``` python
            version_control_repository, 
            packaging, 
           data_validation
```


>The following section discusses the typical steps for training and evaluating an ML model to serve as a prediction
service. Data science steps for ML

>In any ML project, after you define the business use case and establish the success criteria, the process of delivering
an ML model to production involves the following steps. These steps can be completed manually or can be completed by an
automatic pipeline.

>    Data extraction: You select and integrate the relevant data from various data sources for the ML task.
    Data analysis: You perform exploratory data analysis (EDA) to understand the available data for building the ML model. This process leads to the following:
        Understanding the data schema and characteristics that are expected by the model.
        Identifying the data preparation and feature engineering that are needed for the model.
    Data preparation: The data is prepared for the ML task. This preparation involves data cleaning, where you split the data into training, validation, and test sets. You also apply data transformations and feature engineering to the model that solves the target task. The output of this step are the data splits in the prepared format.
    Model training: The data scientist implements different algorithms with the prepared data to train various ML models. In addition, you subject the implemented algorithms to hyperparameter tuning to get the best performing ML model. The output of this step is a trained model.
    Model evaluation: The model is evaluated on a holdout test set to evaluate the model quality. The output of this step is a set of metrics to assess the quality of the model.
    Model validation: The model is confirmed to be adequate for deployment—that its predictive performance is better than a certain baseline.
    Model serving: The validated model is deployed to a target environment to serve predictions. This deployment can be one of the following:
        Microservices with a REST API to serve online predictions.
        An embedded model to an edge or mobile device.
        Part of a batch prediction system.
    Model monitoring: The model predictive performance is monitored to potentially invoke a new iteration in the ML process.

**OPEN CODING TRACE:**
- largely general ML engineering steps
- model serving can be integrated as modular monolith (mobile-> offline support), microservice (then more online support), or batch prediction component
- again model monitoring
- data ingestion and processing decisions
- data extraction, selection, preparation, transformation
- hyperparameter tuning of the ML model


**AXIAL CODING TRACE:**

added:

``` python
model_in_batch_prediction_system =  CClass(practice, "Model in batch prediction system")
add_decision_option_link(integrating_ml_model_decision, model_in_batch_prediction_system,
                         "call an external batch prediction system")                         
```

added codes to source:

``` python
offline_support, online_support,
model_as_module_in_modular_monolith, model_as_a_service, 
model_in_batch_prediction_system, integrating_ml_model_decision,
            data_ingestion_decision, data_processing_automation_decision
            data_extraction, data_processing_tasks_decision,
            data_preparation, 
            data_transformation, 
            data_selection, model_hyperparameter_tuning
```

>The level of automation of these steps defines the maturity of the ML process, which reflects the velocity of training
new models given new data or training new models given new implementations. The following sections describe three levels
of MLOps, starting from the most common level, which involves no automation, up to automating both ML and CI/CD
pipelines. MLOps level 0: Manual process

>Many teams have data scientists and ML researchers who can build state-of-the-art models, but their process for building
and deploying ML models is entirely manual. This is considered the basic level of maturity, or level 0. The following
diagram shows the workflow of this process.


>Workflow of manual ML steps for MLOps level 0.

>Figure 2. Manual ML steps to serve the model as a prediction service. Characteristics

>The following list highlights the characteristics of the MLOps level 0 process, as shown in Figure 2:

>    Manual, script-driven, and interactive process: Every step is manual, including data analysis, data preparation, model training, and validation. It requires manual execution of each step, and manual transition from one step to another. This process is usually driven by experimental code that is written and executed in notebooks by data scientist interactively, until a workable model is produced.

>    Disconnection between ML and operations: The process separates data scientists who create the model and engineers who serve the model as a prediction service. The data scientists hand over a trained model as an artifact to the engineering team to deploy on their API infrastructure. This handoff can include putting the trained model in a storage location, checking the model object into a code repository, or uploading it to a models registry. Then engineers who deploy the model need to make the required features available in production for low-latency serving, which can lead to training-serving skew.

>    Infrequent release iterations: The process assumes that your data science team manages a few models that don't change frequently—either changing model implementation or retraining the model with new data. A new model version is deployed only a couple of times per year.

>    No CI: Because few implementation changes are assumed, CI is ignored. Usually, testing the code is part of the notebooks or script execution. The scripts and notebooks that implement the experiment steps are source controlled, and they produce artifacts such as trained models, evaluation metrics, and visualizations.

>    No CD: Because there aren't frequent model version deployments, CD isn't considered.

>    Deployment refers to the prediction service: The process is concerned only with deploying the trained model as a prediction service (for example, a microservice with a REST API), rather than deploying the entire ML system.

>    Lack of active performance monitoring: The process doesn't track or log the model predictions and actions, which are required in order to detect model performance degradation and other model behavioral drifts.


> The engineering team might have their own complex setup for API configuration, testing, and deployment, including
security, regression, and load and canary testing. In addition, production deployment of a new version of an ML model
usually goes through A/B testing or online experiments before the model is promoted to serve all the prediction request
traffic. Challenges

> MLOps level 0 is common in many businesses that are beginning to apply ML to their use cases. This manual,
data-scientist-driven process might be sufficient when models are rarely changed or trained. In practice, models often
break when they are deployed in the real world. The models fail to adapt to changes in the dynamics of the environment,
or changes in the data that describes the environment. For more information, see Why Machine Learning Models Crash and
Burn in Production.

**OPEN CODING TRACE:**
- all stages manual: ingestion, processing, deployment, model building,
productin monitoring, model adjustment
- force: low latency during model serving  (i.e. integration decision)
- force: modifiability for ci/cd automation

**AXIAL CODING TRACE:**

added modifiability force to automated_integration_and_delivery_in_ml_context_decision decision options

added codes to source:

``` python
manual_data_ingestion, no_data_processing_automation, no_integration_or_delivery_automation,
            model_building_in_development_tool, no_in_production_monitoring, manual_model_adjustment, low_latency, modifiability
```

**SELECTIVE CODING TRACE**
Multiple models in production is positive for shadow testing and enables a/b testing, canary testing
Added n_versions_in_production and the above for this source

>To address these challenges and to maintain your model's accuracy in production, you need to do the following:

>    Actively monitor the quality of your model in production: Monitoring lets you detect performance degradation and model staleness. It acts as a cue to a new experimentation iteration and (manual) retraining of the model on new data.

>    Frequently retrain your production models: To capture the evolving and emerging patterns, you need to retrain your model with the most recent data. For example, if your app recommends fashion products using ML, its recommendations should adapt to the latest trends and products.

>    Continuously experiment with new implementations to produce the model: To harness the latest ideas and advances in technology, you need to try out new implementations such as feature engineering, model architecture, and hyperparameters. For example, if you use computer vision in face detection, face patterns are fixed, but better new techniques can improve the detection accuracy.

>To address the challenges of this manual process, MLOps practices for CI/CD and CT are helpful. By deploying an ML
training pipeline, you can enable CT, and you can set up a CI/CD system to rapidly test, build, and deploy new
implementations of the ML pipeline. These features are discussed in more detail in the next sections. MLOps level 1: ML
pipeline automation
> 

**OPEN CODING TRACE:**
- building as CI/CD task

**AXIAL CODING TRACE:**

added codes to source:

``` python
building, 
```

>The goal of level 1 is to perform continuous training of the model by automating the ML pipeline; this lets you achieve
continuous delivery of model prediction service. To automate the process of using new data to retrain models in
production, you need to introduce automated data and model validation steps to the pipeline, as well as pipeline
triggers and metadata management.

**OPEN CODING TRACE:**
- model validation mentioned (again)
- again data processing (assuming pipeline with validation)

**AXIAL CODING TRACE:**

added codes to source:

``` python
data_pipeline, batch_or_realtime_data_processing_decision, real_time_data_processing,
            batch_based_data_processing,
```

> The following figure is a schematic representation of an automated ML pipeline for CT.

>Workflow of ML pipeline for CT.

>Figure 3. ML pipeline automation for CT. Characteristics

>The following list highlights the characteristics of the MLOps level 1 setup, as shown in Figure 3:

>    Rapid experiment: The steps of the ML experiment are orchestrated. The transition between steps is automated, which leads to rapid iteration of experiments and better readiness to move the whole pipeline to production.

>    CT of the model in production: The model is automatically trained in production using fresh data based on live pipeline triggers, which are discussed in the next section.

>    Experimental-operational symmetry: The pipeline implementation that is used in the development or experiment environment is used in the preproduction and production environment, which is a key aspect of MLOps practice for unifying DevOps.

>    Modularized code for components and pipelines: To construct ML pipelines, components need to be reusable, composable, and potentially shareable across ML pipelines. Therefore, while the EDA code can still live in notebooks, the source code for components must be modularized. In addition, components should ideally be containerized to do the following:
        Decouple the execution environment from the custom code runtime.
        Make code reproducible between development and production environments.
        Isolate each component in the pipeline. Components can have their own version of the runtime environment, and have different languages and libraries.
>
>    Continuous delivery of models: An ML pipeline in production continuously delivers prediction services to new models that are trained on new data. The model deployment step, which serves the trained and validated model as a prediction service for online predictions, is automated.

 >   Pipeline deployment: In level 0, you deploy a trained model as a prediction service to production. For level 1, you deploy a whole training pipeline, which automatically and recurrently runs to serve the trained model as the prediction service.


**OPEN CODING TRACE:**
- loose coupling in model building pipeline
- possibly use data science tools from pipeline
- experimental operational symmetry in model building pipeline
- reproducibility in data processing and model building pipelines

**AXIAL CODING TRACE:**

added:

``` python
experimental_operational_symmetry = CClass(force, "Experimental Operational Symmetry")
```

added experimental_operational_symmetry, reproducibility  to model_building_decision options

added reproducibility  to 
data_processing_automation_decision options

added codes to source:

``` python
loose_coupling,
model_building_in_development_tool, development_tool_facade, 
            development_tool_facade_can_be_combined_with_model_building_in_development_tool,
development_tool_facade_can_be_combined_with_model_building_in_development_tool, 
reproducibility,
experimental_operational_symmetry
```


>Additional components

>This section discusses the components that you need to add to the architecture to enable ML continuous training. Data
and model validation

>When you deploy your ML pipeline to production, one or more of the triggers discussed in the ML pipeline triggers
section automatically executes the pipeline. The pipeline expects new, live data to produce a new model version that is
trained on the new data (as shown in Figure 3). Therefore, automated data validation and model validation steps are
required in the production pipeline to ensure the following expected behavior:

 >   Data validation: This step is required before model training to decide whether you should retrain the model or stop the execution of the pipeline. This decision is automatically made if the following was identified by the pipeline.
        Data schema skews: These skews are considered anomalies in the input data, which means that the downstream pipeline steps, including data processing and model training, receives data that doesn't comply with the expected schema. In this case, you should stop the pipeline so the data science team can investigate. The team might release a fix or an update to the pipeline to handle these changes in the schema. Schema skews include receiving unexpected features, not receiving all the expected features, or receiving features with unexpected values.
        Data values skews: These skews are significant changes in the statistical properties of data, which means that data patterns are changing, and you need to trigger a retraining of the model to capture these changes.

>    Model validation: This step occurs after you successfully train the model given the new data. You evaluate and validate the model before it's promoted to production. This offline model validation step consists of the following.
        Producing evaluation metric values using the trained model on a test dataset to assess the model's predictive quality.
        Comparing the evaluation metric values produced by your newly trained model to the current model, for example, production model, baseline model, or other business-requirement models. You make sure that the new model produces better performance than the current model before promoting it to production.
        Making sure that the performance of the model is consistent on various segments of the data. For example, your newly trained customer churn model might produce an overall better predictive accuracy compared to the previous model, but the accuracy values per customer region might have large variance.
        Making sure that you test your model for deployment, including infrastructure compatibility and consistency with the prediction service API.

>In addition to offline model validation, a newly deployed model undergoes online model validation—in a canary deployment
or an A/B testing setup—before it serves prediction for the online traffic. 
>

**OPEN CODING TRACE:**
- model and data validation again, model training
- online_support, offline_support
- a/b testing
- canary testing

**AXIAL CODING TRACE:**

added codes to source:

``` python
 online_support, offline_support, a_b_testing, 
            canary_testing, model_training, 
            
```

> 
> >Feature store

>An optional additional component for level 1 ML pipeline automation is a feature store. A feature store is a centralized
repository where you standardize the definition, storage, and access of features for training and serving. A feature
store needs to provide an API for both high-throughput batch serving and low-latency real-time serving for the feature
values, and to support both training and serving workloads.

>The feature store helps data scientists do the following:

>    Discover and reuse available feature sets for their entities, instead of re-creating the same or similar ones.
    Avoid having similar features that have different definitions by maintaining features and their related metadata.
    Serve up-to-date feature values from the feature store.

>    Avoid training-serving skew by using the feature store as the data source for experimentation, continuous training, and online serving. This approach makes sure that the features used for training are the same ones used during serving:
        For experimentation, data scientists can get an offline extract from the feature store to run their experiments.
        For continuous training, the automated ML training pipeline can fetch a batch of the up-to-date feature values of the dataset that are used for the training task.
        For online prediction, the prediction service can fetch in a batch of the feature values related to the requested entity, such as customer demographic features, product features, and current session aggregation features.

**OPEN CODING TRACE:**
- feature store as an option
- discover and reuse features force
- maintainability force (maintenance of features and metadata)
- Avoid training-serving skew force
- online/offline forces
- high throughput force
- low latency force
- Near real-time support force
- feature engineering
- data access API is needed (service for online, maybe local for offline); near-realtime force for it as well

**AXIAL CODING TRACE:**

added:

``` python
batch_feature_api = CClass(practice, "Batch feature API")
real_time_feature_api = CClass(practice, "Real-time feature API")

feature_store_can_include_batch_feature_api = \
    feature_store.add_links(batch_feature_api, role_name="to", stereotype_instances=can_include)[0]
feature_store_can_include_real_time_feature_api = \
    feature_store.add_links(real_time_feature_api, role_name="to", stereotype_instances=can_include)[0]

feature_reuse = CClass(force, "Feature reuse")
feature_discovery = CClass(force, "Feature discovery")
avoid_training_serving_skew = CClass(force, "Avoid training/serving skew")
   
add_links({batch_feature_api: data_access_interface_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to design data access api")
add_links({real_time_feature_api: data_access_interface_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet,
          label="how to design data access api")         
```

added codes to source:

``` python
feature_store, 
feature_persistence_and_access_decision, store_features_without_specific_support, 
feature_store_can_include_batch_feature_api, 
feature_store_can_include_real_time_feature_api,
feature_reuse, feature_discovery, maintainability, avoid_training_serving_skew,
online_support, offline_support, data_throughput, low_latency, near_real_time_support
batch_feature_api, real_time_feature_api, data_access_interface_decision, 
            service_based_data_access_api, local_data_access_api,
```

**SELECTIVE CODING TRACE**:

Experimentation with an extract of features mentioned (incremental)
Continuous training with batched features mentioned.

Added codes to source:

``` python
batch_based_learning, incremental_learning, when_and_how_to_train_the_model_decision,
```

>Metadata management

>Information about each execution of the ML pipeline is recorded in order to help with data and artifacts lineage,
reproducibility, and comparisons. It also helps you debug errors and anomalies. Each time you execute the pipeline, the
ML metadata store records the following metadata:

>    The pipeline and component versions that were executed.
    The start and end date, time, and how long the pipeline took to complete each of the steps.
    The executor of the pipeline.
    The parameter arguments that were passed to the pipeline.
    The pointers to the artifacts produced by each step of the pipeline, such as the location of prepared data, validation anomalies, computed statistics, and extracted vocabulary from the categorical features. Tracking these intermediate outputs helps you resume the pipeline from the most recent step if the pipeline stopped due to a failed step, without having to re-execute the steps that have already completed.
    A pointer to the previous trained model if you need to roll back to a previous model version or if you need to produce evaluation metrics for a previous model version when the pipeline is given new test data during the model validation step.
    The model evaluation metrics produced during the model evaluation step for both the training and the testing sets. These metrics help you compare the performance of a newly trained model to the recorded performance of the previous model during the model validation step.


**OPEN CODING TRACE:**
- metadata store
- used by CI/CD and model building pipelines
- rollback to previous version mentioned as in S4, but here it is its own practice
- can use n versions in production or not.

**AXIAL CODING TRACE:**

added:

``` python
ml_metadata_store = CClass(practice, "Machine learning/data science metadata store")
ci_cd_pipeline_can_use_ml_metadata_store = \
    ci_cd_pipeline.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                             label="store all pipeline parameters and versions as metadata")[0]
model_building_pipeline_can_use_ml_metadata_store = \
    model_building_pipeline.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                         label = "store all model building, model versioning, "+
                                                 "and valiation metrics metadata")[0]
 
rollback_to_previous_model_version_can_use_n_versions_in_production = \
    rollback_to_previous_model_version.add_links(n_versions_in_production, role_name="to", stereotype_instances=can_use,
                                                 label="keep previous version in production to "+
                                                       "enable immediate rollback")[0]

rollback_to_previous_model_version = CClass(practice, "Rollback to previous model version")

n_versions_in_production_can_use_metadata_store = \
    n_versions_in_production.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                       label="manage version metadata")[0]
rollback_to_previous_model_version_can_use_metadata_store = \
    rollback_to_previous_model_version.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                                 label="manage version metadata")[0]
single_model_in_production_can_use_metadata_store = \
    single_model_in_production.add_links(ml_metadata_store, role_name="to", stereotype_instances=can_use,
                                         label="manage version metadata")[0] 
        
```

added codes to source:

``` python
ml_metadata_store, ci_cd_pipeline_can_use_ml_metadata_store, 
model_building_pipeline_can_use_ml_metadata_store, 
rollback_to_previous_model_version_can_use_n_versions_in_production,
rollback_to_previous_model_version,
n_versions_in_production_can_use_metadata_store, 
rollback_to_previous_model_version_can_use_metadata_store, 
single_model_in_production_can_use_metadata_store,
rolling_upgrades, safe_model_transition, n_versions_in_production, single_model_in_production,
deploying_model_versions_in_production_decision
```


>ML pipeline triggers

>You can automate the ML production pipelines to retrain the models with new data, depending on your use case:

>    On demand: Ad-hoc manual execution of the pipeline.
    On a schedule: New, labelled data is systematically available for the ML system on a daily, weekly, or monthly basis. The retraining frequency also depends on how frequently the data patterns change, and how expensive it is to retrain your models.
    On availability of new training data: New data isn't systematically available for the ML system and instead is available on an ad-hoc basis when new data is collected and made available in the source databases.
    On model performance degradation: The model is retrained when there is noticeable performance degradation.
    On significant changes in the data distributions (concept drift). It's hard to assess the complete performance of the online model, but you notice significant changes on the data distributions of the features that are used to perform the prediction. These changes suggest that your model has gone stale, and that needs to be retrained on fresh data.


**OPEN CODING TRACE:**
- various kinds of pipeline triggers listed
- pipeline trigger decision for model building pipeline, as well as CI/CD pipeline

**AXIAL CODING TRACE:**

added:

``` python
on_demand_trigger = CClass(practice, "On-demand trigger")
on_schedule_trigger = CClass(practice, "On schedule trigger")
on_availability_of_new_training_data_trigger = CClass(practice, "On availability of new training data trigger")
on_model_performance_degradation_trigger = CClass(practice, "On model performance degradation trigger")
on_changes_in_the_data_distribution_trigger = CClass(practice, "On changes in the data distribution trigger")

trigger_pipeline_or_orchestrator_decision = CClass(decision, "How to trigger a machine learning/data science " +
                                                   "pipeline or orchestrator?")

add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_demand_trigger,
                         "ad-hoc manual execution of the pipeline or triggered by other pipeline or service")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_schedule_trigger,
                         "frequent pipeline runs e.g. on a daily, weekly, or monthly basis")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_availability_of_new_training_data_trigger,
                         "when new data is collected and made available in data stores")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_model_performance_degradation_trigger,
                         "when a noticeable performance degradation is observed")
add_decision_option_link(trigger_pipeline_or_orchestrator_decision, on_changes_in_the_data_distribution_trigger,
                         "when a significant changes in the data distributions of features is observed")

add_links({trigger_pipeline_or_orchestrator_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({automated_integration_and_delivery_in_ml_context_decision: trigger_pipeline_or_orchestrator_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)


on_model_performance_degradation_trigger_requires_model_performance_monitoring = \
    on_model_performance_degradation_trigger.add_links(model_performance_monitoring, role_name="to",
                                                       stereotype_instances=requires)[0]

add_links({model_building_decision: trigger_pipeline_or_orchestrator_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)

```

added codes to source:

``` python
on_demand_trigger, on_schedule_trigger, on_availability_of_new_training_data_trigger, on_model_performance_degradation_trigger, on_changes_in_the_data_distribution_trigger,
trigger_pipeline_or_orchestrator_decision, 
on_model_performance_degradation_trigger_requires_model_performance_monitoring
```


> Challenges

> Assuming that new implementations of the pipeline aren't frequently deployed and you are managing only a few pipelines,
you usually manually test the pipeline and its components. In addition, you manually deploy new pipeline
implementations. You also submit the tested source code for the pipeline to the IT team to deploy to the target
environment. This setup is suitable when you deploy new models based on new data, rather than based on new ML ideas.

>However, you need to try new ML ideas and rapidly deploy new implementations of the ML components. If you manage many ML
pipelines in production, you need a CI/CD setup to automate the build, test, and deployment of ML pipelines.

>MLOps level 2: CI/CD pipeline automation

>For a rapid and reliable update of the pipelines in production, you need a robust automated CI/CD system. This automated
CI/CD system lets your data scientists rapidly explore new ideas around feature engineering, model architecture, and
hyperparameters. They can implement these ideas and automatically build, test, and deploy the new pipeline components to
the target environment.

**OPEN CODING TRACE:**
- idea to set up a CI/CD pipeline for automating the build, test, and deployment of ML pipelines
  (data and model building pipelines)
- force: reliability better with pipelines
- model building pipeline per model as a parallelization is used here as a consequence
- hyperparameter tuning again

**AXIAL CODING TRACE:**

added:

``` python
deployment_of_models_to_be_served = CClass(practice,
                                           "Build, test, and deploy models to be served and other system components")
deployment_of_ml_pipelines = CClass(practice,
                                       "Build, test, and deploy machine learning or data science pipelines " +
                                       "and components needed during pipeline runs")
no_automated_deployment = CClass(practice, "No automated deployment")

ci_cd_pipeline_model_deployment_realizes_model_building_pipeline = \
    ci_cd_pipeline_model_deployment.add_links(model_building_pipeline, role_name="to", stereotype_instances=realizes)[0]

ci_cd_pipeline_model_deployment_realizes_data_pipelines = \
    ci_cd_pipeline_model_deployment.add_links(data_pipeline, role_name="to", stereotype_instances=realizes)[0]

how_to_approach_deployment_decision = CClass(decision, "What to build, test, and deploy" +
                                                         " automatically by CI/CD pipelines?")
add_decision_option_link(how_to_approach_deployment_decision, no_automated_deployment)
add_decision_option_link(how_to_approach_deployment_decision, semi_automated_model_deployment)
add_decision_option_link(how_to_approach_deployment_decision, ci_cd_pipeline_model_deployment)

add_links({how_to_approach_deployment_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({how_to_approach_deployment_decision: automated_integration_and_delivery_in_ml_context_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)
```

added codes to source:

``` python
semi_automated_model_deployment, ci_cd_pipeline_model_deployment, 
no_automated_deployment, 
ci_cd_pipeline_model_deployment_realizes_model_building_pipeline, 
ci_cd_pipeline_model_deployment_realizes_data_pipelines,
how_to_approach_deployment_decision,
model_building_parallelization_decision,
model_building_pipeline_per_model,
no_model_building_parallelization,
ci_cd_pipeline_can_include_model_building_pipeline,
```

**SELECTIVE CODING TRACE:**
- Correction of the following upon developing the theory further:
`deployment_of_models_to_be_served = CClass(practice,
                                           "Build, test, and deploy models to be served and other system components")`
becomes
`semi_automated_model_deployment = CClass(practice,
                                           "Manually build, test, and deploy model pipelines and other system components")`

`deployment_of_ml_pipelines = CClass(practice,
                                       "Build, test, and deploy machine learning or data science pipelines " +
                                       "and components needed during pipeline runs")`
becomes
`ci_cd_pipeline_model_deployment = CClass(practice,
                                         "Build, test, and deploy machine learning or data science models " +
                                         "based on CI/CD pipeline automation")`

- Semi-automated prompts us to consider `automated_integration_and_delivery_decision`. Added:
add_decision_option_link(ci_cd_pipeline_model_deployment, automated_integration_and_delivery_decision)

>The following diagram shows the implementation of the ML pipeline using CI/CD, which has the characteristics of the
automated ML pipelines setup plus the automated CI/CD routines.

>Workflow of the ML pipeline with CI/CD integration.

>Figure 4. CI/CD and automated ML pipeline.

>This MLOps setup includes the following components:

>    Source control
    Test and build services
    Deployment services
    Model registry
    Feature store
    ML metadata store
    ML pipeline orchestrator

**OPEN CODING TRACE:**
- ML pipeline orchestrator component
- features store
- model registry

**AXIAL CODING TRACE:**
- feature store, testing, building, and metadata store already added.

added codes to source:

``` python
model_registry, ci_cd_pipeline_can_use_model_registry, model_building_pipeline_can_use_model_registry,
deployment, 
ml_orchestrator, ci_cd_pipeline_can_be_combined_with_ml_orchestrator
```

**SELECTIVE CODING TRACE**:

Aspects of AutoML mentioned.

- added:
``` python
auto_ml_decision,
auto_ml,
no_auto_ml
```

>Characteristics

>The following diagram shows the stages of the ML CI/CD automation pipeline:

>Characteristics of the CI/CD automated ML pipeline.

>Figure 5. Stages of the CI/CD automated ML pipeline.

>The pipeline consists of the following stages:

>    Development and experimentation: You iteratively try out new ML algorithms and new modeling where the experiment steps are orchestrated. The output of this stage is the source code of the ML pipeline steps that are then pushed to a source repository.

>    Pipeline continuous integration: You build source code and run various tests. The outputs of this stage are pipeline components (packages, executables, and artifacts) to be deployed in a later stage.

>    Pipeline continuous delivery: You deploy the artifacts produced by the CI stage to the target environment. The output of this stage is a deployed pipeline with the new implementation of the model.

>    Automated triggering: The pipeline is automatically executed in production based on a schedule or in response to a trigger. The output of this stage is a trained model that is pushed to the model registry.

>    Model continuous delivery: You serve the trained model as a prediction service for the predictions. The output of this stage is a deployed model prediction service.

>    Monitoring: You collect statistics on the model performance based on live data. The output of this stage is a trigger to execute the pipeline or to execute a new experiment cycle.

>The data analysis step is still a manual process for data scientists before the pipeline starts a new iteration of the
experiment. The model analysis step is also a manual process. 
> 

**OPEN CODING TRACE:**
- forces: iterative development, process and work automation through continuous delivery and automated triggering, observability, 

**AXIAL CODING TRACE:**
added:

``` python
add_force_relations({no_automated_deployment: {process_and_work_automation: very_negative,
                                               iterative_development: negative,
                                               observability: very_negative,
                                               },
                     semi_automated_model_deployment: {process_and_work_automation: positive,
                                                         iterative_development: positive,
                                                         observability: very_positive,
                                                         },
                     ci_cd_pipeline_model_deployment: {process_and_work_automation: very_positive,
                                                     iterative_development: very_positive,
                                                     observability: very_positive
                                                     },
                     })
```



>Continuous integration

>In this setup, the pipeline and its components are built, tested, and packaged when new code is committed or pushed to
the source code repository. Besides building packages, container images, and executables, the CI process can include the
following tests:

>    Unit testing your feature engineering logic.

>    Unit testing the different methods implemented in your model. For example, you have a function that accepts a categorical data column and you encode the function as a one-hot feature.

>    Testing that your model training converges (that is, the loss of your model goes down by iterations and overfits a few sample records).

>    Testing that your model training doesn't produce NaN values due to dividing by zero or manipulating small or large values.

>    Testing that each component in the pipeline produces the expected artifacts.

>    Testing integration between pipeline components.


**OPEN CODING TRACE:**
- detailed list of possible tests

**AXIAL CODING TRACE:**
- testing_trained_model_quality coded above e.g. covers model training doesn't produce NaN values 

added:

``` python
unit_testing_feature_engineering_logic = CClass(practice, "Unit testing feature engineering logic")
unit_testing_different_methods_implemented_in_a_model = CClass(practice,
                                                               "Unit testing different methods implemented in a model")
testing_model_training_converges = CClass(practice, "Testing model training converges")
unit_testing_pipeline_components = CClass(practice, "Unit testing pipeline components")
integration_testing_pipeline_components = CClass(practice, "Integration testing pipeline components")
no_testing = CClass(practice, "No testing")


what_to_test_decision = CClass(decision, "What to test in a machine learning/data science pipeline?")
add_decision_option_link(what_to_test_decision, no_testing)
add_decision_option_link(what_to_test_decision, unit_testing_feature_engineering_logic)
add_decision_option_link(what_to_test_decision, unit_testing_different_methods_implemented_in_a_model)
add_decision_option_link(what_to_test_decision, testing_model_training_converges)
add_decision_option_link(what_to_test_decision, unit_testing_pipeline_components)
add_decision_option_link(what_to_test_decision, integration_testing_pipeline_components)

add_links({what_to_test_decision: model_and_application}, role_name="context",
          stereotype_instances=decide_for_some_instances_of)

add_links({testing: what_to_test_decision},
          role_name="next decision", stereotype_instances=consider_if_not_decided_yet)
```

added codes to source:

``` python
unit_testing_different_methods_implemented_in_a_model, unit_testing_feature_engineering_logic, testing_model_training_converges,  
unit_testing_pipeline_components, integration_testing_pipeline_components, no_testing, what_to_test_decision
```


>Continuous delivery

>In this level, your system continuously delivers new pipeline implementations to the target environment that in turn
delivers prediction services of the newly trained model. For rapid and reliable continuous delivery of pipelines and
models, you should consider the following:

>    Verifying the compatibility of the model with the target infrastructure before you deploy your model. For example, you need to verify that the packages that are required by the model are installed in the serving environment, and that the memory, compute, and accelerator resources that are available.

>    Testing the prediction service by calling the service API with the expected inputs, and making sure that you get the response that you expect. This test usually captures problems that might occur when you update the model version and it expects a different input.

>    Testing prediction service performance, which involves load testing the service to capture metrics such as queries per seconds (QPS) and model latency.

>    Validating the data either for retraining or batch prediction.

>    Verifying that models meet the predictive performance targets before they are deployed.

>    Automated deployment to a test environment, for example, a deployment that is triggered by pushing code to the development branch.

>    Semi-automated deployment to a pre-production environment, for example, a deployment that is triggered by merging code to the main branch after reviewers approve the changes.

>    Manual deployment to a production environment after several successful runs of the pipeline on the pre-production environment.


**OPEN CODING TRACE:**
- run time testing of services serving the model
- performance testing of services serving the model
- data validation
- model validation
- automated provisioning and deployment to testing environment, pre-production environment,
production environment 

**AXIAL CODING TRACE:**
- data_validation, model_validation already coded.
- deployment already coded.

added:

``` python
prediction_service_api_testing = CClass(practice, "Prediction service API testing")
prediction_service_performance_testing = CClass(practice, "Prediction service performance testing")

automated_provisioning_can_include_test_environment_provisioning = \
    automated_provisioning.add_links(test_environment_provisioning, role_name="to", stereotype_instances=can_include)[0]
automated_provisioning_can_include_pre_production_environment_provisioning = \
    automated_provisioning.add_links(pre_production_environment_provisioning, role_name="to",
                                     stereotype_instances=can_include)[0]
automated_provisioning_can_include_production_environment_provisioning = \
    automated_provisioning.add_links(production_environment_provisioning, role_name="to", 
                                     stereotype_instances=can_include)[0]
```

added codes to source:

``` python
prediction_service_api_testing, prediction_service_performance_testing,
test_environment_provisioning, pre_production_environment_provisioning,
production_environment_provisioning,
automated_provisioning_can_include_test_environment_provisioning, 
automated_provisioning_can_include_pre_production_environment_provisioning,
automated_provisioning_can_include_test_production_environment_provisioning,
automated_provisioning, no_automated_provisioning, ci_cd_pipeline_enables_automated_provisioning,
provisioning_decision

```

>To summarize, implementing ML in a production environment doesn't only mean deploying your model as an API for
prediction. Rather, it means deploying an ML pipeline that can automate the retraining and deployment of new models.
Setting up a CI/CD system enables you to automatically test and deploy new pipeline implementations. This system lets
you cope with rapid changes in your data and business environment. You don't have to immediately move all of your
processes from one level to another. You can gradually implement these practices to help improve the automation of your
ML system development and production.


**OPEN CODING TRACE:**
- just a summary.
